from typing import List, Optional, Tuple

import numpy as np
from ase import Atoms


def q_l_global(atoms: Atoms, l: int, r_cut: float, p_1: int, p_2: int, filter: Optional[fastatomstruct.Filter] = None) -> float:
    """Implementation of the global bond orientational parameter.

    The global bond orientational parameter :math:`Q_l` (as first introduced
    by Steinhardt *et al.*) is defined as

    .. math::

        Q_l = \sqrt{\frac{4\pi}{2l + 1}} \sum_{m = -l}^l \left|\frac{1}{N} \sum_i q_{lm}(i)\right|^2,

    with :math:`q_{lm}(i)` the local average of spherical harmonics, which are in
    turn defined as

    .. math::

        q_{lm}(i) = \frac{1}{N_i} \sum_j f(\vec{r}_{ij}) Y_{lm}(\hat{\vec{r}}_{ij}),

    where :math:`\hat{r}_{ij}` denotes the normalized distance vector between
    atoms :math:`i` and :math:`j`. :math:`r_\mathrm{cut}`, :math:`p_1` and :math:`p_2` are parameters
    for the radial cutoff function :math:`f`, which is defined as

    .. math::

        f(r) = \frac{1 - (r / r_c)^{p_1}}{1 - (r / r_c)^{p_2}}.

    For :math:`p_1 \rightarrow \infty` and :math:`p_2 \rightarrow \infty`, this function
    becomes a sharp cutoff.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        l (int): Angular momentum index
        r_cut (float): Cutoff radius
        p_1 (int): First exponent of the cutoff function
        p_2 (int): Second exponent of the cutoff function
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        float, value of :math:`Q_l`

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.q_l_global(atoms, 4, 3.2, 48, 96)
    15.470256244398556
    """
    ...


def q_l(atoms: Atoms, l: int, r_cut: float, p_1: int, p_2: int, filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Local bond orientational parameter.

    The local bond orientational parameter is defined as

    .. math::

        q_l(i) = \sqrt{\frac{4\pi}{2l + 1} \sum_{m=-l}^l \left|q_{lm}\right|^2},

    with :math:`q_{lm}(i)` the local average of spherical harmonics, which are in
    turn defined as

    .. math::

        q_{lm}(i) = \frac{1}{N_i} \sum_j f(r_{ij}) Y_{lm}(\hat{\vec{r}}_{ij}),

    where :math:`\hat{r}_{ij}` denotes the normalized distance vector between
    atoms :math:`i` and :math:`j`. :math:`r_\mathrm{cut}`, :math:`p_1` and :math:`p_2` are parameters
    for the radial cutoff function :math:`f`, which is defined as

    .. math::

        f(r) = \frac{1 - (r / r_c)^{p_1}}{1 - (r / r_c)^{p_2}}.

    For :math:`p_1 \rightarrow \infty` and :math:`p_2 \rightarrow \infty`, this function
    becomes a sharp cutoff.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        l (int): Angular momentum index
        r_cut (float): Cutoff radius
        p_1 (int): First exponent of the cutoff function
        p_2 (int): Second exponent of the cutoff function
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        np.ndarray of floats, values of :math:`q_l`

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.q_l(atoms, 4, 3.2, 48, 96)
    array([0.03066405, 0.03077848, 0.03239637, 0.03255228, 0.03221065,
           0.03088525, 0.0320593 , 0.03041014, 0.02809627, 0.0320796 ,
           0.03061929, 0.0317437 , 0.03149879, 0.03133172, 0.03001349,
           0.03287695, 0.03220788, 0.03214616, 0.02995588, 0.0314777 ,
           0.03072696, 0.03285761, 0.03240999, 0.03223181, 0.03257093,
           0.03330183, 0.03247856, 0.03291686, 0.03252548, 0.03264648,
           0.0310145 , 0.03235084, 0.02920417, 0.03120152, 0.03217064,
           0.03260268, 0.0318103 , 0.03170456, 0.03185368, 0.03282919,
           0.03190343, 0.0321001 , 0.03207043, 0.03058873, 0.03000755,
           0.03056225, 0.03347534, 0.03159053, 0.0282027 , 0.03249406,
           0.03048561, 0.03232488, 0.03155685, 0.03294293, 0.03055524,
           0.03242173, 0.0333785 , 0.0320526 , 0.03054413, 0.03175122,
           0.03199091, 0.03207626, 0.03284527, 0.03223664, 0.0323601 ,
           0.03075725, 0.03183552, 0.03236343, 0.03261055, 0.02926757,
           0.02985864, 0.0319282 , 0.03202976, 0.03082561, 0.03271937,
           0.03158466, 0.03189446, 0.03291091, 0.03301897, 0.03233473,
           0.0307242 , 0.03157726, 0.03329316, 0.02992685, 0.03231233,
           0.02922886, 0.03142811, 0.03095213, 0.03066173, 0.03209566,
           0.0324924 , 0.03232818, 0.03153572, 0.03199383, 0.02873217,
           0.02697654, 0.03250349, 0.03305962, 0.03209492, 0.03154451,
           0.02753824, 0.02960509, 0.03167802, 0.03073506, 0.03111497,
           0.03156074, 0.03170319, 0.03167071, 0.03067202, 0.03225529,
           0.03145943, 0.02875677, 0.0322366 , 0.03122062, 0.03230477,
           0.0309693 , 0.03227595, 0.03172356, 0.03181167, 0.0304437 ,
           0.03305966, 0.03127027, 0.03109714, 0.03080434, 0.03198244,
           0.03193362, 0.03184303, 0.02713633, 0.03179225, 0.03167141,
           0.03095199, 0.0329414 , 0.03211338, 0.02671768, 0.03137277,
           0.03234054, 0.03223452, 0.03166622, 0.03305815, 0.03197315,
           0.03227503, 0.03174525, 0.03194392, 0.0323852 , 0.03070922,
           0.03285023, 0.02658146, 0.03214794, 0.03266306, 0.03257127,
           0.02970597, 0.03299539, 0.03187859, 0.03248672, 0.03248273,
           0.03125455, 0.03127503, 0.03080308, 0.03198665, 0.03175511,
           0.03224295, 0.03095405, 0.03263247, 0.03167989, 0.03194748,
           0.03230611, 0.03187082, 0.03205136, 0.03233821, 0.03120753,
           0.03280628, 0.03224879, 0.03416126, 0.03177875, 0.03098086,
           0.03105586, 0.03365781, 0.03158097, 0.03011997, 0.03164216,
           0.03151952, 0.03188848, 0.03201688, 0.02929834, 0.03167002,
           0.03170435, 0.0316324 , 0.031591  , 0.03245158, 0.02926124,
           0.03116223, 0.03183879, 0.03105355, 0.03160849, 0.03461491,
           0.03096321, 0.03140959, 0.02974125, 0.0325996 , 0.03122526,
           0.03139527, 0.03074977, 0.03258278, 0.03329818, 0.03208158,
           0.03211836, 0.03128247, 0.03217344, 0.03215197, 0.03156444,
           0.0319983 , 0.03037679, 0.03098777, 0.03216577, 0.03182811,
           0.02860384, 0.03176324, 0.0317403 , 0.03135506, 0.03369258,
           0.03273636, 0.02437022, 0.03055305, 0.0297607 , 0.03229257,
           0.03096657, 0.0316618 , 0.03228087, 0.03272197, 0.03158444,
           0.03199783, 0.03064851, 0.03216606, 0.03258162, 0.03151896,
           0.03204917, 0.03154017, 0.0325791 , 0.03273132, 0.03260538,
           0.02938942, 0.03230763, 0.03143792, 0.03101338, 0.03214409,
           0.0314844 , 0.03059686, 0.0307613 , 0.03264335, 0.03292901,
           0.02915667, 0.03236356, 0.0311953 , 0.03081446, 0.03243918,
           0.03227337, 0.03281752, 0.03144328, 0.03215313, 0.03315175,
           0.03012641, 0.0322154 , 0.02824831, 0.0319183 , 0.03146026,
           0.03119672, 0.03245559, 0.03220978, 0.03209589, 0.03083016,
           0.03243092, 0.03180317, 0.03099999, 0.03238761, 0.03014412,
           0.03261338, 0.02524012, 0.03106275, 0.03234553, 0.03245146,
           0.03191523, 0.03166196, 0.03224755, 0.03027243, 0.03268333,
           0.02918048, 0.0293349 , 0.03036061, 0.03069006, 0.03273597,
           0.0318668 , 0.03271231, 0.03175444, 0.03374568, 0.03199945,
           0.03070458, 0.03247144, 0.02921717, 0.03222483, 0.02974489,
           0.03227236, 0.03245493, 0.03099169, 0.0308125 , 0.03299066,
           0.03370813, 0.03185129, 0.03201637, 0.03118289, 0.03259671,
           0.03259001, 0.03123411, 0.03234885, 0.03208167, 0.03225207,
           0.03182264, 0.03513544, 0.02732037, 0.03265918, 0.03192159,
           0.03174079, 0.03025551, 0.03137922, 0.03091808, 0.03242879,
           0.0277702 , 0.03037641, 0.0323894 , 0.03221323, 0.0318261 ,
           0.03228694, 0.03146924, 0.03266115, 0.03224339, 0.03213946,
           0.02755624, 0.03237154, 0.03257062, 0.03153718, 0.02985715,
           0.02285653, 0.0293093 , 0.03199774, 0.03135969, 0.03256725,
           0.03235009, 0.033446  , 0.02401419, 0.03098639, 0.03074289,
           0.03090236, 0.03082822, 0.03277135, 0.03253281, 0.03182146,
           0.03188528, 0.03182423, 0.0319413 , 0.03106253, 0.02968627,
           0.0323163 , 0.03187316, 0.03128458, 0.03243256, 0.03064163,
           0.03367722, 0.03135681, 0.03230078, 0.03241147, 0.03226453,
           0.0320935 , 0.0278087 , 0.03041577, 0.030921  , 0.03267095,
           0.03091002, 0.02973751, 0.03223354, 0.03208016, 0.03215956,
           0.03160335, 0.03174945, 0.03232391, 0.03220363, 0.02806145,
           0.03056179, 0.0302479 , 0.03341277, 0.03122013, 0.03191627,
           0.03058342, 0.02888386, 0.0320222 , 0.03210752, 0.02881538,
           0.030285  , 0.02849879, 0.03152319, 0.03201023, 0.03052745,
           0.03164701, 0.03233717, 0.03075602, 0.03221944, 0.03145295,
           0.03175731, 0.03125165, 0.03181228, 0.03184885, 0.02901128,
           0.03215691, 0.02857147, 0.03173312, 0.03205166, 0.03155986,
           0.0317861 , 0.03275808, 0.03055513, 0.03289854, 0.03085925,
           0.03017929, 0.03385005, 0.03238612, 0.02788997, 0.03204204,
           0.03120085, 0.03208455, 0.03269098, 0.03076184, 0.03145105,
           0.03205821, 0.03218099, 0.03233821, 0.03190587, 0.03104696,
           0.0314549 , 0.03211891, 0.03133587, 0.03126657, 0.03152677,
           0.03151355, 0.0317432 , 0.03155641, 0.03139033, 0.03327719,
           0.03145099, 0.03189376, 0.03193105, 0.03201754, 0.03084169,
           0.03119499, 0.03242395, 0.03160912, 0.0329967 , 0.03033413,
           0.03255113, 0.03007015, 0.03309998, 0.03174779, 0.03201106,
           0.032312  , 0.03204305, 0.0315872 , 0.03040491, 0.02628055,
           0.03183673, 0.03082321, 0.03005789, 0.02244099, 0.03209323,
           0.03057224, 0.03258145, 0.03193024, 0.03082637, 0.03228409,
           0.03172198, 0.03129977, 0.03264497, 0.03260552, 0.03131848,
           0.03228375, 0.03194201, 0.03245168, 0.03099566, 0.03121766,
           0.03250185, 0.03292308, 0.03259912, 0.03309885, 0.03146099,
           0.03186152, 0.03242588, 0.02464734, 0.0321034 , 0.03225078,
           0.03214629, 0.03190498, 0.03390153, 0.02958538, 0.0294802 ,
           0.03302265, 0.03081354, 0.03278828, 0.03027757, 0.02529274,
           0.03158341, 0.03130699, 0.03246764, 0.03244246, 0.03237262,
           0.03083684, 0.02687324, 0.02912212, 0.03230346, 0.02907678,
           0.03228925, 0.03185242, 0.03073456, 0.0320701 , 0.03206213,
           0.03108895, 0.03250206, 0.0288546 , 0.03170712, 0.03162076,
           0.03147254, 0.03025816, 0.03161468, 0.03271597, 0.03111142,
           0.03082953, 0.0323657 , 0.02744053, 0.03186703, 0.02728237,
           0.03242957, 0.03042029, 0.03080642, 0.02996402, 0.02997427])
    """


def qdot(atoms: Atoms, l: int, r_cut: float, p_1: int, p_2: int, filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Implementation of the Local bond order correlation parameter.

    "Bond order correlation" here is a term for the correlation between
    bond orientional parameters. It is defined as

    .. math::

        \dot{q}_l(i) = \frac{1}{N_i^\mathrm{eff}} \sum_{j} f(r_{ij}) C_{ij},

    where :math:`N_i^\mathrm{eff}` is the effective coordination number

    .. math::

        N_i^\mathrm{eff} = \sum_j f(r_{ij}),

    where

    .. math::

        C_{ij} \sum_{m = -l}^l \frac{q_{lm}(i) \, q_{lm}^*(j)}{\sqrt{\sum_m \left|q_{lm}(i)\right|^2} \, \sqrt{\sum_m \left|q_{lm}(j)\right|^2}}

    are the bond order correlators, with :math:`q_{lm}(i)` the local average of
    spherical harmonics, which are in turn defined as

    .. math::

        q_{lm}(i) = \frac{1}{N_i} \sum_j f(r_{ij}) Y_{lm}(\hat{\vec{r}}_{ij}),

    where :math:`\hat{r}_{ij}` denotes the normalized distance vector between
    atoms :math:`i` and :math:`j`. :math:`r_\mathrm{cut}`, :math:`p_1` and :math:`p_2` are parameters
    for the radial cutoff function :math:`f`, which is defined as

    .. math::

        f(r) = \frac{1 - (r / r_c)^{p_1}}{1 - (r / r_c)^{p_2}}.

    For :math:`p_1 \rightarrow \infty` and :math:`p_2 \rightarrow \infty`, this function
    becomes a sharp cutoff.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        l (int): Angular momentum index
        r_cut (float): Cutoff radius
        p_1 (int): First exponent of the cutoff function
        p_2 (int): Second exponent of the cutoff function
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        np.ndarray of floats, values of :math:`q_l`

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.qdot(atoms, 4, 3.2, 48, 96)
    array([0.93074931, 0.88040204, 0.94854139, 0.9500958 , 0.95666451,
           0.93984924, 0.97383854, 0.94550999, 0.9620986 , 0.96789602,
           0.91860531, 0.97485862, 0.89748376, 0.9183081 , 0.96353066,
           0.85456477, 0.95899747, 0.83823261, 0.8157547 , 0.97120092,
           0.96216483, 0.95710621, 0.95312623, 0.97450635, 0.96789501,
           0.90517486, 0.96345246, 0.95920604, 0.96421362, 0.93846943,
           0.96342169, 0.9481619 , 0.89453617, 0.96195208, 0.98005757,
           0.96533463, 0.96346126, 0.92271432, 0.95840263, 0.96417697,
           0.93875022, 0.96647939, 0.92858232, 0.90703077, 0.93772954,
           0.88553929, 0.8511439 , 0.94615439, 0.93218294, 0.96407731,
           0.95725358, 0.91707666, 0.97439081, 0.8617152 , 0.95777445,
           0.97609547, 0.88656155, 0.94243366, 0.97629491, 0.96071184,
           0.96671587, 0.9054203 , 0.9625537 , 0.9582114 , 0.96786281,
           0.93600994, 0.96771239, 0.97898286, 0.95567905, 0.89188261,
           0.96454509, 0.9552257 , 0.96080261, 0.92825584, 0.87489002,
           0.95312328, 0.91222359, 0.96287196, 0.96252531, 0.96963602,
           0.94881132, 0.85243743, 0.93150865, 0.89509333, 0.98046074,
           0.93036275, 0.93695242, 0.86858084, 0.89824549, 0.97764027,
           0.94979049, 0.96963166, 0.950274  , 0.97736454, 0.88075313,
           0.84497102, 0.95867277, 0.95224872, 0.96819875, 0.93727963,
           0.8747107 , 0.94823839, 0.9591061 , 0.97162712, 0.94863278,
           0.96018852, 0.95692253, 0.97372862, 0.97418654, 0.91276668,
           0.93934523, 0.94037201, 0.97051159, 0.93092624, 0.89585316,
           0.93215658, 0.97361331, 0.96918797, 0.94318347, 0.92088637,
           0.9726926 , 0.90812961, 0.87114374, 0.97041419, 0.97299673,
           0.97372962, 0.96058264, 0.75252446, 0.93198139, 0.96520358,
           0.96895434, 0.94079873, 0.97196352, 0.89265877, 0.91871339,
           0.93910698, 0.94628447, 0.87562954, 0.92905645, 0.93155226,
           0.98175263, 0.97663471, 0.90315028, 0.94118216, 0.96994061,
           0.93643945, 0.80984377, 0.94098223, 0.94918481, 0.94366804,
           0.92173119, 0.88336927, 0.98234253, 0.96083778, 0.97665125,
           0.94240832, 0.96394718, 0.85690951, 0.96692701, 0.93130712,
           0.94949769, 0.91969467, 0.9371281 , 0.93674363, 0.97784218,
           0.98001784, 0.96557908, 0.93871116, 0.9538552 , 0.93988204,
           0.92511209, 0.95485204, 0.91524044, 0.95883562, 0.90592291,
           0.95640569, 0.94145079, 0.96150196, 0.93782448, 0.93724224,
           0.94598515, 0.9665694 , 0.96770541, 0.93323636, 0.93096864,
           0.97167064, 0.94512135, 0.95775302, 0.9616459 , 0.9319298 ,
           0.9439924 , 0.95157784, 0.93066066, 0.91663836, 0.93498163,
           0.93228979, 0.97245906, 0.87398993, 0.98054174, 0.94668625,
           0.89952025, 0.89709806, 0.97004063, 0.8964992 , 0.94906029,
           0.95398541, 0.93903863, 0.92103301, 0.96315409, 0.93187168,
           0.88698255, 0.94801505, 0.97028004, 0.98258125, 0.97168283,
           0.78861712, 0.77205075, 0.97448945, 0.96034923, 0.91468599,
           0.97830254, 0.58827697, 0.95164027, 0.93385115, 0.94199682,
           0.96634143, 0.96647825, 0.96301645, 0.9582172 , 0.97078051,
           0.96698654, 0.92427308, 0.90473151, 0.97018893, 0.88231616,
           0.97200627, 0.83428228, 0.95106254, 0.93409203, 0.96728004,
           0.95917149, 0.95086963, 0.94823959, 0.97328187, 0.72896902,
           0.95302877, 0.9472142 , 0.96525042, 0.93413294, 0.86474199,
           0.84407571, 0.97337316, 0.84675967, 0.96749718, 0.9601057 ,
           0.95189614, 0.96750414, 0.92986917, 0.89921394, 0.9067761 ,
           0.93149323, 0.85113662, 0.92751559, 0.96996148, 0.98119623,
           0.9495879 , 0.96671837, 0.97007107, 0.94221914, 0.97567362,
           0.90048158, 0.96239035, 0.94155591, 0.94425775, 0.92533371,
           0.96475551, 0.78984954, 0.97184291, 0.9683978 , 0.96704371,
           0.96175792, 0.95957572, 0.96842922, 0.74778862, 0.97692837,
           0.88488143, 0.95164203, 0.94166277, 0.96268317, 0.97801229,
           0.95706433, 0.9801232 , 0.93525252, 0.88234671, 0.95991132,
           0.93624803, 0.96176381, 0.94150445, 0.95495686, 0.93737267,
           0.91878198, 0.97604118, 0.96930163, 0.98271019, 0.91014299,
           0.88873706, 0.94148279, 0.96302934, 0.91942637, 0.97514819,
           0.97445807, 0.93985034, 0.96998877, 0.96839372, 0.98530747,
           0.93138532, 0.84336281, 0.63478894, 0.97468248, 0.95846076,
           0.96842292, 0.92903967, 0.92792945, 0.93925132, 0.94233619,
           0.92960406, 0.93211295, 0.95859677, 0.97291603, 0.9742509 ,
           0.92302005, 0.93934516, 0.93333744, 0.87561887, 0.95149112,
           0.84843858, 0.94710234, 0.9357063 , 0.95255014, 0.9431574 ,
           0.78899784, 0.66249482, 0.91981328, 0.93299198, 0.95444028,
           0.93949935, 0.96069485, 0.79645701, 0.93874778, 0.83253116,
           0.97571996, 0.93632558, 0.94911459, 0.96836686, 0.952001  ,
           0.92451612, 0.97547872, 0.95795743, 0.95698079, 0.87275851,
           0.98030823, 0.91024846, 0.91389624, 0.95470211, 0.9121267 ,
           0.88614014, 0.94265643, 0.92901972, 0.93654042, 0.98304611,
           0.97190506, 0.89381592, 0.87459399, 0.89833589, 0.98078199,
           0.97081768, 0.85289709, 0.97130687, 0.9475324 , 0.94724549,
           0.9717687 , 0.97118356, 0.96366714, 0.96689857, 0.83028455,
           0.90164448, 0.88336684, 0.94864971, 0.97032553, 0.95029164,
           0.95829113, 0.90987303, 0.95597077, 0.96080899, 0.85719597,
           0.91082379, 0.90375538, 0.95374528, 0.96900844, 0.83243418,
           0.89003442, 0.95605234, 0.88264812, 0.95798632, 0.97363501,
           0.95919538, 0.9598461 , 0.95535573, 0.95524033, 0.82842642,
           0.96057804, 0.83824241, 0.96619588, 0.97816694, 0.95458391,
           0.97079114, 0.97337539, 0.9588261 , 0.90937292, 0.95538196,
           0.89272932, 0.88155291, 0.87503891, 0.85106442, 0.94483368,
           0.95450254, 0.96527072, 0.96758765, 0.95888599, 0.97009959,
           0.96321172, 0.94131346, 0.95086913, 0.93224719, 0.96442782,
           0.85717644, 0.97394171, 0.92658201, 0.93579737, 0.92945626,
           0.87386349, 0.95795114, 0.96692435, 0.84815052, 0.9626444 ,
           0.92751976, 0.92444872, 0.89096114, 0.97065447, 0.90568462,
           0.95324605, 0.96459908, 0.97391712, 0.96385259, 0.94247746,
           0.96371208, 0.9086509 , 0.9410501 , 0.97445504, 0.96157802,
           0.90297722, 0.93804726, 0.97282922, 0.91694484, 0.82359582,
           0.94755344, 0.78666694, 0.95064405, 0.8212462 , 0.9668883 ,
           0.92155746, 0.98428667, 0.97496754, 0.91339764, 0.93314438,
           0.95154859, 0.94162004, 0.94489567, 0.94162457, 0.97105046,
           0.98148459, 0.94939038, 0.94387293, 0.93131337, 0.90570642,
           0.98398807, 0.97961887, 0.91615038, 0.61668456, 0.95258801,
           0.96916922, 0.97275159, 0.71477787, 0.93162664, 0.97846348,
           0.97471202, 0.95927884, 0.88267176, 0.92251993, 0.9235943 ,
           0.92665766, 0.90614527, 0.89316531, 0.95342934, 0.78994957,
           0.92662498, 0.93568074, 0.92384578, 0.96688795, 0.91501176,
           0.93913094, 0.71438013, 0.9563588 , 0.9543706 , 0.82772155,
           0.95710159, 0.91739189, 0.85809953, 0.97289239, 0.97380738,
           0.93593489, 0.97870654, 0.86888224, 0.98081876, 0.96911051,
           0.94622051, 0.86112157, 0.976464  , 0.89012292, 0.93765565,
           0.96879067, 0.95066715, 0.78451946, 0.96498826, 0.75011839,
           0.90396105, 0.89631392, 0.96508114, 0.95514805, 0.92246258])
    """
    ...


def q_tetrahedral(atoms: Atoms, r_cut: float, n_neighbours: int, filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Tetrahedral order parameter.

    The tetrahedral order parameter is defined as

    .. math::

        q = 1 - \frac{3}{8} \sum_{i > k} \left(\frac{1}{3} + \theta_{ijk}\right)^2,

    where the sum runs over all pairs of atoms bonded to a central atom :math:`j` and
    forming a bond angle :math:`\theta_{ijk}`. The parameter is calculated
    for a fixed number of nearest neighbours.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_cut (float): Cutoff radius
        n_neighbours (int): Number of nearest neighbours
        filter (fastatomstruct.Filter): Filter instance

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.q_tetrahedral(atoms, 3.2, 4)
    array([0.75314876,        nan, 0.61997984,        nan, 0.64238884,
           0.6789242 ,        nan,        nan,        nan, 0.61900927,
           nan,               nan,        nan,        nan, 0.7633785 ,
           nan,               nan,        nan,        nan,        nan,
           nan,               nan,        nan, 0.65320697, 0.66741293,
           nan,               nan,        nan,        nan, 0.65822003,
           0.72327292, 0.69133763,        nan, 0.71341097,        nan,
                  nan,        nan,        nan,        nan, 0.69707237,
           0.72445889, 0.68332159,        nan,        nan,        nan,
           0.6805243 ,        nan,        nan, 0.80949055, 0.62028419,
                  nan,        nan,        nan,        nan,        nan,
                  nan, 0.49616694, 0.6732165 ,        nan,        nan,
           0.72079846,        nan, 0.62577129, 0.63839512,        nan,
                  nan, 0.73058656, 0.61658821,        nan,        nan,
           0.70801976,        nan, 0.64658967,        nan, 0.68865111,
           0.69431552,        nan,        nan, 0.65366387,        nan,
           0.67493455, 0.5286997 , 0.59856394, 0.76045954, 0.68055592,
                  nan,        nan,        nan, 0.73822487,        nan,
                  nan,        nan, 0.67110814, 0.66277585,        nan,
           0.76509463, 0.67166402,        nan, 0.66001629, 0.71265648,
           0.69119043, 0.66380563, 0.7155508 ,        nan,        nan,
                  nan,        nan,        nan, 0.71494996, 0.69158548,
                  nan, 0.63157592,        nan,        nan,        nan,
           0.70413764, 0.6652314 ,        nan,        nan,        nan,
           0.64845741,        nan, 0.63545845,        nan,        nan,
           0.71347701, 0.70888986,        nan,        nan,        nan,
                  nan,        nan, 0.64556679,        nan,        nan,
                  nan,        nan,        nan, 0.65685347, 0.65867979,
           0.62592987, 0.71380732, 0.64597048, 0.6613329 ,        nan,
                  nan,        nan, 0.62652205,        nan, 0.60211913,
                  nan,        nan,        nan,        nan,        nan,
                  nan, 0.68617278, 0.56460962,        nan,        nan,
           0.58122629,        nan,        nan,        nan, 0.72124883,
                  nan,        nan,        nan,        nan,        nan,
           0.62691218, 0.69155699,        nan, 0.6631233 ,        nan,
                  nan,        nan, 0.61096032,        nan,        nan,
                  nan,        nan,        nan, 0.78059314, 0.70211886,
           0.69505764,        nan,        nan,        nan, 0.70609159,
                  nan, 0.62853116,        nan, 0.6548643 ,        nan,
                  nan, 0.68295401, 0.76151119, 0.6710223 , 0.71904714,
                  nan,        nan, 0.6128829 ,        nan,        nan,
                  nan,        nan,        nan,        nan, 0.70230306,
                  nan, 0.7275413 , 0.71799258,        nan,        nan,
                  nan,        nan, 0.70723123,        nan,        nan,
           0.67125496,        nan, 0.67210687,        nan, 0.68133094,
           0.66567756,        nan, 0.59758164,        nan, 0.70122185,
           0.69701188,        nan,        nan,        nan, 0.66272183,
           0.698581  , 0.65518162,        nan,        nan, 0.64675271,
           0.75312219, 0.69519417,        nan, 0.70959679,        nan,
                  nan,        nan,        nan, 0.63672152,        nan,
                  nan, 0.68971829,        nan,        nan,        nan,
           0.64517471, 0.64972213,        nan,        nan,        nan,
                  nan,        nan,        nan,        nan,        nan,
           0.74871899,        nan, 0.6135499 ,        nan,        nan,
           0.66285797,        nan,        nan, 0.6464939 ,        nan,
                  nan, 0.76312992,        nan,        nan,        nan,
           0.7034656 , 0.74290639,        nan,        nan,        nan,
           0.59646751, 0.75610767,        nan,        nan, 0.66683183,
                  nan,        nan,        nan,        nan,        nan,
           0.61768964,        nan,        nan, 0.69997568, 0.69998832,
                  nan,        nan, 0.69670876,        nan,        nan,
           0.63381717,        nan, 0.68015466,        nan,        nan,
                  nan,        nan,        nan, 0.68235795,        nan,
                  nan,        nan, 0.45549028, 0.63433507,        nan,
           0.71707856,        nan, 0.71938086,        nan, 0.61893993,
           0.66629521,        nan,        nan,        nan,        nan,
                  nan,        nan, 0.62082321,        nan, 0.70745124,
           0.6604696 , 0.66493729,        nan,        nan,        nan,
           0.86980182,        nan,        nan,        nan,        nan,
                  nan,        nan, 0.81697697,        nan, 0.69363947,
           0.72206935,        nan, 0.63839984,        nan, 0.5789441 ,
                  nan, 0.72598572,        nan, 0.74851452, 0.73875735,
                  nan,        nan, 0.73151063,        nan, 0.63500589,
                  nan, 0.68006004,        nan, 0.69396847,        nan,
           0.68877588,        nan,        nan,        nan, 0.62947143,
           0.70881352,        nan, 0.69267411,        nan, 0.65719532,
                  nan, 0.69409817, 0.65873937,        nan,        nan,
                  nan, 0.70932557,        nan, 0.70317808, 0.67691817,
                  nan, 0.56344672, 0.72865813,        nan, 0.64712833,
                  nan, 0.68457114,        nan, 0.67637263,        nan,
           0.6010237 , 0.67009225,        nan,        nan, 0.73356464,
           0.70128704, 0.71292779,        nan,        nan, 0.6533144 ,
           0.60545014,        nan,        nan,        nan, 0.7334633 ,
           0.66123734,        nan,        nan,        nan,        nan,
                  nan,        nan,        nan, 0.60373193,        nan,
           0.6754361 , 0.55106031,        nan,        nan,        nan,
                  nan,        nan,        nan,        nan, 0.74686557,
           0.72049668,        nan, 0.69844306,        nan, 0.63763053,
           0.73063015, 0.56731768,        nan,        nan,        nan,
                  nan,        nan,        nan, 0.68895463,        nan,
                  nan,        nan, 0.68260716,        nan,        nan,
           0.67232506, 0.75558901, 0.63782509, 0.66712424, 0.69027139,
           0.61709895, 0.60198663,        nan,        nan, 0.62541504,
           0.73110942,        nan,        nan, 0.78461065, 0.67674399,
                  nan,        nan,        nan,        nan,        nan,
           0.73605266, 0.74101595,        nan, 0.66931541, 0.69590138,
           0.6513575 , 0.6958255 , 0.5841596 ,        nan,        nan,
                  nan,        nan, 0.60461152,        nan,        nan,
           0.69444851,        nan, 0.83908583, 0.62186386,        nan,
                  nan, 0.72771316,        nan, 0.76388783,        nan,
           0.65433934,        nan,        nan, 0.71960111, 0.8380811 ,
           0.70347446,        nan,        nan,        nan,        nan,
                  nan,        nan,        nan,        nan,        nan,
                  nan,        nan,        nan, 0.62693729, 0.68916236,
                  nan,        nan,        nan,        nan, 0.66696749,
           0.70181677,        nan,        nan,        nan,        nan,
           0.68279653,        nan, 0.4026759 ,        nan, 0.77371227,
           0.6804065 , 0.6404297 , 0.69180815, 0.80101402,        nan])
    """
    ...


def bond_length_ratio(atoms: Atoms, r_min: float, r_max: float,
                      cutoff_angle: float, filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Angular limited bond length ratio.

    The angular limited bond length ratio is defined as

    .. math::

        \mathrm{BLR} = \frac{\sum_{i_1, i_2, i_3} \frac{\mathrm{max}(r_{i_1, i_2}, r_{i_2, i_3})}{\mathrm{min}(r_{i_1, i_2}, r_{i_2, i_3})} \Theta(\beta - \delta) \Theta(r_{i_1, i_2} - r_\mathrm{min}) \Theta(r_{i_2, i_3} - r_\mathrm{min}) \Theta(r_\mathrm{max} - r_{i_1, i_2}) \Theta(r_\mathrm{max} - r_{i_2, i_3})}{\sum_{i_1, i_2, i_3} \Theta(\beta - \delta) \Theta(r_{i_1, i_2} - r_\mathrm{min}) \Theta(r_{i_2, i_3} - r_\mathrm{min}) \Theta(r_\mathrm{max} - r_{i_1, i_2}) \Theta(r_\mathrm{max} - r_{i_2, i_3})},

    with :math:`i_1`, :math:`i_2` and :math:`i_3` distinct indices of atoms, and

    .. math::

        \cos\beta = \frac{\vec{r}_{i_1 i_2} \cdot \vec{r}_{i_2 i_3}}{r_{i_1 i_2} r_{i_2 i_3}}

    the alignment angle and :math:`\delta` and angular threshold (typically 25° in
    phase-change materials).

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_min (float): Minimum radius to consider
        r_max (float): Maximum radius to consider
        cutoff_angle (float): Cutoff angle :math:`\delta` in degrees.
        filter (fastatomstruct.Filter): Filter instance

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.bond_length_ratio(atoms, 2.5, 4, 25)
    1.1065865743550205
    """
    ...


def bond_length_ratio_list(atoms: Atoms, r_min: float, r_max: float,
                           cutoff_angle: float, filter: Optional[fastatomstruct.Filter] = None]) -> np.ndarray:
    """(Angular limited) bond length ratio for each set of atoms.

    The angular limited bond length ratio for a triplet of atoms
    :math:`i_1`, :math:`i_2`, and :math:`i_3` is defined as

    .. math::

        \mathrm{BLR}(i_1, i_2, i_3) = \frac{\mathrm{max}(r_{i_1, i_2}, r_{i_2, i_3})}{\mathrm{min}(r_{i_1, i_2}, r_{i_2, i_3})} \Theta(\beta - \delta) \Theta(r_{i_1, i_2} - r_\mathrm{min}) \Theta(r_{i_2, i_3} - r_\mathrm{min}) \Theta(r_\mathrm{max} - r_{i_1, i_2}) \Theta(r_\mathrm{max} - r_{i_2, i_3}),

    with :math:`i_1`, :math:`i_2` and :math:`i_3` distinct indices of atoms, and

    .. math::

        \cos\beta = \frac{\vec{r}_{i_1 i_2} \cdot \vec{r}_{i_2 i_3}}{r_{i_1 i_2} r_{i_2 i_3}}

    the alignment angle and :math:`\delta` and angular threshold (typically 25° in
    phase-change materials).

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_min (float): Minimum radius to consider
        r_max (float): Maximum radius to consider
        cutoff_angle (float): Cutoff angle :math:`\delta` in degrees.
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        For each atom :math:`i_2`, a list of tuples
        :math:`(i_1, i_3, r_{i_1, i_2}, r_{i_2, i_3}, \frac{\mathrm{max}(r_{i_1, i_2}, r_{i_2, i_3})}{\mathrm{min}(r_{i_1, i_2}, r_{i_2, i_3})})`
        is returned (under the restrictions described above).

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.bond_length_ratio_list(atoms, 2.5, 4, 25)
    [[(2, 251, 2.9182822534253523, 3.2642249277049435, 1.1185432539548013), (99, 87, 3.473728477611225, 2.9892034529681553, 1.162091685048054), (206, 364, 2.8675345151801683, 3.1418041578635356, 1.0956465009336198), (282, 89, 3.4439717143736956, 3.021147524496202, 1.1399548305566451), (338, 528, 3.102057576273999, 3.5687633935730907, 1.150450404553635), (443, 127, 3.035237755535734, 2.9442968584178857, 1.030887135873492)], [(10, 25, 3.3798045197911324, 3.218576284688825, 1.0500930289797046), (26, 239, 3.3279022062705366, 2.9063389361475354, 1.1450495896675423), (63, 278, 3.1304544745456697, 3.140447605846574, 1.0031922301960179), (94, 117, 2.998646400333901, 3.238029520920562, 1.0798303929933204), (138, 315, 2.9766155957930542, 3.5591587845414234, 1.1957065566584062), (149, 345, 3.2551179306418545, 2.9831983446226693, 1.091150354286476)], [(0, 282, 2.9182822534253523, 3.4439717143736956, 1.180136606159775), (251, 307, 3.2642249277049435, 3.1347144034757415, 1.0413149357675462), (332, 392, 3.185476740560508, 3.0499540147004445, 1.0444343505531095), (441, 39, 3.2378763170187073, 3.150029713250805, 1.027887547662922), (459, 143, 3.116380488876878, 3.0126371025145566, 1.0344360713992833), (495, 419, 3.1075904045375635, 3.293156185452739, 1.059713719235399)], [(219, 198, 3.0698019639035174, 3.3708741864174394, 1.098075454395463), (225, 331, 3.0629574486549846, 3.2575635316482447, 1.063535353087819), (257, 249, 3.089247696773982, 3.249167962016847, 1.0517667344738544), (369, 444, 3.4965535415377267, 2.927071438464555, 1.1945569539539163), (383, 316, 2.985689929339029, 3.5865210815797957, 1.2012369557657911), (389, 518, 3.1549015481471754, 2.9808921548419507, 1.0583749375241824)], [(20, 461, 3.1196711017730467, 2.9941942547357097, 1.0419067155842943), (113, 161, 3.1959164504516417, 3.34798364041703, 1.0475817163317578), (325, 100, 3.0235980412432073, 2.9908183859185455, 1.0109600955641425), (445, 466, 3.4527904914873053, 2.945113631390019, 1.1723793794189448), (533, 521, 2.8972093637887455, 3.435732804060332, 1.185876604915893)], [(109, 337, 3.3645028736306775, 3.098682299497014, 1.0857850364901274), (234, 403, 3.1747895988805253, 3.3977882959836343, 1.070240464811193), (373, 436, 3.7158129231622987, 2.886802082855901, 1.2871727318023343), (383, 330, 2.8348379712712966, 3.5939184854393775, 1.267768571558842), (389, 144, 3.085719004967335, 3.480052570340591, 1.1277930896295043), (536, 232, 2.9237407698223867, 3.230791796494345, 1.1050199216843055)], [(64, 257, 2.9845562292124654, 3.394637143330579, 1.1374009677232053), (122, 17, 3.0903527402931243, 2.9503508957347497, 1.0474526080137694), (198, 312, 3.045622388587091, 3.1940000634227967, 1.0487183425600377), (254, 490, 3.259401316558566, 3.137866884388211, 1.0387315449151218), (360, 328, 3.0230843061052886, 3.0098183851171116, 1.0044075486593391), (453, 294, 3.1867427394594463, 2.9972587553559586, 1.0632190943691093)], [(91, 279, 2.7473184058764906, 3.42296143213926, 1.2459281839402285), (106, 476, 2.986028732316778, 3.3444066323344708, 1.120018235638221), (281, 130, 3.6086358701979115, 2.8916348192678414, 1.247956984800596), (319, 19, 3.2937563409351998, 3.0745729576743233, 1.0712890493340812), (390, 477, 3.2434027105192733, 3.12461261058374, 1.0380175448096078)], [(36, 193, 2.9051824962686053, 3.150530354474014, 1.084451788664062), (165, 354, 3.5671471053931905, 3.138441254092852, 1.1365983354766516), (345, 538, 3.1725789213690248, 3.5269870427718897, 1.11170978884583), (388, 116, 3.1855174717881694, 3.065655673803655, 1.0390982584928719), (415, 392, 3.2284944458546883, 3.1023432693563335, 1.0406631908675046), (486, 26, 3.639768024261627, 3.003139989722938, 1.2119874653586904)], [(124, 164, 2.8847318362692635, 3.41114167505157, 1.1824813773550253), (144, 257, 3.1882368714976415, 3.1954917788659447, 1.0022755233254972), (240, 427, 3.4029994978804954, 2.809290099124834, 1.2113378746255565), (417, 30, 3.0623682304285205, 3.7138078303502016, 1.2127241242411022), (470, 119, 3.234958914758185, 3.07554276082464, 1.0518335026793126), (518, 373, 3.160926302000579, 3.247165256651446, 1.0272828109267482)], [(1, 138, 3.3798045197911324, 2.9766155957930542, 1.1354521304557827), (25, 503, 3.218576284688825, 3.4057142421753928, 1.0581430859280256), (134, 197, 3.5651044108211516, 2.8531216415113043, 1.2495451855087079), (199, 162, 2.9278889446082474, 3.1436349941357413, 1.0736865549237424), (248, 482, 2.8263874375632896, 3.234837412702951, 1.144513087523414), (506, 202, 3.2516528233106405, 3.3723670953921556, 1.0371239731425606)], [(38, 324, 3.2406885688551395, 3.0796662184477857, 1.0522856501275364), (72, 529, 3.0741102004064413, 3.4281753326837094, 1.115176460567502), (116, 388, 3.150100709860709, 3.065655673803655, 1.0275455057717817), (185, 489, 2.9835662994666445, 3.284938671259404, 1.1010107842572945), (289, 380, 3.270994141043471, 2.8664150959691543, 1.141144611484655), (382, 153, 3.380663794831617, 2.987514325396019, 1.1315975177402648)], [(83, 94, 3.974590146027631, 2.867031536624152, 1.3863084850149914), (139, 293, 3.111727276895743, 3.1804758304375054, 1.0220933736874094), (265, 209, 3.538367548005047, 2.872580339488628, 1.231773224708118), (447, 457, 2.8092387429973984, 3.3950573920852594, 1.2085328812113723), (531, 244, 2.901443117861705, 3.708078687206466, 1.2780118501648354)], [(115, 468, 2.971593517427857, 3.1374190947153466, 1.0558035869693996), (141, 67, 3.018433502261906, 3.326475326143656, 1.1020535399076754), (142, 191, 2.969232624117523, 3.1968346934062666, 1.0766534987659946), (161, 230, 3.176075588889058, 3.0810536976116962, 1.030840712497487), (169, 501, 3.1659410096102434, 3.0717759742918753, 1.0306549162785466), (283, 151, 3.6729268526725485, 3.214955296882028, 1.142450365090512)], [(158, 208, 3.4717993803038536, 2.970174945965188, 1.168887167747507), (240, 101, 3.1261454948224388, 3.0224623827131674, 1.0343041861173465), (313, 130, 2.9936814390045714, 3.3936783275755507, 1.1336137116526273), (321, 238, 3.0739316644302273, 3.411229790093119, 1.1097285699503057), (401, 225, 3.3724089410369094, 3.1195086569446957, 1.0810705504949323), (518, 19, 2.9764558343472927, 3.3604554697014994, 1.1290123746917327)], [(61, 525, 2.9180032632757515, 3.305610723129084, 1.1328331139075576), (120, 47, 3.345725942855324, 3.161473107910798, 1.0582806902527429), (121, 423, 3.636696127537174, 2.8492184720480767, 1.2763837393357351), (128, 484, 3.038507275641341, 3.3755193280258093, 1.1109136894573783), (236, 194, 3.0592673672961133, 3.422904152826007, 1.118864009539411), (270, 320, 3.6844097202822934, 2.921339561248391, 1.2612055678689456)], [(60, 179, 2.892993229604277, 3.494459921579365, 1.20790463172199), (187, 200, 3.1927847348862914, 3.315434297451441, 1.0384146044125702), (228, 275, 3.245796314530629, 3.264229398632157, 1.0056790637228243), (287, 308, 3.4145516921573895, 3.001105686360515, 1.1377645604671345), (310, 414, 3.275678363742696, 2.9114248400587095, 1.1251117729958784), (334, 37, 3.128632121115204, 3.8956693571498615, 1.2451669631779034)], [(122, 6, 2.9503508957347497, 3.0903527402931243, 1.0474526080137694), (214, 367, 3.7720150513603423, 2.95442797688108, 1.276732782412374), (223, 246, 3.1649698675555653, 3.244814784700271, 1.0252277021538827), (231, 166, 3.0831525375371696, 3.4025915291230717, 1.103607910311525), (300, 22, 3.498362568036368, 3.015505176090646, 1.1601248758497262), (349, 145, 3.4579590317432674, 3.1371255653115204, 1.1022698836091656)], [(28, 66, 3.318291022061406, 2.9520911517242228, 1.1240476162544295), (151, 59, 3.294217853981194, 3.225507331731531, 1.0213022371934208), (183, 169, 3.9697454362400424, 3.027145077130417, 1.3113826179758663), (449, 437, 3.085273073769242, 3.165586550174065, 1.026031237587247), (466, 212, 2.9940080841216705, 3.3897054178830555, 1.1321630812755363)], [(292, 259, 2.9410499602341855, 3.5547862934695837, 1.20867932933262), (318, 369, 3.3168786486747996, 2.985247183390596, 1.1110901191463622), (319, 7, 3.0745729576743233, 3.2937563409351998, 1.0712890493340812), (377, 103, 3.0503922175727975, 3.151348552946932, 1.0330961817934565), (417, 487, 3.5720858147372283, 2.9370897025884433, 1.2161990870041068), (518, 14, 3.3604554697014994, 2.9764558343472927, 1.1290123746917327)], [(4, 113, 3.1196711017730467, 3.1959164504516417, 1.0244401881452379), (324, 422, 3.0636997056461075, 3.468919414288381, 1.1322648260518131), (351, 185, 3.0476135717573363, 3.229131495583902, 1.059560675772256), (394, 418, 3.185813171002452, 2.8287032058806267, 1.1262451162707439), (461, 170, 2.9941942547357097, 3.0112431873090935, 1.0056939968228242), (477, 390, 3.586096311790499, 3.12461261058374, 1.1476930931033222)], [(86, 402, 3.383813441697338, 3.164999972600058, 1.0691353778804378), (157, 49, 3.508200462951603, 3.0396119786127955, 1.1541606256442836), (166, 171, 3.206937302750879, 3.066333256722596, 1.0458541307341673), (273, 223, 2.8675582106300412, 3.7614635037691992, 1.3117304785044817), (428, 44, 3.277287687316394, 2.9683909470503957, 1.1040620139921056), (510, 118, 2.8353344757561993, 3.346309910822877, 1.1802169865445584)], [(111, 329, 3.321571228071647, 3.004512920858778, 1.1055273568676296), (170, 266, 3.130159813213813, 3.732707886210436, 1.1924975429219289), (300, 17, 3.015505176090646, 3.498362568036368, 1.1601248758497262), (307, 351, 3.0829645094065925, 3.3987988580447452, 1.1024450160468906), (426, 263, 3.0756918560987794, 3.1444651982098106, 1.0223602835812895), (448, 332, 3.0202977558449122, 3.196709630771691, 1.058408769329244)], [(46, 460, 3.3464459920722014, 3.333916060095908, 1.0037583225703448), (50, 372, 3.1166044902166976, 3.3223938706022893, 1.0660299954747492), (123, 60, 3.187366516461131, 3.0214975824526973, 1.0548962656702805), (275, 129, 2.982592751339909, 3.5480032285932936, 1.1895701238458982), (284, 429, 2.9422042232503465, 3.320777572973555, 1.1286699769960178), (464, 359, 3.9016401787835564, 3.010156147040547, 1.296158733366532)], [(27, 442, 3.4449891696583856, 2.9186037394267608, 1.1803552236711008), (202, 213, 3.163734676694957, 3.0248307292823786, 1.0459212299279743), (358, 91, 2.87742930552752, 3.5711697122237127, 1.241097289640973), (476, 181, 2.8066976209087193, 3.5647963538210146, 1.2701034579802177), (500, 248, 3.3810140938077438, 3.1397777724992073, 1.0768322915785586), (527, 522, 2.9844632888820466, 3.6159434384678164, 1.2115891832002788)], [(10, 1, 3.218576284688825, 3.3798045197911324, 1.0500930289797046), (39, 368, 3.1407757784455432, 2.985321843980429, 1.0520727554982288), (163, 333, 2.957286324089136, 3.401176617330192, 1.1501005464453218), (416, 78, 3.21544374146933, 3.3201328519079913, 1.0325582155546664), (503, 459, 3.4057142421753928, 2.8535159371536993, 1.1935150590301224), (526, 237, 3.3131349517058823, 3.273137615172774, 1.0122198762275374)], [(1, 94, 3.3279022062705366, 2.998646400333901, 1.109801477726741), (152, 84, 3.4189587735602154, 3.1334339411342667, 1.0911220206935628), (196, 247, 3.3372922511155165, 2.852041702288446, 1.1701414633726117), (199, 416, 3.1670244662786793, 3.6331585286290076, 1.1471836000364233), (239, 535, 2.9063389361475354, 3.425529336320636, 1.1786406924931156), (486, 8, 3.003139989722938, 3.639768024261627, 1.2119874653586904)], [(24, 476, 3.4449891696583856, 2.8066976209087193, 1.2274172835700796), (117, 255, 2.978827605861224, 3.2422415693317173, 1.088428737182438), (242, 149, 3.2394840724468446, 3.1735042237025324, 1.0207908495131395), (357, 279, 3.169839083784378, 3.236736798443314, 1.021104451327248), (442, 452, 2.9186037394267608, 3.4299534232323894, 1.1752035320512753), (480, 177, 3.3331643809653517, 3.0629140493906934, 1.0882330771339859)], [(18, 466, 3.318291022061406, 2.9940080841216705, 1.1083106420652393), (66, 305, 2.9520911517242228, 3.4081645936396794, 1.1544916530267295), (182, 229, 3.1693985513724354, 3.2247423617637536, 1.0174619283420046), (186, 70, 3.3471396254720545, 2.938788473021369, 1.1389522097964606), (226, 471, 3.2049530458481486, 3.205020457075717, 1.000021033452473), (267, 473, 3.10231849511714, 3.2173104890362794, 1.0370664695130851)], [(140, 220, 3.2915077489049276, 2.9756658500494986, 1.106141588058409), (148, 101, 3.175498448864526, 3.7199427084169594, 1.171451590457275), (179, 71, 2.9089835298407305, 3.5383050937522866, 1.2163372729532134), (263, 98, 3.2527051357101393, 3.119034785065078, 1.0428563192963147), (348, 187, 3.0442734927040664, 3.2644519455338115, 1.072325450836604), (425, 111, 3.1184179142998456, 3.285750838754249, 1.053659557202734)], [(186, 381, 2.985424835075639, 3.4709898351454567, 1.1626451935298903), (267, 362, 2.9426803211503123, 3.422718168388607, 1.1631294584695646), (272, 422, 3.2841098483848894, 3.0051852055567467, 1.0928144602576892), (377, 505, 3.016247817801816, 3.2495160632417037, 1.0773372280829), (417, 9, 3.7138078303502016, 3.0623682304285205, 1.2127241242411022), (469, 433, 3.10832802994952, 3.0705243808612455, 1.0123117892578566)], [(80, 299, 2.9855491418379967, 3.299104346262454, 1.1050242985554821), (126, 232, 3.606139440288883, 2.9247999932243465, 1.2329524920141348), (233, 280, 2.9496629943236368, 3.022066890585477, 1.0245464978206578), (262, 291, 2.994822956170686, 3.4474034430476546, 1.1511209488842902), (427, 298, 3.090983443980383, 3.11587187111503, 1.0080519444978318), (438, 144, 3.267516251131788, 3.239737747426439, 1.0085743062775423)], [(54, 523, 3.3469163890282965, 3.0235403483911765, 1.106952778324651), (296, 131, 3.1811745701769287, 3.3456786776640586, 1.0517117510712342), (304, 285, 3.2648345676273656, 3.3579319369060863, 1.0285151873242928), (394, 325, 3.2637076959866436, 3.1516737559072543, 1.0355474420121693), (407, 173, 3.2410708053103465, 3.087396935437208, 1.049774574856011), (461, 327, 3.130450488688438, 3.2780115129259384, 1.0471373129109363)], [(75, 398, 3.0305463916862316, 3.3073730406252664, 1.0913454582640476), (126, 401, 3.5425560604879798, 2.8324151008252474, 1.250719239371315), (178, 195, 3.175564369794431, 3.254871878877083, 1.0249743037291308), (194, 496, 3.0216199233128562, 3.3400228459626757, 1.1053749084036777), (208, 61, 3.520439616344382, 2.999047205943419, 1.1738526854021116), (233, 180, 3.1705519308351504, 3.1340477576381454, 1.011647612295645)], [(152, 65, 3.1171613564427094, 3.520885617677185, 1.129516638720045), (192, 82, 3.3588054855493454, 2.988749157646431, 1.1238164557758756), (196, 243, 2.8866751297091824, 3.415431252338681, 1.1831713299455238), (314, 218, 3.0895549219632876, 3.089808114125526, 1.000081951015157), (457, 108, 3.1565278644537402, 3.0723647859199916, 1.027393582597825)], [(98, 306, 3.043039875352036, 3.205905299912042, 1.0535206343759018), (289, 251, 3.001260939592734, 3.183916336280324, 1.0608595521562267), (382, 458, 3.1028590957340003, 3.042233270060502, 1.0199280660921484), (419, 436, 2.907348597903755, 3.751010480560893, 1.2901825681534822), (465, 337, 3.5027637251728874, 2.9440209979967995, 1.1897889748599868), (491, 140, 3.19387010806388, 3.1828888208780963, 1.0034501007744134)], [(8, 486, 2.9051824962686053, 3.639768024261627, 1.252853488184139), (193, 406, 3.150530354474014, 2.7199861888244623, 1.1582890999294473), (320, 342, 3.0455042176619225, 3.2339308003943428, 1.0618704061021063), (494, 458, 3.077568949388999, 3.245500587540031, 1.0545663284601219), (519, 135, 2.87149701265598, 3.316883501369983, 1.1551060254463035)], [(52, 92, 3.147683005536353, 2.8996357511927076, 1.0855442806020088), (215, 520, 2.876625767784697, 3.496684518530334, 1.2155507183762548), (250, 80, 3.164685867059168, 3.373019414179183, 1.0658307193420156), (291, 352, 2.8019706415195613, 3.4187402807798977, 1.2201199506236977), (334, 16, 3.8956693571498615, 3.128632121115204, 1.2451669631779034), (344, 228, 3.588840422510059, 3.1328416909197405, 1.1455543486005022), (513, 311, 3.190457899044174, 3.103837140880805, 1.0279076363326163)], [(11, 72, 3.2406885688551395, 3.0741102004064413, 1.0541875071448883), (110, 147, 2.828566332947847, 3.2784801857011536, 1.1590607395388248), (324, 449, 3.0796662184477857, 3.298521965913676, 1.0710647621988718), (351, 171, 3.48211300565591, 2.9388719636411316, 1.184846787725222), (431, 387, 3.4230038077954377, 2.853350087854791, 1.199643823015396), (530, 426, 3.008249822504741, 3.2911670962045645, 1.094047133846171)], [(25, 163, 3.1407757784455432, 2.957286324089136, 1.0620465637235594), (248, 149, 3.3389567585642377, 3.02177306894235, 1.10496608527023), (368, 411, 2.985321843980429, 3.1225758557580945, 1.045976286293695), (441, 2, 3.150029713250805, 3.2378763170187073, 1.027887547662922), (521, 533, 3.0251419395034724, 3.435732804060332, 1.1357261486461858)], [(148, 210, 2.964149625378467, 3.3685529925674706, 1.136431495807965), (207, 59, 3.024758973861427, 3.2443680594912236, 1.0726038297687708), (260, 327, 2.9757449671413876, 3.3686674438791635, 1.1320417176459958), (399, 115, 3.9921446716183704, 3.6959380595128204, 1.0801438247438038), (425, 256, 3.1011763916750605, 3.194195550728337, 1.0299947978783088), (515, 102, 3.7596091047745603, 3.194631336885613, 1.176852258777294)], [(78, 533, 3.007363979491309, 3.097957167333639, 1.0301237856342396), (100, 522, 2.9973258262687343, 3.2229277128491907, 1.075267721848345), (166, 231, 2.907414386585913, 3.4025915291230717, 1.1703152962377097), (323, 163, 3.445453115205379, 2.9265452057431185, 1.1773107445748467), (355, 129, 3.4292607814026264, 2.9493032173810203, 1.162735917145816), (428, 372, 3.0829160513391756, 3.2363988564049233, 1.0497849446789436)], [(184, 326, 2.8629129738292907, 3.528025370917465, 1.2323201589318844), (316, 340, 3.162126716443487, 2.8655628893584533, 1.1034923463680915), (363, 224, 2.8056164804160364, 3.3473518933602904, 1.1930896174604455), (364, 206, 3.2040972148680726, 3.1418041578635356, 1.0198271610433214), (400, 238, 3.3520239412124013, 3.035598205147131, 1.1042383460132312), (474, 225, 3.725014117288715, 2.8235328067957193, 1.3192742469020715)], [(74, 136, 3.605322539716054, 2.8883012854260763, 1.2482501593265067), (127, 392, 3.6828543536474854, 3.6171605270102027, 1.018161711692564), (341, 61, 3.1519412680291485, 3.379598187480926, 1.0722275258619038), (452, 455, 2.8660303855669613, 3.309464821301094, 1.1547207726642479), (484, 242, 2.838550026015317, 3.5563582499848794, 1.2528784828137072), (507, 282, 3.3263598361885505, 3.051907528425515, 1.0899281204318225), (512, 85, 3.2552456555829954, 2.958975925017555, 1.1001257658301773)], [(50, 261, 3.2199867155264656, 3.0403540217572878, 1.059082821435825), (167, 197, 3.2660908617788658, 3.183681671402428, 1.0258848713163387), (275, 228, 3.2488882477997443, 3.264229398632157, 1.0047219693822347), (355, 162, 3.05019794342888, 3.340544281800449, 1.0951893430382345), (428, 21, 2.9683909470503957, 3.277287687316394, 1.1040620139921056), (534, 92, 3.1740998023237177, 3.3880606245526472, 1.0674083474225642)], [(53, 475, 3.13962136190888, 2.923933570547004, 1.0737663104027104), (119, 260, 2.8609727031815924, 3.5295764403472374, 1.233698048367296), (210, 409, 3.0321108513094845, 3.3056241653773113, 1.0902055787141502), (274, 273, 3.3683866811100054, 2.927402073541498, 1.1506402593460676), (353, 144, 3.4342162407358168, 2.8149725066754336, 1.2199821606043777), (402, 497, 3.108107611415448, 3.3027656053411603, 1.062629103706311)], [(23, 275, 3.3464459920722014, 2.982592751339909, 1.1219922634656823), (71, 179, 2.96001240251731, 3.5383050937522866, 1.1953683338431873), (216, 362, 3.096390401316349, 3.0213073853775305, 1.024851167511854), (261, 497, 3.200397618115541, 3.4655920409523864, 1.0828629609445208), (356, 472, 3.411824626277363, 3.0512526694826603, 1.1181717792174313), (460, 81, 3.333916060095908, 3.035301980251455, 1.0983803528569223)], [(120, 15, 3.161473107910798, 3.345725942855324, 1.0582806902527429), (141, 58, 2.9074090999535485, 3.28253533143852, 1.129024233806988), (182, 362, 3.4601819597784864, 2.96734800702388, 1.1660856601881682), (222, 434, 3.072079666271771, 3.173408548510842, 1.032983806817107), (226, 381, 3.2832403619994173, 3.0027611703910564, 1.0934070929030408), (283, 81, 3.2335245572314344, 3.0918664171169916, 1.0458163843464272)], [(105, 59, 3.2887539358062767, 3.1076210765937273, 1.0582866619668734), (123, 108, 3.09407719269586, 3.1610094064364813, 1.0216323671234275), (322, 286, 3.516994263410728, 2.9314608769303043, 1.1997411567346477), (344, 218, 2.9180771418688347, 3.241789867054261, 1.1109335735305852), (409, 210, 3.0359950620756333, 3.3056241653773113, 1.0888107845331407)], [(157, 21, 3.0396119786127955, 3.508200462951603, 1.1541606256442836), (328, 69, 3.004385266108441, 3.8006114238665196, 1.265021322910901), (370, 64, 2.9253914593024737, 3.314433775477454, 1.1329881219615452), (433, 469, 3.3252551078936508, 3.0705243808612455, 1.0829600079452737), (456, 217, 3.4467780534866264, 2.8585769726262877, 1.205767095478956), (462, 237, 3.054939485919205, 3.29331922587049, 1.0780309204323106)], [(23, 123, 3.1166044902166976, 3.187366516461131, 1.022704846401448), (44, 355, 3.2199867155264656, 3.05019794342888, 1.0556648372487976), (221, 504, 3.8718860490644866, 2.68197383390531, 1.4436703297088125), (261, 466, 3.0403540217572878, 3.265999695059311, 1.074216907533552), (372, 86, 3.3223938706022893, 2.952112913154284, 1.1254291310464701), (430, 290, 2.9861271819926913, 3.084356031551921, 1.0328950656059064)], [(167, 84, 3.298687795357158, 3.0053580600169183, 1.0976022588598273), (323, 416, 2.8722794723512184, 3.4144909846861813, 1.188773939846151), (333, 340, 3.286308026959834, 3.012836219994319, 1.0907688924975918), (428, 511, 2.935853915562805, 3.630256445025419, 1.236524891712637), (454, 147, 3.5150334477700254, 2.9326911743662105, 1.19856924537227), (510, 224, 3.0430856446771584, 3.284753526381704, 1.079415405914474)], [(37, 215, 3.147683005536353, 2.876625767784697, 1.0942274941659855), (55, 398, 3.0952191523283625, 3.0958134575487923, 1.000192007477074), (92, 150, 2.8996357511927076, 3.648145528192271, 1.2581392427278766), (155, 275, 3.4578806009925485, 3.0635402895715016, 1.1287204587331225), (191, 142, 3.1757888868798094, 3.1968346934062666, 1.006626953892749), (322, 180, 3.017535830608052, 3.0837141854361168, 1.0219312573381205)], [(45, 353, 3.13962136190888, 3.4342162407358168, 1.0938313397918225), (159, 147, 3.4952607218571887, 2.9231171875187254, 1.1957306182528127), (200, 224, 3.037308634563409, 3.032407579016425, 1.0016162258599068), (263, 426, 3.4841151475264738, 3.1444651982098106, 1.1080151720267188), (348, 60, 3.2993229664989334, 2.9115639298468405, 1.1331789533031105), (475, 238, 2.923933570547004, 2.9452721120778285, 1.007297888620237)], [(32, 296, 3.3469163890282965, 3.1811745701769287, 1.0521008247724517), (69, 501, 3.0628085891785957, 3.2904890729532625, 1.0743371572677116), (266, 450, 3.499960000892225, 2.8841524044147593, 1.2135142357716089), (360, 370, 2.980906007456529, 3.238852375125205, 1.0865328752478076), (453, 227, 3.3527425620921076, 3.1330831755778297, 1.070109656911284), (523, 79, 3.0235403483911765, 3.065079966624696, 1.01373873454529)], [(52, 155, 3.0952191523283625, 3.4578806009925485, 1.1171682620247991), (107, 58, 3.1994365460046432, 3.067402408187351, 1.0430442831579168), (286, 233, 2.88053965307213, 3.692140697545869, 1.281753123449683), (291, 190, 3.0145882945515963, 3.403744249819916, 1.129090913001838), (398, 133, 3.0958134575487923, 3.5352218190275853, 1.1419363173860961), (413, 264, 3.150639890119308, 3.025198504332868, 1.0414655056872386)], [(36, 519, 3.4739260348378753, 2.87149701265598, 1.2097961514592281), (244, 531, 3.0145031775174886, 3.708078687206466, 1.2300795417506085), (270, 317, 3.0795045638868705, 2.887524204671388, 1.0664861471654161), (345, 149, 3.1857967632202504, 2.9831983446226693, 1.067913157354345), (408, 411, 2.995784295616037, 3.0413975520219747, 1.0152258146464974)], [(129, 152, 3.1119931541549604, 3.3025898020996145, 1.061245844223719), (155, 464, 3.0088106883822476, 3.4026019884029313, 1.1308793875072327), (191, 107, 3.24503007244727, 3.123001280810733, 1.0390742047998323), (218, 478, 3.0483891808498598, 3.2861077443789393, 1.0779816976855974), (365, 100, 3.125629006355453, 2.9126757945281554, 1.07311256962665), (535, 161, 3.345474459818699, 3.0751574581586096, 1.087903466842948)], [(107, 55, 3.067402408187351, 3.1994365460046432, 1.0430442831579168), (141, 47, 3.28253533143852, 2.9074090999535485, 1.129024233806988), (142, 535, 3.276800903645684, 3.089094945862324, 1.0607640623137151), (442, 484, 3.0629626023982492, 3.309906628992037, 1.0806226058393382), (455, 227, 3.0175109184766953, 3.0655946913007805, 1.0159349126227384), (480, 527, 3.468568995112218, 2.8548439916389876, 1.2149767221153427)], [(105, 48, 3.1076210765937273, 3.2887539358062767, 1.0582866619668734), (135, 519, 2.9340395938380768, 3.316883501369983, 1.1304835518702392), (151, 18, 3.225507331731531, 3.294217853981194, 1.0213022371934208), (207, 40, 3.2443680594912236, 3.024758973861427, 1.0726038297687708), (229, 235, 3.1877120932423804, 3.079502894572144, 1.0351385279945549), (385, 115, 3.454512501724665, 2.933036898791437, 1.177793741070255)], [(16, 310, 2.892993229604277, 3.275678363742696, 1.132279996448787), (123, 23, 3.0214975824526973, 3.187366516461131, 1.0548962656702805), (179, 160, 3.494459921579365, 3.0512097312048057, 1.1452703122441656), (253, 195, 2.878770210330181, 3.3618227097050295, 1.1677982138489076), (348, 53, 2.9115639298468405, 3.2993229664989334, 1.1331789533031105), (409, 250, 3.702332572027588, 3.0025107782376232, 1.2330788614856323)], [(15, 121, 2.9180032632757515, 3.636696127537174, 1.2462961139579458), (208, 33, 2.999047205943419, 3.520439616344382, 1.1738526854021116), (317, 392, 3.979943222581933, 3.9536899322036336, 1.0066401996182), (339, 222, 3.0259542005620568, 3.3101128367031714, 1.093907117327927), (341, 43, 3.379598187480926, 3.1519412680291485, 1.0722275258619038), (525, 313, 3.305610723129084, 3.0476161020408923, 1.0846545668647114)], [(145, 311, 3.213869533024464, 3.322432720751312, 1.0337795876936806), (196, 138, 3.1103737829875664, 3.190715841768162, 1.0258303549303407), (211, 374, 2.895029503481432, 3.316338328284098, 1.1455283354784533), (239, 162, 3.1480452859286903, 3.174032509251582, 1.0082550347795347), (314, 228, 3.1320717708240733, 2.8758463135908254, 1.089095671080323), (414, 214, 3.3000992697356835, 3.121886577927251, 1.0570849348174447)], [(1, 149, 3.1304544745456697, 3.2551179306418545, 1.0398227979706613), (134, 163, 3.0716362432430757, 3.345502973928838, 1.0891598838528518), (152, 129, 3.0723944714507914, 3.3025898020996145, 1.0749237550021122), (278, 429, 3.140447605846574, 3.0550117659339753, 1.0279657973384202), (295, 326, 3.299438532626326, 3.007889682584272, 1.0969280395255605)], [(6, 453, 2.9845562292124654, 3.1867427394594463, 1.0677442456161503), (190, 291, 2.8941873122502497, 3.403744249819916, 1.176062183471284), (219, 330, 3.283238194919659, 3.495315045832326, 1.0645938059690052), (257, 144, 3.394637143330579, 3.1954917788659447, 1.0623207250232105), (370, 49, 3.314433775477454, 2.9253914593024737, 1.1329881219615452)], [(134, 201, 3.0607587106275296, 3.31327730607646, 1.0825019608935975), (152, 34, 3.520885617677185, 3.1171613564427094, 1.129516638720045), (167, 430, 3.4048959948538573, 2.9203111406261097, 1.1659360358854958), (199, 77, 2.9768770427598126, 3.399384245430149, 1.1419296788552062), (323, 446, 2.9263605379473026, 3.3418251517534565, 1.1419731466504746), (355, 365, 3.3203393469333964, 3.243894348882502, 1.023565810050882)], [(28, 18, 2.9520911517242228, 3.318291022061406, 1.1240476162544295), (103, 397, 2.8733136463853044, 3.550936878313788, 1.2358333670885353), (279, 357, 3.1042302508244948, 3.236736798443314, 1.0426857987044051), (305, 439, 3.4081645936396794, 2.9780279908864693, 1.1444367225793506), (309, 319, 3.1090535846118237, 3.2257112717365386, 1.0375219287638235), (524, 235, 3.2853184474249453, 3.019026654169993, 1.0882045187932143)], [(120, 524, 3.1466162555198802, 2.938311667693787, 1.0708926116029027), (128, 245, 2.844405997945784, 3.9111861785513002, 1.3750449764822392), (141, 13, 3.326475326143656, 3.018433502261906, 1.1020535399076754), (357, 358, 3.235161893017223, 3.1764504281928736, 1.0184833562341318), (442, 93, 2.882986538664661, 3.562320484730432, 1.2356354901262994), (493, 405, 3.007931449815717, 3.3414213646605444, 1.1108701845134334)], [(154, 107, 3.116026228599773, 3.136010449901598, 1.0064133674865776), (202, 506, 3.1449007720146405, 3.3723670953921556, 1.0723286169794792), (264, 397, 3.465128170219541, 2.9857630944241786, 1.1605502716175178), (456, 511, 2.999811588522652, 3.25867886704957, 1.086294512467833), (462, 168, 3.0832078877456577, 3.461591956464979, 1.1227241504613508), (527, 375, 3.0131714734575383, 3.268209964321089, 1.0846412137875778)], [(54, 453, 3.0628085891785957, 3.3527425620921076, 1.0946627791034333), (328, 49, 3.8006114238665196, 3.004385266108441, 1.265021322910901), (394, 171, 3.1378152343196124, 3.524441809197798, 1.1232152137734202), (412, 106, 3.2238490624033513, 3.148663117975456, 1.0238786880688078), (477, 437, 2.7849501053822183, 3.444861340858145, 1.2369562148350797), (501, 169, 3.2904890729532625, 3.0717759742918753, 1.0712008624625715)], [(124, 102, 3.323455581425549, 3.073603511923633, 1.0812896225985715), (186, 28, 2.938788473021369, 3.3471396254720545, 1.1389522097964606), (319, 97, 2.9823743972681838, 3.2659642435249667, 1.0950886134606532), (390, 412, 3.0678985125992235, 3.449065575640786, 1.1242437001994), (417, 132, 3.2222168092159738, 3.1562320514194386, 1.0209061807628688), (437, 327, 3.164414220873724, 3.346776810378991, 1.0576291777171054)], [(46, 261, 2.96001240251731, 3.200397618115541, 1.0812108812090782), (105, 180, 3.260562029233435, 3.003252404306952, 1.0856769895721972), (123, 228, 2.987055561154018, 3.5808298907629754, 1.1987824857799292), (135, 354, 3.265405628849882, 3.331922527861232, 1.0203701795647293), (179, 29, 3.5383050937522866, 2.9089835298407305, 1.2163372729532134), (404, 193, 2.9780131805371033, 3.5385316786672907, 1.1882189446955687)], [(11, 38, 3.0741102004064413, 3.2406885688551395, 1.0541875071448883), (76, 259, 2.973616277035373, 3.2747345415273905, 1.1012633226477444), (368, 482, 3.4833562945027334, 2.846856195892433, 1.2235799966042085), (441, 503, 3.18805468486839, 2.9365714675410746, 1.0856383779884282), (495, 376, 3.0651928149213448, 3.3767202501444666, 1.1016338788563669), (529, 281, 3.4281753326837094, 3.00833323643055, 1.1395597040809584)], [(136, 243, 3.0250031822478287, 3.117601552007677, 1.0306109991233265), (204, 300, 3.241952225849331, 3.008710102342345, 1.077522298783589), (332, 376, 2.9716826510510073, 3.2532322912133718, 1.0947441814026768), (367, 82, 3.460162683449088, 3.1923083249683812, 1.0839061679555528), (371, 156, 2.952897820531345, 3.5423003903093675, 1.1996014104111348), (459, 503, 3.4081667665448494, 2.8535159371536993, 1.1943745335953506)], [(43, 452, 3.605322539716054, 2.8660303855669613, 1.2579498660837978), (88, 149, 2.991722323200221, 3.340135863983737, 1.116459184089926), (127, 317, 2.7729076703834727, 3.3876511537322975, 1.2216963406011316), (136, 189, 2.8883012854260763, 3.089841060776554, 1.069777961311521), (315, 326, 3.1369662879401305, 2.9754396404389114, 1.0542866490403389), (371, 206, 3.5802724869024773, 2.893935180896006, 1.237164021688271)], [(33, 208, 3.0305463916862316, 3.520439616344382, 1.161651782002772), (139, 308, 2.948548325145212, 3.3986255230881954, 1.152643656576603), (265, 481, 3.1695144510532582, 3.1538784563710016, 1.004957703633338), (286, 322, 3.5543682459888606, 2.9314608769303043, 1.212490425494212), (331, 225, 3.0997946146784106, 3.2575635316482447, 1.0508965710898244), (398, 93, 3.3073730406252664, 3.024628397678057, 1.093480786983375)], [(72, 441, 2.973616277035373, 3.18805468486839, 1.0721136783818008), (97, 319, 3.450176908273754, 3.2659642435249667, 1.056403760425119), (259, 514, 3.2747345415273905, 3.4065781596211426, 1.0402608566959624), (382, 110, 3.1304691298917766, 3.253930202955873, 1.0394385211740977), (411, 408, 3.0560376426610816, 3.0413975520219747, 1.0048136063729565), (419, 252, 3.5042965039249507, 2.8209730184315753, 1.2422297133041331)], [(199, 65, 3.399384245430149, 2.9768770427598126, 1.1419296788552062), (248, 500, 3.170240749098018, 3.1397777724992073, 1.0097022715638129), (388, 494, 2.985766560918502, 3.464467831972416, 1.1603277621632457), (486, 243, 3.2614726317137017, 3.216672771382615, 1.0139273912875604), (508, 113, 3.148715476938977, 3.1317016777688336, 1.0054327649695756), (521, 185, 3.2656070472777854, 3.107583792498345, 1.050850842754717)], [(41, 428, 3.007363979491309, 3.0829160513391756, 1.025122357108449), (162, 239, 3.066698229696414, 3.174032509251582, 1.0349999483208994), (171, 296, 3.397173782885968, 2.819462194035446, 1.2049013425584025), (325, 501, 3.0988343895170893, 3.278577997272235, 1.0580036185099766), (416, 25, 3.3201328519079913, 3.21544374146933, 1.0325582155546664), (533, 431, 3.097957167333639, 2.964039885966948, 1.0451806610298038)], [(106, 358, 3.081674890455086, 3.1317976422666005, 1.0162647760043606), (181, 174, 3.0450987338632114, 3.4171870343650457, 1.1221925241254094), (281, 451, 3.478460778163109, 3.1001421615061404, 1.1220326671965166), (282, 507, 3.1394930703056216, 3.051907528425515, 1.0286986224400096), (338, 441, 2.8771110934075033, 3.3310830782579157, 1.1577874368113994), (523, 54, 3.065079966624696, 3.0235403483911765, 1.01373873454529)], [(31, 126, 2.9855491418379967, 3.606139440288883, 1.2078647072842457), (137, 353, 3.042575567561341, 3.301591968646269, 1.0851306386097528), (180, 362, 3.070815142469946, 3.015138496149572, 1.018465701124998), (250, 37, 3.373019414179183, 3.164685867059168, 1.0658307193420156), (299, 179, 3.299104346262454, 3.0235727713247336, 1.0911278132779985), (410, 434, 3.060880382596036, 3.286316004925123, 1.0736505822347386)], [(113, 509, 3.007910635776715, 3.4181004362623217, 1.136370341461187), (283, 47, 3.0918664171169916, 3.2335245572314344, 1.0458163843464272), (354, 329, 3.0128464704994413, 3.1534483454189997, 1.0466674542816152), (387, 388, 3.2827906620368275, 3.2600060772808077, 1.006989123399127), (445, 324, 3.2178384123229913, 3.140389686724783, 1.0246621385637595), (460, 46, 3.035301980251455, 3.333916060095908, 1.0983803528569223)], [(192, 34, 2.988749157646431, 3.3588054855493454, 1.1238164557758756), (205, 223, 3.4863663348362106, 3.0952362150380393, 1.1263651923875426), (361, 90, 3.328510619054082, 2.9132184594383976, 1.1425544171842654), (363, 118, 3.1182711826832055, 2.8688106452151203, 1.0869560833107474), (367, 73, 3.1923083249683812, 3.460162683449088, 1.0839061679555528)], [(12, 531, 3.974590146027631, 2.901443117861705, 1.3698666437951093), (94, 242, 2.867031536624152, 3.4075559351677263, 1.18853102647766), (201, 347, 3.137631146705446, 3.118274239974471, 1.006207570355048), (255, 413, 3.0077393645481805, 3.005284211061149, 1.0008169455248177)], [(152, 26, 3.1334339411342667, 3.4189587735602154, 1.0911220206935628), (167, 51, 3.0053580600169183, 3.298687795357158, 1.0976022588598273), (275, 155, 3.274266926053315, 3.0635402895715016, 1.0687853321854917), (278, 463, 2.8247435083217547, 3.6068117681996603, 1.2768634594871768), (284, 356, 3.1259377624169575, 3.1338139361798816, 1.0025196195067025), (457, 287, 3.286717100561865, 2.810868747235087, 1.1692887132474068)], [(293, 232, 3.28975875278027, 3.099990454425687, 1.0612157686110488), (315, 247, 3.746544006026854, 3.321169291275405, 1.1280798048653808), (352, 291, 2.9973612644970866, 3.4187402807798977, 1.1405833261655607), (502, 330, 3.7045030454841044, 2.89350683377068, 1.280281422614293), (512, 43, 2.958975925017555, 3.2552456555829954, 1.1001257658301773)], [(21, 166, 3.383813441697338, 3.206937302750879, 1.0551542241860283), (172, 337, 2.8904003288323894, 3.2353000717461717, 1.1193259423178616), (274, 246, 3.202585151156336, 3.0723624474724454, 1.0423852022377833), (372, 50, 2.952112913154284, 3.3223938706022893, 1.1254291310464701), (402, 306, 3.164999972600058, 3.180520120096998, 1.0049036801362718), (504, 433, 3.8935704552319867, 2.8820887701096494, 1.3509543826729027)], [(99, 0, 2.9892034529681553, 3.473728477611225, 1.162091685048054), (252, 491, 3.2170249092898984, 3.1625652639491992, 1.0172200858466058), (313, 102, 3.2455910841335984, 2.9497667229417783, 1.1002873748934279), (321, 536, 3.291857348924555, 2.821342850488343, 1.1667696991717158), (436, 97, 3.1995743239178007, 2.921272624069847, 1.095267280963401)], [(74, 371, 2.991722323200221, 3.5802724869024773, 1.196726199867603), (138, 295, 3.186117560709512, 2.987833648265123, 1.0663637724809485), (149, 248, 3.340135863983737, 3.02177306894235, 1.1053562884365824), (243, 211, 3.1138555369376864, 3.1330579498590096, 1.006166764223175), (245, 270, 3.3930475412230554, 3.0562400838633694, 1.110203206592962), (484, 442, 3.059134699492522, 3.309906628992037, 1.0819747916105542)], [(198, 219, 2.9659968409869086, 3.3708741864174394, 1.1365063306324397), (269, 102, 3.486821410172158, 2.8892631318875877, 1.2068203036579015), (277, 452, 3.13762354975632, 2.910912350305406, 1.0778832105429514), (282, 0, 3.021147524496202, 3.4439717143736956, 1.1399548305566451), (453, 304, 3.111111594828166, 3.1685842466260765, 1.018473349491369), (523, 412, 3.077438120403443, 3.2398145652119354, 1.0527635125242438)], [(140, 256, 3.207385638772465, 2.9498032339900373, 1.0873218938179854), (253, 108, 3.195365308638267, 3.2313670693024044, 1.0112668684756672), (348, 210, 3.460921519198629, 3.118018677582383, 1.1099745951111886), (361, 82, 2.9132184594383976, 3.328510619054082, 1.1425544171842654), (475, 465, 3.2967486865999818, 2.9802851928897, 1.1061856410471365)], [(7, 281, 2.7473184058764906, 3.6086358701979115, 1.3135120641564773), (279, 120, 3.42296143213926, 2.844312023537952, 1.203440903744992), (309, 377, 2.931539171743158, 3.3434612658195215, 1.140513931400557), (358, 24, 3.5711697122237127, 2.87742930552752, 1.241097289640973), (381, 366, 3.099734100953242, 3.190171728935117, 1.0291759309142237)], [(52, 37, 2.8996357511927076, 3.147683005536353, 1.0855442806020088), (150, 433, 3.648145528192271, 2.9171697386390814, 1.2505770507184146), (217, 177, 3.126436143745149, 2.9363640599676897, 1.0647304216696996), (359, 362, 3.124443377854147, 3.2332423573305014, 1.0348218758731602), (413, 255, 3.1743458236989808, 3.005284211061149, 1.0562547834962128), (534, 44, 3.3880606245526472, 3.1740998023237177, 1.0674083474225642)], [(107, 191, 3.3306042705064445, 3.123001280810733, 1.0664754737602342), (117, 202, 3.162124490867587, 2.9607226342498714, 1.0680245607230758), (268, 138, 3.091276377204276, 3.434482275089957, 1.1110240095051203), (393, 350, 3.1496268005280887, 3.288349748425666, 1.0440442492660775), (398, 75, 3.024628397678057, 3.3073730406252664, 1.093480786983375), (442, 67, 3.562320484730432, 2.882986538664661, 1.2356354901262994)], [(1, 26, 2.998646400333901, 3.3279022062705366, 1.109801477726741), (83, 12, 2.867031536624152, 3.974590146027631, 1.3863084850149914), (117, 154, 3.238029520920562, 3.2348445500343246, 1.0009845823615244), (242, 538, 3.4075559351677263, 2.905957443141829, 1.172610405293336), (268, 512, 2.9245739560747106, 3.2478332285572344, 1.110532090259189), (506, 526, 3.6797890631069325, 2.785726285015436, 1.320944230199753)], [(137, 101, 3.210319450990008, 3.4257515458891272, 1.0671061239194384), (249, 391, 3.0978376636386686, 3.1687908930455437, 1.0229041147764775), (250, 322, 3.048482905819008, 2.8700086817590424, 1.0621859526747417), (396, 225, 2.9816158304041585, 3.135638599010506, 1.0516574828439482)], [(318, 389, 3.3027819678555606, 2.9038180944479026, 1.1373928601693324), (377, 309, 3.0607790057577065, 3.3434612658195215, 1.0923563117526796), (397, 209, 2.974564914733546, 3.45691442261032, 1.162158003507542), (403, 146, 2.9545018363495466, 3.6995560479066225, 1.2521759175745284), (444, 462, 3.0969526171237685, 3.4133899540508272, 1.1021770030246518), (469, 467, 3.3097645139914116, 2.942517665628377, 1.124807015656305)], [(76, 382, 3.450176908273754, 3.1304691298917766, 1.1021277530991114), (241, 305, 3.218991588390329, 2.928901964930394, 1.0990438146900656), (292, 234, 3.163031138245725, 3.0142275902274607, 1.049367057915834), (319, 70, 3.2659642435249667, 2.9823743972681838, 1.0950886134606532), (436, 87, 2.921272624069847, 3.1995743239178007, 1.095267280963401), (499, 528, 2.9363395436192707, 3.2768631840781297, 1.1159687547711654)], [(35, 419, 3.043039875352036, 2.907348597903755, 1.0466718292901362), (110, 212, 3.187656288969435, 3.1962993527197385, 1.002711416466139), (159, 172, 3.21574604530358, 3.3423422558139357, 1.0393676020204525), (263, 29, 3.119034785065078, 3.2527051357101393, 1.0428563192963147), (306, 487, 3.205905299912042, 3.0511846555034943, 1.0507083844071758), (530, 307, 3.1779215889423913, 2.83193656894518, 1.122172588112057)], [(0, 206, 3.473728477611225, 2.8675345151801683, 1.2113990116673345), (87, 321, 2.9892034529681553, 3.291857348924555, 1.1012490118917388), (130, 269, 3.0207640003589113, 3.2509586519955556, 1.0762041164451424), (495, 289, 3.435410994004304, 3.024746876869135, 1.1357680936132548), (529, 499, 3.0861366116095033, 2.8548652815472364, 1.0810095423966646)], [(41, 323, 2.9973258262687343, 3.445453115205379, 1.1495090340226717), (325, 4, 2.9908183859185455, 3.0235980412432073, 1.0109600955641425), (365, 57, 2.9126757945281554, 3.125629006355453, 1.07311256962665), (418, 407, 3.481468509706455, 2.906931178035758, 1.1976439401151966), (522, 527, 3.2229277128491907, 3.6159434384678164, 1.1219436985979387)], [(137, 95, 3.4257515458891272, 3.210319450990008, 1.0671061239194384), (148, 29, 3.7199427084169594, 3.175498448864526, 1.171451590457275), (240, 14, 3.0224623827131674, 3.1261454948224388, 1.0343041861173465), (410, 229, 3.1810761909498813, 2.9864000926493626, 1.0651875476362624), (470, 487, 2.951011634577805, 3.25709060932503, 1.1037200162685956), (515, 294, 3.1357760424842356, 3.4323016058156246, 1.0945620986046167)], [(124, 70, 3.073603511923633, 3.323455581425549, 1.0812896225985715), (240, 438, 3.3381177300953238, 3.007080393963774, 1.1100859613850211), (269, 89, 2.8892631318875877, 3.486821410172158, 1.2068203036579015), (298, 339, 3.3506137930496003, 2.912216084512726, 1.1505374930343562), (313, 87, 2.9497667229417783, 3.2455910841335984, 1.1002873748934279), (515, 40, 3.194631336885613, 3.7596091047745603, 1.176852258777294)], [(66, 279, 2.8733136463853044, 3.1042302508244948, 1.0803659582133294), (267, 343, 2.9780404180252473, 3.133914635979198, 1.0523412029636965), (280, 233, 3.1032296190093924, 3.022066890585477, 1.0268566948920814), (377, 19, 3.151348552946932, 3.0503922175727975, 1.0330961817934565), (397, 264, 3.550936878313788, 2.9857630944241786, 1.1892895604962945), (496, 336, 3.1375064422967984, 3.042748022773085, 1.0311423814310305)], [(132, 318, 3.137544757187182, 3.1044817962266333, 1.0106500740319158), (353, 137, 3.105072923156939, 3.301591968646269, 1.0632896715641476), (391, 173, 3.02615423982134, 3.356642030605456, 1.1092104911359797), (402, 159, 3.185817224560477, 2.984894306050817, 1.0673132439236994), (487, 437, 3.2841396845126662, 3.0250655488778344, 1.0856424865672536)], [(48, 344, 3.2887539358062767, 2.9180771418688347, 1.1270277569495808), (59, 385, 3.1076210765937273, 3.454512501724665, 1.1116260369527313), (71, 404, 3.260562029233435, 2.9780131805371033, 1.094878306967524), (180, 233, 3.003252404306952, 3.1340477576381454, 1.0435512357015415), (216, 343, 3.3961414617425776, 3.2962451021039727, 1.0303061078725129), (299, 148, 3.0550982228439048, 3.5729074508533065, 1.1694902062845587)], [(7, 319, 2.986028732316778, 3.2937563409351998, 1.1030558096404064), (79, 282, 3.081674890455086, 3.1394930703056216, 1.018761933658095), (358, 357, 3.1317976422666005, 3.1764504281928736, 1.0142578771130168), (412, 69, 3.148663117975456, 3.2238490624033513, 1.0238786880688078), (476, 237, 3.3444066323344708, 2.902045101487362, 1.1524309634679313), (528, 368, 3.154855562587185, 3.473946735876745, 1.101142878638756)], [(55, 286, 3.1994365460046432, 2.88053965307213, 1.1107073435328016), (58, 141, 3.067402408187351, 3.28253533143852, 1.0701352136507905), (93, 268, 3.3306042705064445, 3.091276377204276, 1.0774204128323894), (154, 68, 3.136010449901598, 3.116026228599773, 1.0064133674865776), (191, 57, 3.123001280810733, 3.24503007244727, 1.0390742047998323), (490, 254, 3.158648319706941, 3.137866884388211, 1.0066227906040641)], [(123, 48, 3.1610094064364813, 3.09407719269586, 1.0216323671234275), (253, 90, 3.2313670693024044, 3.195365308638267, 1.0112668684756672), (284, 430, 3.194918030235297, 3.1892925570822745, 1.0017638623777334), (404, 519, 3.687105499657023, 2.9272621178104443, 1.2595747668865855), (406, 531, 3.1706741115832973, 3.160786186448746, 1.0031283119297798), (457, 34, 3.0723647859199916, 3.1565278644537402, 1.027393582597825)], [(5, 389, 3.3645028736306775, 3.085719004967335, 1.090346485929069), (316, 420, 3.341528126298138, 3.4138015733001112, 1.021628860889476), (337, 172, 3.098682299497014, 3.2353000717461717, 1.0440889897848946), (400, 421, 3.0038338182328124, 3.2992999358229596, 1.0983630039041152), (440, 419, 3.1658211225777917, 3.0655056573162645, 1.032723953721015)], [(38, 351, 2.828566332947847, 3.48211300565591, 1.2310522702244555), (98, 263, 3.187656288969435, 3.119034785065078, 1.0220008780385965), (147, 454, 3.2784801857011536, 2.9326911743662105, 1.1179084297580948), (212, 479, 3.1962993527197385, 3.0313304458151533, 1.0544212878976391), (382, 76, 3.253930202955873, 3.1304691298917766, 1.0394385211740977), (422, 272, 3.2256423017000744, 3.0051852055567467, 1.0733589050470802)], [(22, 307, 3.321571228071647, 3.0829645094065925, 1.0773952207159794), (329, 354, 3.004512920858778, 3.1534483454189997, 1.0495705721637074), (395, 349, 3.157735145800857, 3.3105327913654996, 1.0483883665062386), (425, 29, 3.285750838754249, 3.1184179142998456, 1.053659557202734), (537, 392, 3.1854773249538106, 3.0956718264202094, 1.0290100190101388)], [(174, 181, 3.011192002735864, 3.4171870343650457, 1.1348286762386155), (175, 213, 3.3267863922226293, 2.987684561939487, 1.1134998769960542), (234, 292, 3.266707681747268, 3.0142275902274607, 1.0837627829890426), (340, 333, 3.883754868164459, 3.012836219994319, 1.2890693634092671), (369, 378, 3.022632870993381, 3.4691879750730155, 1.1477371295617766), (383, 474, 3.0284949428780124, 3.321428334556755, 1.0967257324855773)], [(4, 20, 3.1959164504516417, 3.1196711017730467, 1.0244401881452379), (81, 460, 3.007910635776715, 3.035301980251455, 1.0091064355931794), (161, 346, 3.34798364041703, 2.8275996074375875, 1.1840373833730378), (508, 77, 3.1317016777688336, 3.148715476938977, 1.0054327649695756), (509, 365, 3.4181004362623217, 2.8974169572899404, 1.1797060922358222)], [(158, 427, 3.1619824187900414, 3.295730849069048, 1.0422989164911887), (232, 293, 3.307199283896692, 3.099990454425687, 1.0668417637142027), (321, 373, 3.2176235365785955, 3.263169832281719, 1.0141552593662198), (432, 244, 3.1727956403744177, 3.3911979509571, 1.068835921167904), (492, 276, 2.8307377082945337, 3.2040255535362903, 1.1318694572612507)], [(13, 169, 2.971593517427857, 3.1659410096102434, 1.0654017755263543), (271, 493, 3.094290643406826, 3.1274406644300687, 1.0107132861271055), (311, 513, 3.2016711374251052, 3.103837140880805, 1.0315203382470437), (346, 211, 3.742689264607889, 3.252580874895862, 1.1506829218282602), (385, 59, 2.933036898791437, 3.454512501724665, 1.177793741070255), (399, 40, 3.6959380595128204, 3.9921446716183704, 1.0801438247438038)], [(11, 289, 3.150100709860709, 3.270994141043471, 1.038377640055866), (368, 528, 2.9535099091215455, 3.473946735876745, 1.176209609166333), (388, 8, 3.065655673803655, 3.1855174717881694, 1.0390982584928719), (431, 445, 3.3526642850514214, 2.9209266070213094, 1.147808464955026), (458, 379, 2.936113536996145, 3.194254999983586, 1.087919441715983), (521, 416, 3.3030504226293638, 3.018016802491453, 1.0944440136657316)], [(27, 357, 2.978827605861224, 3.169839083784378, 1.0641230387241325), (93, 393, 3.162124490867587, 3.1496268005280887, 1.003967990854473), (94, 1, 3.238029520920562, 2.998646400333901, 1.0798303929933204), (154, 217, 3.2348445500343246, 3.0627772618118527, 1.0561801507304784), (202, 237, 2.9607226342498714, 3.2231624662351823, 1.0886404653206574), (255, 265, 3.2422415693317173, 2.9908871878049306, 1.084040074313622)], [(316, 383, 3.172951957017077, 3.5865210815797957, 1.130342069519237), (333, 197, 3.288114951185095, 3.2705462574969792, 1.0053717918368663), (337, 465, 3.5866980649813254, 2.9440209979967995, 1.218299077153941), (363, 82, 2.8688106452151203, 3.1182711826832055, 1.0869560833107474), (424, 503, 3.0559218970508684, 3.2716228988294294, 1.0705845924880228), (510, 21, 3.346309910822877, 2.8353344757561993, 1.1802169865445584)], [(45, 274, 2.8609727031815924, 3.3683866811100054, 1.1773571545663946), (148, 299, 2.9693673480423777, 3.5729074508533065, 1.2032554521113146), (260, 170, 3.5295764403472374, 3.0517971772915495, 1.1565566894847559), (263, 491, 3.433613003334017, 2.9472928542921117, 1.1650057096747892), (306, 422, 3.3277963817927, 3.218521002430938, 1.033952047937306), (470, 9, 3.07554276082464, 3.234958914758185, 1.0518335026793126)], [(15, 270, 3.345725942855324, 3.6844097202822934, 1.1012287865807469), (47, 283, 3.161473107910798, 3.2335245572314344, 1.0227904672477983), (67, 442, 3.1466162555198802, 2.882986538664661, 1.091443270136574), (279, 91, 2.844312023537952, 3.42296143213926, 1.203440903744992), (339, 298, 3.473187056864409, 2.912216084512726, 1.19262683677044), (524, 280, 2.938311667693787, 3.560537386519781, 1.2117630085559865)], [(15, 61, 3.636696127537174, 2.9180032632757515, 1.2462961139579458), (279, 226, 3.3357559212122108, 3.2858014365811354, 1.0152031355501059), (305, 496, 3.1800179260025065, 3.3895001371271025, 1.0658745378168133), (408, 494, 3.398897714183604, 3.1122356998345784, 1.092108066996424), (423, 499, 2.8492184720480767, 3.261985628310783, 1.1448703075289277), (538, 242, 3.1944587443144403, 2.905957443141829, 1.0992792588389364)], [(6, 198, 3.0903527402931243, 3.045622388587091, 1.0146867687450856), (17, 214, 2.9503508957347497, 3.7720150513603423, 1.2784970956551143), (335, 513, 2.9739545879481826, 3.5890192385511375, 1.206817095693215), (407, 418, 3.396126268716366, 2.906931178035758, 1.1682857490321328), (520, 249, 3.385914804066517, 2.9421384818614675, 1.1508346139860404)], [(23, 50, 3.187366516461131, 3.1166044902166976, 1.022704846401448), (48, 322, 3.09407719269586, 3.516994263410728, 1.136686011491), (60, 348, 3.0214975824526973, 2.9115639298468405, 1.0377575953180735), (71, 135, 2.987055561154018, 3.265405628849882, 1.0931854336142064), (108, 406, 3.1610094064364813, 3.1706741115832973, 1.003057474339411), (228, 314, 3.5808298907629754, 2.8758463135908254, 1.2451395173102615)], [(9, 144, 2.8847318362692635, 3.1882368714976415, 1.1052108315277207), (70, 186, 3.323455581425549, 2.938788473021369, 1.1308930914679627), (102, 298, 3.073603511923633, 3.3506137930496003, 1.0901255741188942), (164, 523, 3.41114167505157, 3.084328565224637, 1.1059592397229348), (450, 260, 2.8160009098688645, 3.286542455408161, 1.1670956653068725), (451, 436, 3.4183867518437587, 3.0365815346213734, 1.1257352100937386)], [(140, 491, 3.189585314960798, 3.1828888208780963, 1.0021039044904039), (179, 299, 3.426514514221372, 3.0235727713247336, 1.13326675868962), (188, 432, 2.8411066681526678, 3.42585678924275, 1.2058177285790876), (253, 287, 3.0782664795547707, 3.54265057799015, 1.1508589660835815), (303, 374, 3.449869274325318, 2.9422319968455835, 1.1725347552551877), (404, 356, 3.149710630874017, 3.147884651538419, 1.0005800655162842)], [(31, 80, 3.606139440288883, 2.9855491418379967, 1.2078647072842457), (33, 194, 3.5425560604879798, 3.0216199233128562, 1.1724029329949537), (158, 525, 3.0247302443652977, 3.399659710101203, 1.1239546787467587), (232, 421, 2.9247999932243465, 3.5126677421013923, 1.2009941706232605), (331, 352, 3.1340390727766527, 3.3312093099184645, 1.062912501268571), (401, 389, 2.8324151008252474, 3.490305212602498, 1.2322717851580332)], [(43, 452, 3.6828543536474854, 2.8660303855669613, 1.2850018521066513), (74, 315, 2.7729076703834727, 3.1369662879401305, 1.1312912872812355), (245, 345, 3.3713278169596106, 2.985119264420868, 1.1293779304371176), (392, 537, 3.6171605270102027, 3.0956718264202094, 1.168457359122926), (443, 0, 2.9442968584178857, 3.035237755535734, 1.030887135873492)], [(15, 236, 3.038507275641341, 3.0592673672961133, 1.0068323323828114), (67, 141, 2.844405997945784, 3.326475326143656, 1.1694797889422326), (245, 415, 3.9111861785513002, 2.936872962817174, 1.3317519103037823), (484, 268, 3.3755193280258093, 3.0091321088587852, 1.1217584359584587), (538, 203, 2.878920604129509, 3.4371961929223307, 1.1939183692638256)], [(57, 191, 3.1119931541549604, 3.24503007244727, 1.0427497464493731), (152, 63, 3.3025898020996145, 3.0723944714507914, 1.0749237550021122), (239, 230, 2.897125954566208, 3.2380845800754017, 1.1176885751107242), (275, 23, 3.5480032285932936, 2.982592751339909, 1.1895701238458982), (314, 310, 3.4404919769036355, 3.0029577502969054, 1.1457010930518323), (355, 41, 2.9493032173810203, 3.4292607814026264, 1.162735917145816)], [(99, 495, 3.0207640003589113, 3.435410994004304, 1.1372656035347768), (269, 294, 3.2509586519955556, 3.0997494418321256, 1.0487811073123565), (281, 7, 2.8916348192678414, 3.6086358701979115, 1.247956984800596), (282, 312, 2.9768957020202467, 3.1897788551448696, 1.0715117943098078), (313, 14, 3.3936783275755507, 2.9936814390045714, 1.1336137116526273), (435, 341, 3.4032749814090018, 2.9109775703228946, 1.1691175555954214)], [(204, 189, 3.0359070108694097, 3.5539036708948775, 1.1706233617073556), (251, 289, 2.938454098771702, 3.183916336280324, 1.0835344808044567), (296, 32, 3.3456786776640586, 3.1811745701769287, 1.0517117510712342), (297, 224, 3.4697075027154467, 3.040668334962137, 1.1411002847039062), (459, 206, 3.259043638617252, 3.0415664165780254, 1.0715017172907584), (489, 416, 3.199312171197943, 2.98156925142103, 1.0730296368843808)], [(104, 402, 3.137544757187182, 3.185817224560477, 1.0153854274947687), (144, 438, 3.279692083334005, 3.239737747426439, 1.012332583382499), (318, 234, 3.1044817962266333, 3.134770912191493, 1.0097565770885417), (378, 370, 3.2089228877121156, 3.0897242574658446, 1.0385790511752775), (417, 70, 3.1562320514194386, 3.2222168092159738, 1.0209061807628688), (469, 473, 2.9518445796148325, 3.519831841249142, 1.1924177395912974)], [(83, 201, 3.0752035398452353, 3.137631146705446, 1.0203003170526241), (85, 502, 2.865310018588406, 3.7045030454841044, 1.2928803590018252), (139, 178, 2.8342681908862484, 3.7227141551470075, 1.313465735923512), (268, 484, 3.677456638811395, 3.0091321088587852, 1.2220987666128331), (326, 192, 3.6883569228258617, 3.058801698898756, 1.205817599798562), (398, 55, 3.5352218190275853, 3.0958134575487923, 1.1419363173860961)], [(10, 248, 3.5651044108211516, 2.8263874375632896, 1.2613643704469375), (63, 295, 3.0716362432430757, 3.299438532626326, 1.0741631727664254), (65, 355, 3.0607587106275296, 3.3203393469333964, 1.084809245303968), (163, 424, 3.345502973928838, 2.8946880051608352, 1.155738707578938), (197, 290, 2.8531216415113043, 3.4259981458165476, 1.200789372584125)], [(59, 229, 2.9340395938380768, 3.1877120932423804, 1.0864584445066978), (71, 123, 3.265405628849882, 2.987055561154018, 1.0931854336142064), (160, 425, 2.939341080511237, 3.2583854758202597, 1.1085428286714964), (354, 161, 3.331922527861232, 3.023606603281613, 1.101969589643373), (460, 466, 3.308296690530383, 3.0432168620182027, 1.0871051392428157), (519, 36, 3.316883501369983, 2.87149701265598, 1.1551060254463035)], [(73, 459, 3.0250031822478287, 3.4081667665448494, 1.1266655144515578), (74, 43, 2.8883012854260763, 3.605322539716054, 1.2482501593265067), (189, 395, 3.089841060776554, 3.1281491189372574, 1.0123980675404307), (243, 486, 3.117601552007677, 3.216672771382615, 1.0317780247803439), (247, 192, 2.8707040336162977, 3.550500126099675, 1.2368046599450453), (392, 342, 3.4464082105535496, 3.044790592107404, 1.1319031986919574)], [(80, 180, 3.042575567561341, 3.070815142469946, 1.009281470346927), (101, 515, 3.4257515458891272, 3.1357760424842356, 1.092473282363356), (353, 104, 3.301591968646269, 3.105072923156939, 1.0632896715641476), (438, 401, 3.112532883671786, 3.1789008328636488, 1.0213228106086931)], [(1, 10, 2.9766155957930542, 3.3798045197911324, 1.1354521304557827), (88, 245, 3.186117560709512, 3.3930475412230554, 1.0649473776690972), (196, 62, 3.190715841768162, 3.1103737829875664, 1.0258303549303407), (268, 93, 3.434482275089957, 3.091276377204276, 1.1110240095051203), (295, 447, 2.987833648265123, 3.1144256692476495, 1.042369166387838), (315, 258, 3.5591587845414234, 2.8495653270435537, 1.2490181399820999)], [(12, 463, 3.111727276895743, 3.362314470515118, 1.080529934445078), (75, 286, 2.948548325145212, 3.5543682459888606, 1.2054637923608775), (133, 268, 2.8342681908862484, 3.677456638811395, 1.2974977634919898), (178, 432, 3.7227141551470075, 3.1249274716418287, 1.1912961785289382), (293, 474, 3.1804758304375054, 2.974433989210993, 1.069270940950069)], [(29, 148, 3.2915077489049276, 3.175498448864526, 1.0365326268957502), (90, 361, 3.207385638772465, 2.9132184594383976, 1.1009766975700053), (125, 404, 3.189585314960798, 3.149710630874017, 1.0126597928380856), (256, 537, 2.9498032339900373, 3.219095949691826, 1.0912917555309378), (491, 35, 3.1828888208780963, 3.19387010806388, 1.0034501007744134)], [(13, 161, 3.018433502261906, 3.176075588889058, 1.0522264567064408), (47, 182, 2.9074090999535485, 3.4601819597784864, 1.190125586328312), (58, 107, 3.28253533143852, 3.067402408187351, 1.0701352136507905), (67, 128, 3.326475326143656, 2.844405997945784, 1.1694797889422326), (271, 302, 3.193672165296996, 2.9732652215892164, 1.0741295939922815), (471, 375, 3.1110632067527693, 3.232783051297128, 1.0391248381839873)], [(13, 283, 2.969232624117523, 3.6729268526725485, 1.2369953175238901), (58, 442, 3.276800903645684, 3.0629626023982492, 1.0698142057235707), (177, 522, 2.9885618757958388, 3.249469307542079, 1.0873020009588263), (191, 52, 3.1968346934062666, 3.1757888868798094, 1.006626953892749), (311, 285, 3.4339388333119576, 3.0203257462374817, 1.1369432047486028), (535, 239, 3.089094945862324, 3.425529336320636, 1.1089103431116443)], [(297, 231, 3.647527033142582, 2.891553304084559, 1.2614420865042149), (363, 483, 3.1312985828551896, 3.0754848317742627, 1.0181479519925734), (364, 330, 2.9898111245947003, 3.0373359699183102, 1.0158956012079365), (367, 247, 3.0855678206478374, 3.250385989238422, 1.0534158307873394), (424, 163, 3.3799451909403677, 2.8946880051608352, 1.167637128738705), (459, 2, 3.0126371025145566, 3.116380488876878, 1.0344360713992833)], [(9, 124, 3.1882368714976415, 2.8847318362692635, 1.1052108315277207), (132, 469, 3.279692083334005, 2.9518445796148325, 1.111065299976583), (257, 64, 3.1954917788659447, 3.394637143330579, 1.0623207250232105), (353, 45, 2.8149725066754336, 3.4342162407358168, 1.2199821606043777), (389, 5, 3.480052570340591, 3.085719004967335, 1.1277930896295043), (438, 31, 3.239737747426439, 3.267516251131788, 1.0085743062775423)], [(62, 196, 3.213869533024464, 3.1103737829875664, 1.0332743770549302), (218, 344, 2.952159980558611, 3.241789867054261, 1.0981077883322725), (311, 455, 3.322432720751312, 3.011513096358524, 1.1032436567414392), (346, 329, 3.2395582855693483, 3.0980453100091365, 1.0456781490906581), (349, 17, 3.1371255653115204, 3.4579590317432674, 1.1022698836091656), (535, 509, 3.067297402623201, 3.0621738495350717, 1.0016731751167254)], [(197, 167, 3.0910156202292742, 3.183681671402428, 1.0299791597838255), (340, 316, 3.369415553740101, 2.8655628893584533, 1.17583025877839), (347, 526, 3.4965771761984925, 3.014494102273366, 1.1599217173991372), (384, 498, 3.0810256359088237, 3.003036263749825, 1.0259701732877562), (403, 96, 3.6995560479066225, 2.9545018363495466, 1.2521759175745284)], [(110, 38, 3.2784801857011536, 2.828566332947847, 1.1590607395388248), (153, 259, 3.1115023796781927, 3.040832647247099, 1.0232402570707309), (159, 53, 2.9231171875187254, 3.4952607218571887, 1.1957306182528127), (454, 51, 2.9326911743662105, 3.5150334477700254, 1.19856924537227), (497, 467, 3.1796514647550387, 3.4343306579968096, 1.0800965753840543), (539, 356, 3.584041604758573, 3.016537387801347, 1.1881310071780218)], [(29, 140, 3.175498448864526, 3.2915077489049276, 1.0365326268957502), (40, 532, 2.964149625378467, 3.2513107618704047, 1.0968780840323715), (101, 240, 3.7199427084169594, 3.0224623827131674, 1.2307655935415436), (119, 306, 2.9693673480423777, 3.3277963817927, 1.1207088890455486), (210, 246, 3.3685529925674706, 2.7237751977258364, 1.2367221037110474), (299, 105, 3.5729074508533065, 3.0550982228439048, 1.1694902062845587)], [(1, 63, 3.2551179306418545, 3.1304544745456697, 1.0398227979706613), (88, 74, 3.340135863983737, 2.991722323200221, 1.116459184089926), (242, 27, 3.1735042237025324, 3.2394840724468446, 1.0207908495131395), (248, 39, 3.02177306894235, 3.3389567585642377, 1.10496608527023), (345, 56, 2.9831983446226693, 3.1857967632202504, 1.067913157354345), (486, 230, 3.035752923401623, 3.3559258782101136, 1.1054673956962644)], [(92, 52, 3.648145528192271, 2.8996357511927076, 1.2581392427278766), (264, 213, 3.0517646675947856, 3.234052364503101, 1.0597318983485), (433, 378, 2.9171697386390814, 3.376504623630983, 1.157459087453098), (456, 522, 3.163713408874697, 2.997921556710785, 1.0553022649284438), (473, 267, 3.212711828237726, 3.2173104890362794, 1.001431395358318)], [(18, 449, 3.294217853981194, 3.085273073769242, 1.0677232696153818), (59, 207, 3.225507331731531, 3.2443680594912236, 1.0058473678153346), (182, 524, 3.3427462521398987, 3.164213033402539, 1.0564226292138679), (216, 464, 3.378997929624203, 2.9098095958497163, 1.161243654720121), (283, 13, 3.214955296882028, 3.6729268526725485, 1.142450365090512), (460, 479, 3.0111434331982165, 3.005724338895261, 1.0018029245838782)], [(26, 486, 3.4189587735602154, 3.003139989722938, 1.1384613388853844), (34, 192, 3.1171613564427094, 3.3588054855493454, 1.077520571274629), (63, 201, 3.0723944714507914, 3.49976097804802, 1.1390988398685098), (65, 323, 3.520885617677185, 2.9263605379473026, 1.203161938530962), (84, 284, 3.1334339411342667, 3.1259377624169575, 1.0023980575708946), (129, 57, 3.3025898020996145, 3.1119931541549604, 1.061245844223719)], [(147, 497, 3.1115023796781927, 3.1796514647550387, 1.0219023085188494), (259, 292, 3.040832647247099, 3.5547862934695837, 1.1690174060344205), (379, 472, 3.0715219878472495, 3.348437167898343, 1.0901556886607788), (382, 11, 2.987514325396019, 3.380663794831617, 1.1315975177402648), (465, 475, 3.1547900201437886, 2.9802851928897, 1.0585530632002664), (516, 333, 3.360420558372048, 2.932606238867815, 1.1458819509534284)], [(68, 462, 3.116026228599773, 3.0832078877456577, 1.0106442192836083), (107, 490, 3.136010449901598, 3.158648319706941, 1.0072186844294646), (117, 94, 3.2348445500343246, 3.238029520920562, 1.0009845823615244), (217, 365, 3.0627772618118527, 3.634003068001854, 1.186505827019259), (413, 209, 3.098486227088526, 3.314122523483043, 1.0695940793634375), (480, 471, 3.196646151861502, 3.0108367001341865, 1.0617135601273342)], [(52, 55, 3.4578806009925485, 3.0952191523283625, 1.1171682620247991), (57, 535, 3.0088106883822476, 3.345474459818699, 1.1118926400841342), (92, 413, 3.7900227139742784, 3.1743458236989808, 1.1939539434168724), (275, 84, 3.0635402895715016, 3.274266926053315, 1.0687853321854917), (344, 334, 3.3182828120155867, 2.8722035622700584, 1.1553090649998943), (464, 216, 3.4026019884029313, 2.9098095958497163, 1.1693555458941673)], [(312, 181, 3.1734830337780044, 2.8976582478437454, 1.0951888602251527), (350, 254, 2.909995393435746, 3.4178013436710275, 1.1745040392093988), (371, 73, 3.5423003903093675, 2.952897820531345, 1.1996014104111348), (502, 474, 3.4611499249062287, 2.8864383357192165, 1.199107523647759), (507, 435, 2.7967048951454974, 3.1469051054161548, 1.1252188641277572), (512, 268, 3.0546269397946677, 3.2478332285572344, 1.0632503715087231)], [(21, 510, 3.508200462951603, 2.8353344757561993, 1.237314501322087), (49, 462, 3.0396119786127955, 3.054939485919205, 1.005042586821692), (100, 509, 3.867441398101282, 3.742351340702459, 1.0334255247598807), (173, 520, 3.2344023030080105, 2.9245051705190597, 1.1059656640764113), (418, 394, 3.1984504072422975, 2.8287032058806267, 1.1307126179208191), (504, 467, 3.0565597162645055, 3.4465568561604245, 1.127593495988537), (511, 534, 3.0762338792636443, 3.1837615376182113, 1.0349543183564138)], [(14, 518, 3.4717993803038536, 2.9764558343472927, 1.1664205933246057), (126, 331, 3.0247302443652977, 3.1340390727766527, 1.0361383725424718), (208, 236, 2.970174945965188, 3.5569558865360844, 1.1975577032484246), (427, 410, 3.295730849069048, 3.0366869423404292, 1.085304778413862), (525, 435, 3.399659710101203, 3.0832721091795516, 1.1026142324511996)], [(53, 348, 3.4952607218571887, 3.2993229664989334, 1.0593872613708304), (98, 530, 3.21574604530358, 3.1779215889423913, 1.011902262312827), (147, 539, 2.9231171875187254, 3.584041604758573, 1.2261026071968295), (172, 510, 3.3423422558139357, 3.5032056228928434, 1.0481289331752572), (402, 104, 2.984894306050817, 3.185817224560477, 1.0673132439236994), (465, 440, 3.120641116477984, 3.3739098552632516, 1.0811592007321595)], [(135, 460, 2.939341080511237, 3.308296690530383, 1.1255232380023668), (179, 60, 3.0512097312048057, 3.494459921579365, 1.1452703122441656), (207, 410, 3.3256977198315862, 3.214091474474762, 1.0347240413793957), (303, 320, 2.9943540428130055, 3.170698728151715, 1.058892396429196), (329, 346, 3.1488729823189874, 3.0980453100091365, 1.0164063682818445), (425, 426, 3.2583854758202597, 2.885298605352015, 1.129306155618069)], [(13, 141, 3.176075588889058, 3.018433502261906, 1.0522264567064408), (113, 4, 3.34798364041703, 3.1959164504516417, 1.0475817163317578), (230, 486, 3.0810536976116962, 3.3559258782101136, 1.089213693617702), (346, 395, 2.8275996074375875, 3.3367863311644963, 1.180077378136412), (354, 135, 3.023606603281613, 3.331922527861232, 1.101969589643373), (535, 57, 3.0751574581586096, 3.345474459818699, 1.087903466842948)], [(78, 171, 3.066698229696414, 3.397173782885968, 1.1077626582196414), (199, 10, 3.1436349941357413, 2.9278889446082474, 1.0736865549237424), (239, 62, 3.174032509251582, 3.1480452859286903, 1.0082550347795347), (355, 44, 3.340544281800449, 3.05019794342888, 1.0951893430382345), (508, 387, 3.196567400501707, 3.11968957298559, 1.0246427811862524), (509, 177, 3.0568194301595466, 3.342922791689961, 1.0935951135051118)], [(25, 39, 2.957286324089136, 3.1407757784455432, 1.0620465637235594), (134, 63, 3.345502973928838, 3.0716362432430757, 1.0891598838528518), (323, 41, 2.9265452057431185, 3.445453115205379, 1.1773107445748467), (333, 516, 3.401176617330192, 2.932606238867815, 1.159779506792319), (347, 175, 3.992601574032313, 2.9686098004008095, 1.3449398346300845), (424, 143, 2.8946880051608352, 3.3799451909403677, 1.167637128738705)], [(124, 9, 3.41114167505157, 2.8847318362692635, 1.1824813773550253), (266, 170, 3.067592294969673, 3.732707886210436, 1.216820074926984), (269, 405, 3.376493808194789, 2.888966924829519, 1.1687547473026327), (281, 529, 3.194291038364821, 3.00833323643055, 1.0618142297809114), (390, 381, 2.9341200768159585, 3.3843466314787767, 1.1534451702301816), (523, 485, 3.084328565224637, 3.146700520573147, 1.020222214990888)], [(230, 239, 2.8832139217596824, 3.2380845800754017, 1.1230816262496175), (354, 385, 3.138441254092852, 3.3218137458710184, 1.0584278872637904), (374, 537, 2.9362923747194083, 3.2565045223019746, 1.1090532231529449), (387, 431, 3.306649538048984, 2.853350087854791, 1.1588656968956073), (519, 404, 3.3518727086120816, 2.9272621178104443, 1.1450538331426365)], [(21, 86, 3.206937302750879, 3.383813441697338, 1.0551542241860283), (41, 355, 2.907414386585913, 3.4292607814026264, 1.1794881380598456), (171, 351, 3.066333256722596, 2.9388719636411316, 1.0433708220903728), (231, 17, 3.4025915291230717, 3.0831525375371696, 1.103607910311525), (418, 328, 3.3170847750298806, 3.2617784944761694, 1.0169558664536456), (446, 424, 3.110684708541374, 3.1657290741692012, 1.0176952570849387)], [(44, 534, 3.2660908617788658, 3.1740998023237177, 1.0289817791450044), (51, 510, 3.298687795357158, 3.0430856446771584, 1.0839943992792607), (65, 199, 3.4048959948538573, 2.9768770427598126, 1.1437811995409912), (84, 457, 3.0053580600169183, 3.286717100561865, 1.0936191411892409), (197, 146, 3.183681671402428, 3.0910156202292742, 1.0299791597838255), (430, 539, 2.9203111406261097, 3.4811857299456883, 1.1920598738664976)], [(174, 420, 3.3222692373399756, 3.028603037637261, 1.0969642426073165), (219, 288, 2.860138991215845, 3.6086836458650424, 1.2617161812583768), (369, 318, 3.3536889089337403, 2.985247183390596, 1.123420843537879), (370, 173, 2.9894057925463793, 3.0946842194410475, 1.0352171749841268), (462, 68, 3.461591956464979, 3.0832078877456577, 1.1227241504613508), (485, 523, 3.2001342499988508, 3.146700520573147, 1.0169808753887934)], [(13, 115, 3.1659410096102434, 2.971593517427857, 1.0654017755263543), (113, 508, 3.2704644466536408, 3.1317016777688336, 1.044309063621816), (177, 217, 3.1972546494587193, 2.9363640599676897, 1.088848175554192), (183, 18, 3.027145077130417, 3.9697454362400424, 1.3113826179758663), (471, 357, 3.079449136784337, 3.2090211368760597, 1.0420763566262459), (501, 69, 3.0717759742918753, 3.2904890729532625, 1.0712008624625715)], [(22, 448, 3.130159813213813, 3.0202977558449122, 1.0363745783528444), (260, 119, 3.0517971772915495, 3.5295764403472374, 1.1565566894847559), (266, 164, 3.732707886210436, 3.067592294969673, 1.216820074926984), (461, 20, 3.0112431873090935, 2.9941942547357097, 1.0056939968228242), (532, 399, 2.9135546556875407, 3.0041987741071012, 1.0311111783136158)], [(78, 162, 3.397173782885968, 3.066698229696414, 1.1077626582196414), (166, 21, 3.066333256722596, 3.206937302750879, 1.0458541307341673), (296, 300, 2.819462194035446, 3.530570332397716, 1.2522141066004058), (351, 38, 2.9388719636411316, 3.48211300565591, 1.184846787725222), (394, 69, 3.524441809197798, 3.1378152343196124, 1.1232152137734202), (489, 503, 3.195103419418893, 2.929382320517427, 1.090708917385195)], [(86, 504, 2.8904003288323894, 3.8935704552319867, 1.3470696139883294), (159, 98, 3.3423422558139357, 3.21574604530358, 1.0393676020204525), (200, 187, 2.9401232616568467, 3.315434297451441, 1.1276514630148857), (337, 109, 3.2353000717461717, 3.098682299497014, 1.0440889897848946), (454, 430, 3.2812755239513693, 2.9988308382176925, 1.094184934386477), (510, 446, 3.5032056228928434, 3.019743956840392, 1.1601002180854783)], [(157, 511, 3.2344023030080105, 3.0762338792636443, 1.0514162544046315), (273, 224, 3.1530476263692866, 3.1480692910314505, 1.001581393189794), (370, 168, 3.0946842194410475, 2.9894057925463793, 1.0352171749841268), (391, 104, 3.356642030605456, 3.02615423982134, 1.1092104911359797), (407, 32, 3.087396935437208, 3.2410708053103465, 1.049774574856011), (520, 215, 2.9245051705190597, 3.496684518530334, 1.19564996970401)], [(112, 234, 3.011192002735864, 3.266707681747268, 1.0848553259902562), (168, 370, 3.3222692373399756, 2.9894057925463793, 1.111347695125078), (181, 79, 3.4171870343650457, 3.0450987338632114, 1.1221925241254094), (237, 202, 3.1034638143259983, 3.2231624662351823, 1.038569372504567), (330, 502, 3.410976762988592, 2.89350683377068, 1.1788383297313902), (420, 424, 3.028603037637261, 3.225967083572584, 1.0651666935159965)], [(112, 383, 3.3267863922226293, 3.0284949428780124, 1.098494947150596), (213, 255, 2.987684561939487, 3.230982586641851, 1.0814336385446341), (237, 476, 3.2294285979661566, 2.902045101487362, 1.1128113054862598), (347, 163, 2.9686098004008095, 3.992601574032313, 1.3449398346300845), (403, 505, 3.1072716813554093, 3.3341533033763313, 1.0730163452981218), (444, 433, 3.3146274966796736, 3.2145020843237644, 1.0311480315549313)], [(90, 348, 2.905531477377357, 3.460921519198629, 1.1911492083791113), (195, 406, 3.5593456951059057, 2.8245756504933506, 1.260134666417668), (220, 342, 3.1616155406872832, 2.9246402355738406, 1.0810271643776883), (238, 440, 3.262602681808449, 3.3443852435961494, 1.0250666629570624), (421, 184, 3.125325264365802, 3.0823420509621258, 1.013944984915045)], [(142, 311, 2.9885618757958388, 3.4339388333119576, 1.1490271829816197), (169, 183, 3.1972546494587193, 3.027145077130417, 1.0561947207662599), (217, 92, 2.9363640599676897, 3.126436143745149, 1.0647304216696996), (480, 27, 3.0629140493906934, 3.3331643809653517, 1.0882330771339859), (509, 162, 3.342922791689961, 3.0568194301595466, 1.0935951135051118), (522, 328, 3.249469307542079, 3.3462948556511924, 1.0297973419488466)], [(33, 233, 3.175564369794431, 3.1705519308351504, 1.0015809357703724), (139, 133, 3.7227141551470075, 2.8342681908862484, 1.313465735923512), (195, 253, 3.254871878877083, 3.3618227097050295, 1.0328586914655589), (232, 536, 3.3193961011549935, 3.230791796494345, 1.027424950365663), (531, 336, 3.185339147081731, 3.1407232747039346, 1.0142056044023815)], [(29, 263, 2.9089835298407305, 3.2527051357101393, 1.118158663444969), (60, 16, 3.494459921579365, 2.892993229604277, 1.20790463172199), (71, 46, 3.5383050937522866, 2.96001240251731, 1.1953683338431873), (125, 188, 3.426514514221372, 2.8411066681526678, 1.2060492316711733), (160, 329, 3.0512097312048057, 3.1488729823189874, 1.0320080426184333), (299, 80, 3.0235727713247336, 3.299104346262454, 1.0911278132779985)], [(80, 137, 3.070815142469946, 3.042575567561341, 1.009281470346927), (105, 71, 3.003252404306952, 3.260562029233435, 1.0856769895721972), (229, 222, 3.481597715289835, 2.91840525025631, 1.192979527083863), (233, 33, 3.1340477576381454, 3.1705519308351504, 1.011647612295645), (322, 52, 3.0837141854361168, 3.017535830608052, 1.0219312573381205), (362, 267, 3.015138496149572, 3.422718168388607, 1.13517776140616)], [(79, 281, 3.0450987338632114, 3.478460778163109, 1.142314611832013), (174, 112, 3.4171870343650457, 3.011192002735864, 1.1348286762386155), (312, 156, 2.8976582478437454, 3.1734830337780044, 1.0951888602251527), (386, 503, 3.0894464651200884, 3.167594299848571, 1.0252950927005122), (476, 24, 3.5647963538210146, 2.8066976209087193, 1.2701034579802177), (485, 360, 3.301182238128192, 3.073577049139953, 1.0740522151712213)], [(28, 186, 3.1693985513724354, 3.3471396254720545, 1.0560803796741347), (47, 141, 3.4601819597784864, 2.9074090999535485, 1.190125586328312), (151, 460, 3.3427462521398987, 3.0111434331982165, 1.11012521532044), (229, 410, 3.2247423617637536, 2.9864000926493626, 1.079809222381502), (362, 359, 2.96734800702388, 3.2332423573305014, 1.0896067295366887), (524, 236, 3.164213033402539, 3.15151015189195, 1.0040307284122068)], [(18, 343, 3.9697454362400424, 2.88205664816274, 1.3774002113284918), (169, 177, 3.027145077130417, 3.1972546494587193, 1.0561947207662599), (226, 279, 2.9189783532410742, 3.2858014365811354, 1.1256683123164517), (366, 477, 3.1674241905261806, 2.9792715731147723, 1.0631538994663379)], [(42, 316, 2.8629129738292907, 3.162126716443487, 1.1045137401483716), (82, 205, 2.880462730763372, 3.4863663348362106, 1.210349398935727), (258, 371, 3.03577605706026, 2.970155669117857, 1.0220932487225132), (326, 295, 3.528025370917465, 3.007889682584272, 1.1729237915023238), (421, 176, 3.0823420509621258, 3.125325264365802, 1.013944984915045)], [(11, 382, 2.9835662994666445, 3.380663794831617, 1.1330949124328022), (251, 448, 2.906445751749418, 3.3976232712963927, 1.168995935758075), (351, 20, 3.229131495583902, 3.0476135717573363, 1.059560675772256), (441, 338, 3.0743803175583073, 3.3310830782579157, 1.08349739921035), (489, 446, 3.284938671259404, 2.773626144831022, 1.1843480338477748), (521, 77, 3.107583792498345, 3.2656070472777854, 1.050850842754717)], [(28, 182, 3.3471396254720545, 3.1693985513724354, 1.0560803796741347), (30, 469, 2.985424835075639, 3.10832802994952, 1.0411677404937134), (70, 124, 2.938788473021369, 3.323455581425549, 1.1308930914679627), (309, 241, 3.5295348215655666, 3.0660693898905955, 1.1511594725165397), (381, 375, 3.4709898351454567, 2.8698072605774385, 1.2094853486596355), (449, 324, 3.0011543505702614, 3.298521965913676, 1.0990844123984862)], [(16, 228, 3.1927847348862914, 3.245796314530629, 1.016603555844245), (200, 172, 3.315434297451441, 2.9401232616568467, 1.1276514630148857), (205, 214, 3.071251954799175, 3.34524106820175, 1.0892108877535873), (246, 249, 2.898755790494772, 3.4561891642112825, 1.192300909081191), (348, 29, 3.2644519455338115, 3.0442734927040664, 1.072325450836604), (361, 421, 3.2483343648453022, 3.427308552710056, 1.0550972183780338)], [(125, 179, 2.8411066681526678, 3.426514514221372, 1.2060492316711733), (193, 472, 3.2073873849490195, 3.060664614438363, 1.0479382059107383), (195, 308, 3.3221297840003476, 2.9107983575516987, 1.1413122366863722), (220, 252, 2.914746170069079, 3.618258504629908, 1.2413631560048182), (320, 415, 2.9341048773387417, 3.4035965528464205, 1.1600118929400751), (432, 208, 3.42585678924275, 2.9855109693324553, 1.1474942897325056)], [(136, 74, 3.089841060776554, 2.8883012854260763, 1.069777961311521), (204, 131, 3.5539036708948775, 3.0359070108694097, 1.1706233617073556), (211, 230, 2.948189460036143, 3.136460660056166, 1.0638599393194068), (414, 310, 3.5818481840801604, 2.9114248400587095, 1.2302732788416861), (537, 256, 2.999315348624523, 3.219095949691826, 1.0732769234045674)], [(64, 370, 2.8941873122502497, 3.314433775477454, 1.1452036160370214), (215, 334, 3.5976678374446123, 3.4317090643370167, 1.0483603854511652), (254, 227, 3.210753938262146, 2.962856358147407, 1.0836684436061363), (288, 502, 3.0720014677038416, 3.3624883812011683, 1.094559496976559), (291, 55, 3.403744249819916, 3.0145882945515963, 1.129090913001838), (396, 438, 2.8824299455578726, 3.660284789151241, 1.26986079741231)], [(52, 322, 3.1757888868798094, 3.017535830608052, 1.0524444663312806), (57, 129, 3.24503007244727, 3.1119931541549604, 1.0427497464493731), (107, 93, 3.123001280810733, 3.3306042705064445, 1.0664754737602342), (142, 13, 3.1968346934062666, 2.969232624117523, 1.0766534987659946), (217, 456, 3.454395092767328, 2.8585769726262877, 1.2084317217435774), (513, 335, 2.9215310517796964, 3.5890192385511375, 1.2284720493950698)], [(34, 152, 3.3588054855493454, 3.1171613564427094, 1.077520571274629), (82, 363, 2.988749157646431, 3.1182711826832055, 1.0433365325107344), (247, 136, 3.550500126099675, 2.8707040336162977, 1.2368046599450453), (287, 253, 2.9687921568310376, 3.54265057799015, 1.193296933851935), (310, 478, 3.2169178556324765, 3.186924079010081, 1.0094115127561218), (326, 133, 3.058801698898756, 3.6883569228258617, 1.205817599798562)], [(36, 8, 3.150530354474014, 2.9051824962686053, 1.084451788664062), (188, 220, 3.2073873849490195, 2.914746170069079, 1.100400239954001), (244, 236, 3.170659141190159, 2.9833897587558273, 1.0627706728176305), (404, 71, 3.5385316786672907, 2.9780131805371033, 1.1882189446955687), (406, 447, 2.7199861888244623, 3.349977071083289, 1.2316154710076301), (472, 379, 3.060664614438363, 3.348437167898343, 1.0940228968905783)], [(33, 126, 3.0216199233128562, 3.5425560604879798, 1.1724029329949537), (236, 15, 3.422904152826007, 3.0592673672961133, 1.118864009539411), (265, 255, 3.543154040211609, 2.9908871878049306, 1.1846498439187196), (280, 362, 3.3808066626868283, 3.1229014127964048, 1.0825851398425934), (496, 505, 3.3400228459626757, 3.13278416266613, 1.0661515995152941), (531, 406, 2.88767197657496, 3.160786186448746, 1.0945793746967494)], [(176, 238, 3.5593456951059057, 3.262602681808449, 1.0909528502971049), (178, 33, 3.254871878877083, 3.175564369794431, 1.0249743037291308), (188, 320, 3.3221297840003476, 2.9341048773387417, 1.1322464338812415), (253, 60, 3.3618227097050295, 2.878770210330181, 1.1677982138489076), (308, 326, 2.9107983575516987, 3.525171188960263, 1.211066778231014), (406, 429, 2.8245756504933506, 3.272915854643634, 1.158728339979839)], [(26, 199, 3.3372922511155165, 3.1670244662786793, 1.0537626995464628), (34, 457, 2.8866751297091824, 3.1565278644537402, 1.0934821975522213), (62, 145, 3.1103737829875664, 3.213869533024464, 1.0332743770549302), (243, 415, 3.415431252338681, 2.8907404467069613, 1.1815074079824186), (247, 367, 2.852041702288446, 3.250385989238422, 1.1396698676005856)], [(134, 10, 2.8531216415113043, 3.5651044108211516, 1.2495451855087079), (146, 403, 3.0910156202292742, 3.6995560479066225, 1.1968739412686016), (167, 44, 3.183681671402428, 3.2660908617788658, 1.0258848713163387), (278, 447, 3.0723075884555997, 2.907141659495836, 1.0568138564628486), (290, 379, 3.4259981458165476, 3.0945099510617884, 1.107121385937382), (333, 118, 3.2705462574969792, 3.288114951185095, 1.0053717918368663)], [(6, 122, 3.045622388587091, 3.0903527402931243, 1.0146867687450856), (89, 269, 2.9659968409869086, 3.486821410172158, 1.1755984908641877), (219, 3, 3.3708741864174394, 3.0698019639035174, 1.098075454395463), (312, 206, 3.1940000634227967, 3.0685606306227906, 1.040878916175936), (350, 393, 3.0174072964615104, 3.288349748425666, 1.0897931320978402), (485, 301, 3.2722066794872653, 3.0018180010786932, 1.0900749740028906)], [(10, 506, 2.9278889446082474, 3.2516528233106405, 1.1105792893199071), (26, 196, 3.1670244662786793, 3.3372922511155165, 1.0537626995464628), (65, 167, 2.9768770427598126, 3.4048959948538573, 1.1437811995409912), (77, 388, 3.399384245430149, 2.985766560918502, 1.1385298133905042), (162, 509, 3.1436349941357413, 3.0568194301595466, 1.0284006189962172), (416, 489, 3.6331585286290076, 2.98156925142103, 1.2185390384266361)], [(53, 263, 3.037308634563409, 3.4841151475264738, 1.1471060622152711), (172, 454, 2.9401232616568467, 3.2812755239513693, 1.1160333196718675), (187, 16, 3.315434297451441, 3.1927847348862914, 1.0384146044125702), (224, 297, 3.032407579016425, 3.040668334962137, 1.0027241575317494), (274, 391, 3.2474391015351878, 3.2101355137372476, 1.0116205648136365), (483, 400, 3.225277419614823, 2.97744842648178, 1.0832353604948528)], [(63, 152, 3.49976097804802, 3.0723944714507914, 1.1390988398685098), (83, 133, 3.137631146705446, 3.0752035398452353, 1.0203003170526241), (134, 65, 3.31327730607646, 3.0607587106275296, 1.0825019608935975), (213, 264, 3.132714764040242, 3.234052364503101, 1.0323481734200928), (463, 336, 3.0455962777832166, 3.715348803558551, 1.2199085054906962), (506, 500, 3.081895517744509, 3.07818936967174, 1.0012040026222182)], [(24, 358, 3.163734676694957, 2.87742930552752, 1.0995004014929042), (68, 456, 3.1449007720146405, 2.999811588522652, 1.0483660987400352), (117, 93, 2.9607226342498714, 3.162124490867587, 1.0680245607230758), (213, 481, 3.0248307292823786, 3.276017620252854, 1.0830416355331287), (237, 174, 3.2231624662351823, 3.1034638143259983, 1.038569372504567), (506, 10, 3.3723670953921556, 3.2516528233106405, 1.0371239731425606)], [(358, 375, 3.289238619241666, 3.146759004845712, 1.0452782097950777), (423, 241, 3.0729890156694966, 3.1523943164629853, 1.0258397606983274), (482, 388, 3.346759033695243, 3.1855367826808774, 1.050610701433711), (500, 506, 3.0799489299440057, 3.07818936967174, 1.0005716218402942), (528, 338, 2.8657481118419277, 3.5687633935730907, 1.245316494784082), (538, 128, 3.4371961929223307, 2.878920604129509, 1.1939183692638256)], [(73, 371, 3.241952225849331, 2.952897820531345, 1.0978883872337897), (131, 489, 3.0359070108694097, 3.199312171197943, 1.053824165148503), (189, 211, 3.5539036708948775, 2.948189460036143, 1.2054529463148236), (214, 205, 2.878342707480302, 3.34524106820175, 1.162210830387939), (300, 304, 3.008710102342345, 3.5066457234995747, 1.1654980387673695), (448, 380, 3.375592196038874, 2.865991172318017, 1.177809697616302)], [(82, 184, 3.4863663348362106, 2.880462730763372, 1.210349398935727), (187, 348, 3.071251954799175, 3.2644519455338115, 1.0629059398506007), (214, 204, 3.34524106820175, 2.878342707480302, 1.162210830387939), (224, 510, 2.9343711035554043, 3.284753526381704, 1.1194063090390176), (310, 314, 3.4384744971873773, 3.0029577502969054, 1.1450292621823974)], [(0, 99, 2.8675345151801683, 3.473728477611225, 1.2113990116673345), (312, 198, 3.0685606306227906, 3.1940000634227967, 1.040878916175936), (364, 42, 3.1418041578635356, 3.2040972148680726, 1.0198271610433214), (371, 74, 2.893935180896006, 3.5802724869024773, 1.237164021688271), (386, 517, 3.2875798487998757, 3.05617795148539, 1.075716107173019), (459, 131, 3.0415664165780254, 3.259043638617252, 1.0715017172907584)], [(40, 260, 3.024758973861427, 2.9757449671413876, 1.0164711718447847), (59, 151, 3.2443680594912236, 3.225507331731531, 1.0058473678153346), (160, 303, 3.3256977198315862, 2.9943540428130055, 1.1106561456264217), (299, 409, 3.167605438785687, 3.040957046472888, 1.0416475439729391), (399, 468, 3.0295404923234535, 3.205259789319713, 1.05800196347978), (410, 427, 3.214091474474762, 3.0366869423404292, 1.058420421829062)], [(33, 75, 3.520439616344382, 3.0305463916862316, 1.161651782002772), (61, 341, 2.999047205943419, 3.379598187480926, 1.1268906273910404), (158, 14, 2.970174945965188, 3.4717993803038536, 1.168887167747507), (235, 229, 3.409699540178146, 3.079502894572144, 1.1072240088450633), (236, 305, 3.5569558865360844, 2.97355384989083, 1.1961968964061886), (432, 188, 2.9855109693324553, 3.42585678924275, 1.1474942897325056)], [(265, 12, 2.872580339488628, 3.538367548005047, 1.231773224708118), (280, 524, 3.0198675452235997, 3.560537386519781, 1.1790376012190789), (286, 262, 2.8875955544377367, 3.0620821207452087, 1.0604262484194908), (397, 96, 3.45691442261032, 2.974564914733546, 1.162158003507542), (413, 154, 3.314122523483043, 3.098486227088526, 1.0695940793634375)], [(45, 402, 3.0321108513094845, 3.108107611415448, 1.0250639781435242), (137, 438, 3.7405267779509352, 3.112532883671786, 1.2017629749627958), (148, 40, 3.3685529925674706, 2.964149625378467, 1.136431495807965), (246, 223, 2.7237751977258364, 3.244814784700271, 1.1912931681768255), (348, 90, 3.118018677582383, 3.460921519198629, 1.1099745951111886), (409, 48, 3.3056241653773113, 3.0359950620756333, 1.0888107845331407)], [(62, 314, 2.895029503481432, 3.1320717708240733, 1.0818790506478724), (189, 204, 2.948189460036143, 3.5539036708948775, 1.2054529463148236), (230, 508, 3.136460660056166, 2.976468446276127, 1.053752363469603), (243, 88, 3.1330579498590096, 3.1138555369376864, 1.006166764223175), (346, 115, 3.252580874895862, 3.742689264607889, 1.1506829218282602), (374, 303, 3.316338328284098, 2.9422319968455835, 1.127150521046471)], [(110, 98, 3.1962993527197385, 3.187656288969435, 1.002711416466139), (431, 533, 3.3126500761638735, 2.964039885966948, 1.117613191322896), (458, 411, 3.08547297622838, 3.4367751173775245, 1.113856819961058), (466, 18, 3.3897054178830555, 2.9940080841216705, 1.1321630812755363), (479, 519, 3.0313304458151533, 3.1714219869932405, 1.0462145396822335), (539, 430, 2.833145830699028, 3.4811857299456883, 1.2287351015343846)], [(175, 112, 2.987684561939487, 3.3267863922226293, 1.1134998769960542), (202, 24, 3.0248307292823786, 3.163734676694957, 1.0459212299279743), (255, 398, 3.230982586641851, 2.983470831116972, 1.082961010693781), (264, 150, 3.234052364503101, 3.0517646675947856, 1.0597318983485), (481, 496, 3.276017620252854, 2.984871923918773, 1.0975404318024613)], [(17, 122, 3.7720150513603423, 2.9503508957347497, 1.2784970956551143), (204, 448, 2.878342707480302, 3.375592196038874, 1.1727554843508763), (205, 187, 3.34524106820175, 3.071251954799175, 1.0892108877535873), (297, 446, 3.449564986585881, 2.9193179814989922, 1.1816338639529158), (367, 258, 2.95442797688108, 3.3088360599557882, 1.1199582747821284), (414, 62, 3.121886577927251, 3.3000992697356835, 1.0570849348174447)], [(37, 52, 2.876625767784697, 3.147683005536353, 1.0942274941659855), (190, 288, 3.5976678374446123, 3.0720014677038416, 1.1711152729798917), (335, 285, 2.9805856754573985, 3.140999542418886, 1.0538195792465757), (520, 173, 3.496684518530334, 2.9245051705190597, 1.19564996970401)], [(46, 356, 3.096390401316349, 3.411824626277363, 1.101871593719873), (105, 299, 3.3961414617425776, 3.0550982228439048, 1.1116308589846926), (151, 283, 3.378997929624203, 3.214955296882028, 1.0510248565201727), (343, 272, 3.2962451021039727, 3.178924956809618, 1.0369056039033075), (362, 280, 3.0213073853775305, 3.1229014127964048, 1.0336258494950123), (464, 155, 2.9098095958497163, 3.4026019884029313, 1.1693555458941673)], [(92, 359, 3.126436143745149, 3.124443377854147, 1.0006377986892407), (154, 117, 3.0627772618118527, 3.2348445500343246, 1.0561801507304784), (177, 169, 2.9363640599676897, 3.1972546494587193, 1.088848175554192), (191, 513, 3.454395092767328, 2.9215310517796964, 1.1823920511346369), (365, 355, 3.634003068001854, 3.243894348882502, 1.1202593787475665), (456, 49, 2.8585769726262877, 3.4467780534866264, 1.205767095478956)], [(57, 365, 3.0483891808498598, 3.125629006355453, 1.0253379148537916), (145, 346, 2.952159980558611, 3.2395582855693483, 1.097351873510715), (314, 34, 3.089808114125526, 3.0895549219632876, 1.000081951015157), (344, 48, 3.241789867054261, 2.9180771418688347, 1.1109335735305852), (478, 223, 3.2861077443789393, 2.9138176282118526, 1.1277671301602885), (513, 490, 3.315143376395405, 3.28780308559798, 1.0083156716158541)], [(3, 389, 3.0698019639035174, 3.1549015481471754, 1.027721522510021), (64, 520, 3.283238194919659, 3.0854885892292208, 1.064090208072958), (168, 462, 2.860138991215845, 3.461591956464979, 1.2102880199515955), (198, 89, 3.3708741864174394, 2.9659968409869086, 1.1365063306324397), (288, 352, 3.6086836458650424, 2.80915442138788, 1.2846156189883466), (330, 364, 3.495315045832326, 3.0373359699183102, 1.1507831469583305)], [(140, 29, 2.9756658500494986, 3.2915077489049276, 1.106141588058409), (176, 421, 3.1616155406872832, 3.125325264365802, 1.0116116798258583), (188, 193, 2.914746170069079, 3.2073873849490195, 1.100400239954001), (252, 419, 3.618258504629908, 2.8209730184315753, 1.282627831244417), (342, 392, 2.9246402355738406, 3.044790592107404, 1.0410820979182722), (492, 525, 3.82950071355407, 3.038244680856133, 1.2604319651026172)], [(372, 454, 2.72654208085076, 3.3720772235472243, 1.2367596477715241), (488, 464, 3.1150210505654576, 2.9472576128192305, 1.0569218778217868)], [(47, 226, 3.072079666271771, 3.2832403619994173, 1.0687354231226391), (229, 180, 2.91840525025631, 3.481597715289835, 1.192979527083863), (271, 455, 3.4425956546298444, 3.061968179555648, 1.124308109279383), (339, 61, 3.3101128367031714, 3.0259542005620568, 1.093907117327927), (385, 354, 3.0857804099344457, 3.3218137458710184, 1.0764906456650902), (434, 515, 3.173408548510842, 2.980913406047323, 1.0645758920983774)], [(17, 300, 3.1649698675555653, 3.498362568036368, 1.1053383489992894), (205, 82, 3.0952362150380393, 3.4863663348362106, 1.1263651923875426), (246, 210, 3.244814784700271, 2.7237751977258364, 1.1912931681768255), (273, 21, 3.7614635037691992, 2.8675582106300412, 1.3117304785044817), (478, 218, 2.9138176282118526, 3.2861077443789393, 1.1277671301602885)], [(200, 53, 3.032407579016425, 3.037308634563409, 1.0016162258599068), (205, 310, 2.9343711035554043, 3.4384744971873773, 1.1717926519318502), (273, 173, 3.1480692910314505, 3.1530476263692866, 1.001581393189794), (297, 131, 3.040668334962137, 3.4697075027154467, 1.1411002847039062), (363, 42, 3.3473518933602904, 2.8056164804160364, 1.1930896174604455), (510, 51, 3.284753526381704, 3.0430856446771584, 1.079415405914474)], [(3, 369, 3.0629574486549846, 3.4965535415377267, 1.1415612525316485), (288, 350, 3.2467974565804463, 3.00495655811838, 1.0804806637915259), (331, 75, 3.2575635316482447, 3.0997946146784106, 1.0508965710898244), (396, 95, 3.135638599010506, 2.9816158304041585, 1.0516574828439482), (401, 14, 3.1195086569446957, 3.3724089410369094, 1.0810705504949323), (474, 42, 2.8235328067957193, 3.725014117288715, 1.3192742469020715)], [(28, 267, 3.2049530458481486, 3.10231849511714, 1.0330831766282376), (47, 222, 3.2832403619994173, 3.072079666271771, 1.0687354231226391), (279, 121, 3.2858014365811354, 3.3357559212122108, 1.0152031355501059), (381, 390, 3.0027611703910564, 3.3843466314787767, 1.1270781921820394), (471, 480, 3.205020457075717, 3.0108367001341865, 1.0644949481759924)], [(254, 190, 2.962856358147407, 3.210753938262146, 1.0836684436061363), (277, 507, 3.0687639058669465, 3.377987884291869, 1.1007649946070273), (285, 349, 3.2359940073884235, 3.100394308524693, 1.0437362752508261), (302, 434, 3.2958706371335587, 2.9016360122181775, 1.1358663261881718), (453, 54, 3.1330831755778297, 3.3527425620921076, 1.070109656911284), (455, 58, 3.0655946913007805, 3.0175109184766953, 1.0159349126227384)], [(16, 187, 3.245796314530629, 3.1927847348862914, 1.016603555844245), (123, 71, 3.5808298907629754, 2.987055561154018, 1.1987824857799292), (275, 44, 3.264229398632157, 3.2488882477997443, 1.0047219693822347), (314, 62, 2.8758463135908254, 3.1320717708240733, 1.089095671080323), (344, 37, 3.1328416909197405, 3.588840422510059, 1.1455543486005022), (457, 447, 3.162972285844094, 3.3950573920852594, 1.0733756369854595)], [(59, 135, 3.1877120932423804, 2.9340395938380768, 1.0864584445066978), (180, 322, 3.481597715289835, 3.0837141854361168, 1.1290273695703894), (182, 28, 3.2247423617637536, 3.1693985513724354, 1.0174619283420046), (222, 271, 2.91840525025631, 3.4425956546298444, 1.1796153581917719), (235, 208, 3.079502894572144, 3.409699540178146, 1.1072240088450633), (410, 101, 2.9864000926493626, 3.1810761909498813, 1.0651875476362624)], [(161, 13, 3.0810536976116962, 3.176075588889058, 1.030840712497487), (165, 519, 2.8832139217596824, 3.3518727086120816, 1.1625473515216544), (211, 189, 3.136460660056166, 2.948189460036143, 1.0638599393194068), (239, 129, 3.2380845800754017, 2.897125954566208, 1.1176885751107242), (486, 149, 3.3559258782101136, 3.035752923401623, 1.1054673956962644), (508, 533, 2.976468446276127, 3.174059134900897, 1.0663842712231593)], [(17, 349, 3.0831525375371696, 3.4579590317432674, 1.1215659911868954), (166, 41, 3.4025915291230717, 2.907414386585913, 1.1703152962377097), (273, 274, 3.536002389172755, 2.927402073541498, 1.2078977538247717), (296, 307, 2.8663004191972963, 3.2052241607367087, 1.1182443191472329), (297, 143, 2.891553304084559, 3.647527033142582, 1.2614420865042149), (407, 360, 3.22917791183968, 3.10654096819586, 1.0394770083186904)], [(114, 492, 3.307199283896692, 2.8307377082945337, 1.1683171048331489), (126, 31, 2.9247999932243465, 3.606139440288883, 1.2329524920141348), (178, 531, 3.3193961011549935, 3.185339147081731, 1.042085614084792), (293, 85, 3.099990454425687, 3.28975875278027, 1.0612157686110488), (536, 5, 3.230791796494345, 2.9237407698223867, 1.1050199216843055)], [(31, 438, 2.9496629943236368, 3.267516251131788, 1.1077591770381332), (33, 178, 3.1705519308351504, 3.175564369794431, 1.0015809357703724), (180, 105, 3.1340477576381454, 3.003252404306952, 1.0435512357015415), (235, 339, 2.9933460387622945, 3.813324297458974, 1.2739336675674586), (280, 103, 3.022066890585477, 3.1032296190093924, 1.0268566948920814), (286, 55, 3.692140697545869, 2.88053965307213, 1.281753123449683)], [(5, 536, 3.1747895988805253, 2.9237407698223867, 1.0858656251776349), (112, 174, 3.266707681747268, 3.011192002735864, 1.0848553259902562), (292, 97, 3.0142275902274607, 3.163031138245725, 1.049367057915834), (318, 132, 3.134770912191493, 3.1044817962266333, 1.0097565770885417), (403, 481, 3.3977882959836343, 3.0556383773337377, 1.1119733019417195), (514, 516, 2.922832941983249, 3.328337790845077, 1.1387369230164341)], [(208, 432, 3.409699540178146, 2.9855109693324553, 1.1420824023770166), (229, 59, 3.079502894572144, 3.1877120932423804, 1.0351385279945549), (233, 286, 2.9933460387622945, 3.692140697545869, 1.2334493405488514), (339, 493, 3.813324297458974, 2.875861860153412, 1.3259761709331732), (427, 240, 3.144064388841977, 2.809290099124834, 1.1191668634796506), (524, 66, 3.019026654169993, 3.2853184474249453, 1.0882045187932143)], [(15, 128, 3.0592673672961133, 3.038507275641341, 1.0068323323828114), (194, 265, 3.422904152826007, 3.543154040211609, 1.035130953721366), (208, 158, 3.5569558865360844, 2.970174945965188, 1.1975577032484246), (244, 193, 2.9833897587558273, 3.170659141190159, 1.0627706728176305), (305, 241, 2.97355384989083, 2.928901964930394, 1.0152452644353007), (524, 182, 3.15151015189195, 3.164213033402539, 1.0040307284122068)], [(174, 330, 3.1034638143259983, 3.410976762988592, 1.099087009567527), (175, 403, 3.2294285979661566, 3.1072716813554093, 1.0393132397606963), (202, 117, 3.2231624662351823, 2.9607226342498714, 1.0886404653206574), (462, 49, 3.29331922587049, 3.054939485919205, 1.0780309204323106), (476, 106, 2.902045101487362, 3.3444066323344708, 1.1524309634679313), (526, 25, 3.273137615172774, 3.3131349517058823, 1.0122198762275374)], [(176, 195, 3.262602681808449, 3.5593456951059057, 1.0909528502971049), (252, 376, 3.348906509518689, 3.002209529518101, 1.1154806073965924), (321, 14, 3.411229790093119, 3.0739316644302273, 1.1097285699503057), (400, 42, 3.035598205147131, 3.3520239412124013, 1.1042383460132312), (440, 259, 3.3443852435961494, 2.90106176636602, 1.1528142152538357), (475, 53, 2.9452721120778285, 2.923933570547004, 1.007297888620237)], [(26, 1, 2.9063389361475354, 3.3279022062705366, 1.1450495896675423), (62, 414, 3.1480452859286903, 3.3000992697356835, 1.0483010789224199), (129, 275, 2.897125954566208, 3.5480032285932936, 1.2246630917103303), (162, 78, 3.174032509251582, 3.066698229696414, 1.0349999483208994), (230, 165, 3.2380845800754017, 2.8832139217596824, 1.1230816262496175), (535, 142, 3.425529336320636, 3.089094945862324, 1.1089103431116443)], [(9, 417, 3.4029994978804954, 3.0623682304285205, 1.1112313222385768), (14, 321, 3.1261454948224388, 3.0739316644302273, 1.0169860088291487), (101, 148, 3.0224623827131674, 3.7199427084169594, 1.2307655935415436), (102, 269, 3.3381177300953238, 2.8892631318875877, 1.1553526202767466), (427, 235, 2.809290099124834, 3.144064388841977, 1.1191668634796506), (438, 396, 3.007080393963774, 3.660284789151241, 1.2172221256533977)], [(97, 436, 3.218991588390329, 2.921272624069847, 1.1019141321721992), (305, 236, 2.928901964930394, 2.97355384989083, 1.0152452644353007), (309, 186, 3.0660693898905955, 3.5295348215655666, 1.1511594725165397), (423, 203, 3.1523943164629853, 3.0729890156694966, 1.0258397606983274), (498, 379, 3.3962704448202237, 2.9114306965063483, 1.1665297233060268), (505, 403, 3.095169577403205, 3.3341533033763313, 1.0772118360550795)], [(27, 480, 3.2394840724468446, 3.3331643809653517, 1.028918280325962), (94, 83, 3.4075559351677263, 2.867031536624152, 1.18853102647766), (149, 486, 3.1735042237025324, 3.035752923401623, 1.0453763213860487), (484, 43, 3.5563582499848794, 2.838550026015317, 1.2528784828137072), (500, 517, 2.992581283359904, 3.5945960477628978, 1.2011690602191714), (538, 121, 2.905957443141829, 3.1944587443144403, 1.0992792588389364)], [(88, 484, 3.1138555369376864, 3.059134699492522, 1.0178876848588072), (136, 73, 3.117601552007677, 3.0250031822478287, 1.0306109991233265), (196, 34, 3.415431252338681, 2.8866751297091824, 1.1831713299455238), (211, 346, 3.1330579498590096, 3.252580874895862, 1.038148967222975), (415, 320, 2.8907404467069613, 3.4035965528464205, 1.1774134051791776), (486, 77, 3.216672771382615, 3.2614726317137017, 1.0139273912875604)], [(56, 345, 3.0145031775174886, 3.1857967632202504, 1.0568231564591768), (193, 404, 3.170659141190159, 3.5385316786672907, 1.1160239940957655), (236, 524, 2.9833897587558273, 3.15151015189195, 1.0563521385842105), (432, 114, 3.3911979509571, 3.1727956403744177, 1.068835921167904), (439, 498, 3.3972336288700835, 3.0057396306022124, 1.130248806078201), (531, 12, 3.708078687206466, 2.901443117861705, 1.2780118501648354)], [(88, 138, 3.3930475412230554, 3.186117560709512, 1.0649473776690972), (127, 443, 3.3713278169596106, 2.9442968584178857, 1.1450366518990172), (128, 67, 3.9111861785513002, 2.844405997945784, 1.3750449764822392), (270, 432, 3.0562400838633694, 3.7805428300144, 1.2369914425163369), (345, 482, 2.985119264420868, 3.1965356573462325, 1.0708234325660622), (415, 374, 2.936872962817174, 3.1050739357002395, 1.0572721309408357)], [(187, 361, 2.898755790494772, 3.2483343648453022, 1.120596076253413), (210, 148, 2.7237751977258364, 3.3685529925674706, 1.2367221037110474), (223, 17, 3.244814784700271, 3.1649698675555653, 1.0252277021538827), (249, 257, 3.4561891642112825, 3.249167962016847, 1.0637151432657646), (274, 86, 3.0723624474724454, 3.202585151156336, 1.0423852022377833), (334, 344, 3.677694905995369, 2.8722035622700584, 1.2804436824417442)], [(136, 392, 2.8707040336162977, 3.4464082105535496, 1.200544594704186), (192, 287, 3.550500126099675, 2.9687921568310376, 1.1959409546168995), (196, 26, 2.852041702288446, 3.3372922511155165, 1.1701414633726117), (315, 85, 3.321169291275405, 3.746544006026854, 1.1280798048653808), (367, 143, 3.250385989238422, 3.0855678206478374, 1.0534158307873394), (414, 349, 3.1477674854956144, 3.1887672138631316, 1.013025018066435)], [(10, 134, 2.8263874375632896, 3.5651044108211516, 1.2613643704469375), (39, 441, 3.3389567585642377, 3.150029713250805, 1.0599762740391618), (77, 508, 3.170240749098018, 3.148715476938977, 1.0068362074365533), (149, 88, 3.02177306894235, 3.340135863983737, 1.1053562884365824), (482, 408, 3.234837412702951, 3.1184957778119675, 1.037306972072482), (500, 24, 3.1397777724992073, 3.3810140938077438, 1.0768322915785586)], [(246, 187, 3.4561891642112825, 2.898755790494772, 1.192300909081191), (257, 3, 3.249167962016847, 3.089247696773982, 1.0517667344738544), (353, 470, 3.4159211306402733, 3.093457498754077, 1.1042405244022495), (391, 504, 3.1687908930455437, 3.1195806883122676, 1.0157746215437369), (520, 122, 2.9421384818614675, 3.385914804066517, 1.1508346139860404)], [(37, 513, 3.164685867059168, 3.190457899044174, 1.0081436303846976), (80, 410, 3.373019414179183, 3.060880382596036, 1.1019768800368512), (262, 331, 3.178583225599958, 3.1114446883698683, 1.021577930496738), (322, 359, 2.8700086817590424, 3.4581484975110053, 1.2049261451681355), (409, 60, 3.0025107782376232, 3.702332572027588, 1.2330788614856323)], [(2, 0, 3.2642249277049435, 2.9182822534253523, 1.1185432539548013), (131, 297, 2.938454098771702, 3.4697075027154467, 1.1807935009656312), (185, 521, 2.906445751749418, 3.107583792498345, 1.069204126940219), (289, 35, 3.183916336280324, 3.001260939592734, 1.0608595521562267), (307, 461, 3.1347144034757415, 3.0563451787906124, 1.025641483569647), (448, 537, 3.3976232712963927, 2.943513623693817, 1.1542746885719228)], [(87, 313, 3.2170249092898984, 3.2455910841335984, 1.00887968717967), (220, 188, 3.618258504629908, 2.914746170069079, 1.2413631560048182), (238, 400, 3.348906509518689, 3.035598205147131, 1.1032113880685248), (376, 332, 3.002209529518101, 3.2532322912133718, 1.0836126723425474), (419, 76, 2.8209730184315753, 3.5042965039249507, 1.2422297133041331), (491, 263, 3.1625652639491992, 2.9472928542921117, 1.0730407259474024)], [(60, 409, 2.878770210330181, 3.702332572027588, 1.2860813130350368), (90, 475, 3.195365308638267, 3.2967486865999818, 1.0317282589529395), (108, 284, 3.2313670693024044, 3.194918030235297, 1.0114084426336356), (125, 303, 3.0782664795547707, 3.449869274325318, 1.1207182020265818), (195, 178, 3.3618227097050295, 3.254871878877083, 1.0328586914655589), (287, 192, 3.54265057799015, 2.9687921568310376, 1.193296933851935)], [(6, 360, 3.259401316558566, 3.0230843061052886, 1.0781708303589226), (190, 396, 3.210753938262146, 2.8824299455578726, 1.1139052809280778), (227, 302, 2.962856358147407, 3.2958706371335587, 1.112396363080651), (335, 478, 3.1089477977100253, 3.560956746642364, 1.1453896875545087), (350, 156, 3.4178013436710275, 2.909995393435746, 1.1745040392093988), (490, 107, 3.137866884388211, 3.158648319706941, 1.0066227906040641)], [(117, 27, 3.2422415693317173, 2.978827605861224, 1.088428737182438), (213, 175, 3.230982586641851, 2.987684561939487, 1.0814336385446341), (265, 194, 2.9908871878049306, 3.543154040211609, 1.1846498439187196), (398, 352, 2.983470831116972, 3.150879467424001, 1.0561120405673123), (413, 92, 3.005284211061149, 3.1743458236989808, 1.0562547834962128)], [(140, 90, 2.9498032339900373, 3.207385638772465, 1.0873218938179854), (303, 519, 3.181676493879205, 3.191227033373396, 1.0030017317953488), (342, 317, 2.878727982873827, 3.0066855550598617, 1.0444493446227925), (380, 289, 3.3978113173798925, 2.8664150959691543, 1.1853870439623364), (425, 40, 3.194195550728337, 3.1011763916750605, 1.0299947978783088), (537, 189, 3.219095949691826, 2.999315348624523, 1.0732769234045674)], [(3, 383, 3.089247696773982, 2.985689929339029, 1.034684702660292), (64, 6, 3.394637143330579, 2.9845562292124654, 1.1374009677232053), (144, 9, 3.1954917788659447, 3.1882368714976415, 1.0022755233254972), (249, 246, 3.249167962016847, 3.4561891642112825, 1.0637151432657646), (378, 433, 3.261896222241576, 3.376504623630983, 1.0351355143084988)], [(184, 421, 3.03577605706026, 3.0823420509621258, 1.0153390741038253), (315, 138, 2.8495653270435537, 3.5591587845414234, 1.2490181399820999), (364, 420, 3.622503945686218, 3.042461781810881, 1.1906489564940712), (367, 214, 3.3088360599557882, 2.95442797688108, 1.1199582747821284), (371, 443, 2.970155669117857, 3.284945911017667, 1.105984425386466)], [(76, 72, 3.2747345415273905, 2.973616277035373, 1.1012633226477444), (153, 147, 3.040832647247099, 3.1115023796781927, 1.0232402570707309), (292, 19, 3.5547862934695837, 2.9410499602341855, 1.20867932933262), (440, 238, 2.90106176636602, 3.3443852435961494, 1.1528142152538357), (498, 439, 3.504383987601379, 3.0057396306022124, 1.1658973890893076)], [(40, 207, 2.9757449671413876, 3.024758973861427, 1.0164711718447847), (119, 45, 3.5295764403472374, 2.8609727031815924, 1.233698048367296), (327, 324, 3.3686674438791635, 3.071709328210865, 1.0966752006581506), (426, 380, 3.2048706316714615, 3.1858210881400995, 1.0059794768771786), (450, 124, 3.286542455408161, 2.8160009098688645, 1.1670956653068725), (532, 468, 3.8719363928185238, 3.148261551982456, 1.2298649044519097)], [(46, 71, 3.200397618115541, 2.96001240251731, 1.0812108812090782), (50, 44, 3.0403540217572878, 3.2199867155264656, 1.059082821435825), (343, 267, 3.444289651190979, 3.133914635979198, 1.0990374822748812), (466, 445, 3.265999695059311, 2.945113631390019, 1.108955410157754), (497, 402, 3.4655920409523864, 3.3027656053411603, 1.0493000276337827), (539, 379, 2.9116332517426615, 3.769566238402216, 1.2946569545275206)], [(31, 427, 2.994822956170686, 3.090983443980383, 1.0321089056738941), (250, 409, 3.178583225599958, 3.0025107782376232, 1.0586417369884293), (286, 209, 3.0620821207452087, 2.8875955544377367, 1.0604262484194908), (291, 490, 3.4474034430476546, 3.0102533384727734, 1.1452203703216106), (331, 293, 3.1114446883698683, 3.4271153285196876, 1.1014546848059843)], [(29, 179, 3.2527051357101393, 2.9089835298407305, 1.118158663444969), (53, 200, 3.4841151475264738, 3.037308634563409, 1.1471060622152711), (98, 110, 3.119034785065078, 3.187656288969435, 1.0220008780385965), (119, 470, 3.433613003334017, 3.07554276082464, 1.116425057414376), (426, 22, 3.1444651982098106, 3.0756918560987794, 1.0223602835812895), (491, 252, 2.9472928542921117, 3.1625652639491992, 1.0730407259474024)], [(68, 527, 3.465128170219541, 3.0131714734575383, 1.149993686301362), (213, 201, 3.234052364503101, 3.132714764040242, 1.0323481734200928), (397, 103, 2.9857630944241786, 3.550936878313788, 1.1892895604962945), (413, 55, 3.025198504332868, 3.150639890119308, 1.0414655056872386), (444, 369, 3.248108464429941, 2.927071438464555, 1.1096785755710121)], [(12, 447, 3.538367548005047, 2.8092387429973984, 1.2595467568661265), (75, 331, 3.1695144510532582, 3.0997946146784106, 1.0224917599523222), (194, 236, 3.543154040211609, 3.422904152826007, 1.035130953721366), (209, 359, 2.872580339488628, 3.1505084083885975, 1.0967520612319743), (255, 117, 2.9908871878049306, 3.2422415693317173, 1.084040074313622), (481, 403, 3.1538784563710016, 3.0556383773337377, 1.032150427146744)], [(54, 360, 3.499960000892225, 2.980906007456529, 1.1741262529369658), (164, 281, 3.067592294969673, 3.194291038364821, 1.0413023411236597), (170, 22, 3.732707886210436, 3.130159813213813, 1.1924975429219289), (294, 302, 2.9135073800976095, 3.409766375131361, 1.1703304403564359), (450, 470, 2.8841524044147593, 3.4840431962712333, 1.2079955244175806), (477, 366, 3.4915324946305537, 2.9792715731147723, 1.1719416672647343)], [(28, 226, 3.10231849511714, 3.2049530458481486, 1.0330831766282376), (30, 417, 2.9426803211503123, 3.7138078303502016, 1.2620493648791762), (103, 496, 2.9780404180252473, 3.1375064422967984, 1.0535473001999394), (343, 261, 3.133914635979198, 3.444289651190979, 1.0990374822748812), (362, 180, 3.422718168388607, 3.015138496149572, 1.13517776140616), (473, 150, 3.2173104890362794, 3.212711828237726, 1.001431395358318)], [(93, 107, 3.091276377204276, 3.3306042705064445, 1.0774204128323894), (94, 506, 2.9245739560747106, 3.6797890631069325, 1.2582308118635688), (133, 139, 3.677456638811395, 2.8342681908862484, 1.2974977634919898), (484, 128, 3.0091321088587852, 3.3755193280258093, 1.1217584359584587), (512, 156, 3.2478332285572344, 3.0546269397946677, 1.0632503715087231)], [(89, 198, 3.486821410172158, 2.9659968409869086, 1.1755984908641877), (102, 240, 2.8892631318875877, 3.3381177300953238, 1.1553526202767466), (130, 99, 3.2509586519955556, 3.0207640003589113, 1.0762041164451424), (164, 390, 3.376493808194789, 2.9341200768159585, 1.1507687892101828), (405, 493, 2.888966924829519, 3.3414213646605444, 1.1566146140138747)], [(15, 120, 3.6844097202822934, 3.345725942855324, 1.1012287865807469), (56, 408, 3.0795045638868705, 2.995784295616037, 1.027946026819537), (245, 88, 3.0562400838633694, 3.3930475412230554, 1.110203206592962), (317, 276, 2.887524204671388, 3.0990716275011136, 1.0732625626089949), (320, 303, 2.921339561248391, 3.170698728151715, 1.085357816739648)], [(115, 346, 3.094290643406826, 3.742689264607889, 1.2095467736951735), (141, 471, 3.193672165296996, 3.1110632067527693, 1.0265532883950794), (222, 229, 3.4425956546298444, 2.91840525025631, 1.1796153581917719), (302, 294, 2.9732652215892164, 3.409766375131361, 1.146808683723423), (455, 490, 3.061968179555648, 3.401495787847332, 1.1108854136887067), (493, 341, 3.1274406644300687, 3.3382796900115266, 1.0674158355678605)], [(30, 377, 3.2841098483848894, 3.016247817801816, 1.0888063736017175), (343, 216, 3.178924956809618, 3.2962451021039727, 1.0369056039033075), (422, 110, 3.0051852055567467, 3.2256423017000744, 1.0733589050470802), (449, 366, 3.1095535974346697, 3.049206440639224, 1.0197911023639301), (467, 504, 2.9433991576201515, 3.4465568561604245, 1.1709444324727927), (487, 470, 3.1583318679618064, 3.25709060932503, 1.0312692729871216)], [(21, 428, 2.8675582106300412, 3.277287687316394, 1.1428844496224995), (173, 370, 3.1530476263692866, 3.0946842194410475, 1.0188592446885518), (223, 478, 3.7614635037691992, 2.9138176282118526, 1.2909056034771567), (224, 363, 3.1480692910314505, 3.3473518933602904, 1.0633031181672452), (231, 296, 3.536002389172755, 2.8663004191972963, 1.2336468171619666), (274, 45, 2.927402073541498, 3.3683866811100054, 1.1506402593460676)], [(45, 119, 3.3683866811100054, 2.8609727031815924, 1.1773571545663946), (86, 372, 3.202585151156336, 2.952112913154284, 1.0848450738066202), (200, 483, 3.2474391015351878, 3.225277419614823, 1.006871248279477), (246, 334, 3.0723624474724454, 3.677694905995369, 1.1970250804949512), (273, 231, 2.927402073541498, 3.536002389172755, 1.2078977538247717), (391, 378, 3.2101355137372476, 3.0865705903404836, 1.0400330787131402)], [(23, 46, 2.982592751339909, 3.3464459920722014, 1.1219922634656823), (44, 428, 3.2488882477997443, 2.9683909470503957, 1.0944947298899663), (84, 278, 3.274266926053315, 2.8247435083217547, 1.1591377823888274), (129, 239, 3.5480032285932936, 2.897125954566208, 1.2246630917103303), (155, 52, 3.0635402895715016, 3.4578806009925485, 1.1287204587331225), (228, 16, 3.264229398632157, 3.245796314530629, 1.0056790637228243)], [(99, 529, 2.882490557592832, 3.0861366116095033, 1.070649339502689), (317, 270, 3.0990716275011136, 2.887524204671388, 1.0732625626089949), (376, 380, 3.031858243012121, 3.3130227911993138, 1.0927367065512466), (435, 405, 3.7348930650674763, 2.96508595743958, 1.2596238755562563), (443, 371, 2.8257946704320465, 3.284945911017667, 1.162485705486669), (492, 114, 3.2040255535362903, 2.8307377082945337, 1.1318694572612507)], [(89, 523, 3.13762354975632, 3.077438120403443, 1.019556990911969), (227, 285, 3.0687639058669465, 3.2359940073884235, 1.0544942871629068), (350, 288, 3.1821674428203273, 3.00495655811838, 1.0589728607633888), (405, 298, 3.0793864378240983, 3.2701135256920644, 1.0619367174984164), (452, 442, 2.910912350305406, 3.4299534232323894, 1.1783087260846334), (507, 443, 3.377987884291869, 3.2015314287797194, 1.0551162652741495)], [(63, 1, 3.140447605846574, 3.1304544745456697, 1.0031922301960179), (84, 275, 2.8247435083217547, 3.274266926053315, 1.1591377823888274), (197, 333, 3.0723075884555997, 3.2705462574969792, 1.0645243561504956), (429, 472, 3.0550117659339753, 3.3421805442616788, 1.0939992380814647), (447, 308, 2.907141659495836, 3.675187202751148, 1.264192679000207), (463, 481, 3.6068117681996603, 3.045093794280748, 1.1844665589526087)], [(66, 103, 3.1042302508244948, 2.8733136463853044, 1.0803659582133294), (91, 7, 3.42296143213926, 2.7473184058764906, 1.2459281839402285), (120, 339, 2.844312023537952, 3.473187056864409, 1.2210991720044198), (121, 408, 3.3357559212122108, 3.398897714183604, 1.0189287809008662), (226, 183, 3.2858014365811354, 2.9189783532410742, 1.1256683123164517), (357, 27, 3.236736798443314, 3.169839083784378, 1.021104451327248)], [(103, 377, 3.1032296190093924, 3.151348552946932, 1.015506082322358), (194, 531, 3.3808066626868283, 2.88767197657496, 1.1707724042454333), (209, 413, 3.0198675452235997, 3.314122523483043, 1.097439696891625), (233, 31, 3.022066890585477, 2.9496629943236368, 1.0245464978206578), (362, 216, 3.1229014127964048, 3.0213073853775305, 1.0336258494950123), (524, 120, 3.560537386519781, 2.938311667693787, 1.2117630085559865)], [(7, 91, 3.6086358701979115, 2.7473184058764906, 1.3135120641564773), (79, 181, 3.478460778163109, 3.0450987338632114, 1.142314611832013), (130, 435, 2.8916348192678414, 3.4032749814090018, 1.1769380278353083), (164, 266, 3.194291038364821, 3.067592294969673, 1.0413023411236597), (451, 518, 3.1001421615061404, 3.1322817464501607, 1.0103671326247845), (529, 72, 3.00833323643055, 3.4281753326837094, 1.1395597040809584)], [(0, 2, 3.4439717143736956, 2.9182822534253523, 1.180136606159775), (79, 106, 3.1394930703056216, 3.081674890455086, 1.018761933658095), (89, 453, 3.021147524496202, 3.111111594828166, 1.0297781123240468), (130, 313, 2.9768957020202467, 3.3936783275755507, 1.140005786992288), (312, 330, 3.1897788551448696, 3.0346015623523397, 1.051135969452359), (507, 43, 3.051907528425515, 3.3263598361885505, 1.0899281204318225)], [(13, 142, 3.6729268526725485, 2.969232624117523, 1.2369953175238901), (47, 120, 3.2335245572314344, 3.161473107910798, 1.0227904672477983), (81, 387, 3.0918664171169916, 3.2827906620368275, 1.0617504830942415), (151, 216, 3.214955296882028, 3.378997929624203, 1.0510248565201727), (183, 366, 2.7968257604428253, 3.1674241905261806, 1.1325067994313232), (385, 399, 3.2997236821530174, 2.91040429550998, 1.1337681459732791)], [(23, 464, 2.9422042232503465, 3.9016401787835564, 1.326094275832862), (84, 152, 3.1259377624169575, 3.1334339411342667, 1.0023980575708946), (108, 253, 3.194918030235297, 3.2313670693024044, 1.0114084426336356), (356, 479, 3.1338139361798816, 3.304044022894998, 1.054320419202241), (429, 336, 3.320777572973555, 3.1221794369380085, 1.0636088155875871), (430, 454, 3.1892925570822745, 2.9988308382176925, 1.063511991552608)], [(227, 277, 3.2359940073884235, 3.0687639058669465, 1.0544942871629068), (304, 32, 3.3579319369060863, 3.2648345676273656, 1.0285151873242928), (311, 142, 3.0203257462374817, 3.4339388333119576, 1.1369432047486028), (335, 215, 3.140999542418886, 2.9805856754573985, 1.0538195792465757), (349, 414, 3.100394308524693, 3.1887672138631316, 1.028503763245679), (468, 399, 3.0394688377208783, 3.205259789319713, 1.0545460277602818)], [(55, 107, 2.88053965307213, 3.1994365460046432, 1.1107073435328016), (75, 139, 3.5543682459888606, 2.948548325145212, 1.2054637923608775), (209, 397, 2.8875955544377367, 3.45691442261032, 1.1971601830795306), (233, 235, 3.692140697545869, 2.9933460387622945, 1.2334493405488514), (322, 48, 2.9314608769303043, 3.516994263410728, 1.1997411567346477)], [(16, 334, 3.4145516921573895, 3.128632121115204, 1.0913880443509187), (192, 247, 2.9687921568310376, 3.550500126099675, 1.1959409546168995), (253, 125, 3.54265057799015, 3.0782664795547707, 1.1508589660835815), (457, 84, 2.810868747235087, 3.286717100561865, 1.1692887132474068)], [(190, 215, 3.0720014677038416, 3.5976678374446123, 1.1711152729798917), (219, 168, 3.6086836458650424, 2.860138991215845, 1.2617161812583768), (225, 401, 3.2467974565804463, 3.1195086569446957, 1.040804118094809), (350, 277, 3.00495655811838, 3.1821674428203273, 1.0589728607633888), (352, 398, 2.80915442138788, 3.150879467424001, 1.1216469423803657)], [(11, 116, 3.270994141043471, 3.150100709860709, 1.038377640055866), (35, 465, 3.001260939592734, 3.5027637251728874, 1.1670973619668692), (251, 131, 3.183916336280324, 2.938454098771702, 1.0835344808044567), (380, 256, 2.8664150959691543, 3.3978113173798925, 1.1853870439623364), (495, 99, 3.024746876869135, 3.435410994004304, 1.1357680936132548), (530, 327, 3.4331798474336974, 2.9725959444296794, 1.154943326174922)], [(197, 134, 3.4259981458165476, 2.8531216415113043, 1.200789372584125), (379, 458, 3.0945099510617884, 3.194254999983586, 1.0322329061787547), (384, 505, 3.2498029387353338, 3.0123931085167337, 1.0788110388207262), (429, 406, 3.071636816279431, 3.272915854643634, 1.0655282673060305), (430, 50, 3.084356031551921, 2.9861271819926913, 1.0328950656059064), (516, 337, 3.2284297622478224, 3.0448360231728593, 1.0602967574206672)], [(37, 344, 2.8019706415195613, 3.588840422510059, 1.2808272753934937), (55, 413, 3.0145882945515963, 3.150639890119308, 1.0451310700746779), (190, 64, 3.403744249819916, 2.8941873122502497, 1.176062183471284), (262, 31, 3.4474034430476546, 2.994822956170686, 1.1511209488842902), (352, 85, 3.4187402807798977, 2.9973612644970866, 1.1405833261655607), (490, 455, 3.0102533384727734, 3.401495787847332, 1.1299699411920765)], [(19, 417, 2.9410499602341855, 3.5720858147372283, 1.2145614195730274), (97, 499, 3.163031138245725, 2.9363395436192707, 1.0772021052943486), (234, 112, 3.0142275902274607, 3.266707681747268, 1.0837627829890426), (259, 153, 3.5547862934695837, 3.040832647247099, 1.1690174060344205), (373, 321, 3.047812101965144, 3.263169832281719, 1.0706597792487662), (505, 496, 3.148423981358476, 3.13278416266613, 1.0049923064853072)], [(85, 512, 3.28975875278027, 2.958975925017555, 1.1117896313268425), (139, 12, 3.1804758304375054, 3.111727276895743, 1.0220933736874094), (232, 114, 3.099990454425687, 3.307199283896692, 1.0668417637142027), (331, 262, 3.4271153285196876, 3.1114446883698683, 1.1014546848059843), (474, 383, 2.974433989210993, 3.321428334556755, 1.1166589497714174)], [(266, 477, 2.9135073800976095, 3.4915324946305537, 1.198394937483762), (269, 130, 3.0997494418321256, 3.2509586519955556, 1.0487811073123565), (302, 271, 3.409766375131361, 2.9732652215892164, 1.146808683723423), (453, 6, 2.9972587553559586, 3.1867427394594463, 1.0632190943691093), (515, 101, 3.4323016058156246, 3.1357760424842356, 1.0945620986046167)], [(34, 314, 3.07162976274557, 3.0895549219632876, 1.0058357160863343), (63, 134, 3.299438532626326, 3.0716362432430757, 1.0741631727664254), (133, 398, 3.284837138281858, 3.5352218190275853, 1.0762243819724626), (138, 88, 2.987833648265123, 3.186117560709512, 1.0663637724809485), (326, 184, 3.007889682584272, 3.528025370917465, 1.1729237915023238), (447, 406, 3.1144256692476495, 3.349977071083289, 1.0756323723380246)], [(32, 54, 3.1811745701769287, 3.3469163890282965, 1.0521008247724517), (131, 459, 3.3456786776640586, 3.259043638617252, 1.0265829637935022), (171, 78, 2.819462194035446, 3.397173782885968, 1.2049013425584025), (231, 273, 2.8663004191972963, 3.536002389172755, 1.2336468171619666), (300, 395, 3.530570332397716, 2.97586413575663, 1.1864017210920312), (307, 530, 3.2052241607367087, 2.83193656894518, 1.131813542677818)], [(131, 251, 3.4697075027154467, 2.938454098771702, 1.1807935009656312), (143, 364, 3.647527033142582, 2.9898111245947003, 1.219985772056769), (214, 414, 3.449564986585881, 3.121886577927251, 1.1049616635579982), (224, 200, 3.040668334962137, 3.032407579016425, 1.0027241575317494), (231, 407, 2.891553304084559, 3.22917791183968, 1.1167623668836395), (446, 323, 2.9193179814989922, 3.3418251517534565, 1.1447280402244904)], [(102, 124, 3.3506137930496003, 3.073603511923633, 1.0901255741188942), (339, 120, 2.912216084512726, 3.473187056864409, 1.19262683677044), (405, 277, 3.2701135256920644, 3.0793864378240983, 1.0619367174984164), (427, 31, 3.11587187111503, 3.090983443980383, 1.0080519444978318), (434, 399, 2.8679327352323347, 3.4713076061902726, 1.2103866884831445), (525, 492, 3.1922332835661495, 3.038244680856133, 1.0506834106154428)], [(80, 31, 3.299104346262454, 2.9855491418379967, 1.1050242985554821), (105, 216, 3.0550982228439048, 3.3961414617425776, 1.1116308589846926), (148, 119, 3.5729074508533065, 2.9693673480423777, 1.2032554521113146), (179, 125, 3.0235727713247336, 3.426514514221372, 1.13326675868962), (207, 399, 3.167605438785687, 3.0295404923234535, 1.045572900184723), (409, 334, 3.040957046472888, 3.0796800332461265, 1.0127338157630843)], [(17, 223, 3.498362568036368, 3.1649698675555653, 1.1053383489992894), (22, 426, 3.015505176090646, 3.0756918560987794, 1.0199590703691512), (204, 73, 3.008710102342345, 3.241952225849331, 1.077522298783589), (296, 171, 3.530570332397716, 2.819462194035446, 1.2522141066004058), (304, 453, 3.5066457234995747, 3.1685842466260765, 1.1066916485599105), (395, 346, 2.97586413575663, 3.3367863311644963, 1.1212831564019303)], [(328, 418, 3.0963684429204994, 3.2617784944761694, 1.053420661851099), (412, 390, 3.089324782571677, 3.449065575640786, 1.1164464141480284), (462, 444, 2.9569681641622148, 3.4133899540508272, 1.1543546513014045), (476, 517, 3.1462459279269455, 2.9842909547515726, 1.0542691633057792), (485, 198, 3.0018180010786932, 3.2722066794872653, 1.0900749740028906), (527, 480, 3.5445050311947877, 2.8548439916389876, 1.2415757363889648)], [(227, 254, 3.2958706371335587, 2.962856358147407, 1.112396363080651), (271, 141, 2.9732652215892164, 3.193672165296996, 1.0741295939922815), (294, 266, 3.409766375131361, 2.9135073800976095, 1.1703304403564359), (405, 435, 3.1993324720534164, 2.96508595743958, 1.079001593200392), (434, 410, 2.9016360122181775, 3.286316004925123, 1.1325734830582264), (468, 395, 3.1741542660310587, 3.8093772962903336, 1.2001235532429095)], [(125, 253, 3.449869274325318, 3.0782664795547707, 1.1207182020265818), (160, 207, 2.9943540428130055, 3.3256977198315862, 1.1106561456264217), (256, 380, 3.181676493879205, 3.3978113173798925, 1.0679311123919983), (320, 270, 3.170698728151715, 2.921339561248391, 1.085357816739648), (374, 211, 2.9422319968455835, 3.316338328284098, 1.127150521046471), (519, 479, 3.191227033373396, 3.1714219869932405, 1.0062448474095786)], [(32, 394, 3.2648345676273656, 3.2637076959866436, 1.0003452734575795), (122, 520, 2.9194488753109717, 3.385914804066517, 1.1597787625946554), (285, 311, 3.3579319369060863, 3.0203257462374817, 1.111778072643049), (300, 204, 3.5066457234995747, 3.008710102342345, 1.1654980387673695), (453, 89, 3.1685842466260765, 3.111111594828166, 1.018473349491369)], [(66, 28, 3.4081645936396794, 2.9520911517242228, 1.1544916530267295), (121, 538, 3.1800179260025065, 3.1944587443144403, 1.0045411122352026), (236, 208, 2.97355384989083, 3.5569558865360844, 1.1961968964061886), (241, 97, 2.928901964930394, 3.218991588390329, 1.0990438146900656), (439, 472, 2.9780279908864693, 3.4202928936685315, 1.1485093169491711), (496, 481, 3.3895001371271025, 2.984871923918773, 1.1355596566693227)], [(98, 35, 3.205905299912042, 3.043039875352036, 1.0535206343759018), (119, 148, 3.3277963817927, 2.9693673480423777, 1.1207088890455486), (327, 461, 2.9450350056462007, 3.2780115129259384, 1.1130636840110075), (402, 86, 3.180520120096998, 3.164999972600058, 1.0049036801362718), (422, 466, 3.218521002430938, 3.379528583908741, 1.050025331932336), (487, 417, 3.0511846555034943, 2.9370897025884433, 1.0388462609141627)], [(22, 111, 3.0829645094065925, 3.321571228071647, 1.0773952207159794), (251, 2, 3.1347144034757415, 3.2642249277049435, 1.0413149357675462), (296, 231, 3.2052241607367087, 2.8663004191972963, 1.1182443191472329), (351, 533, 3.3987988580447452, 3.087519286360008, 1.1008186646994897), (530, 98, 2.83193656894518, 3.1779215889423913, 1.122172588112057)], [(139, 75, 3.3986255230881954, 2.948548325145212, 1.152643656576603), (195, 188, 2.9107983575516987, 3.3221297840003476, 1.1413122366863722), (287, 16, 3.001105686360515, 3.4145516921573895, 1.1377645604671345), (326, 315, 3.525171188960263, 2.9754396404389114, 1.1847564108006101), (421, 400, 2.906389174901144, 3.2992999358229596, 1.135188626600627), (447, 278, 3.675187202751148, 2.907141659495836, 1.264192679000207)], [(66, 524, 3.1090535846118237, 3.2853184474249453, 1.0566940575374897), (91, 358, 2.931539171743158, 3.5711697122237127, 1.218189320697433), (186, 449, 3.5295348215655666, 3.0011543505702614, 1.1760590790323415), (241, 498, 3.0660693898905955, 3.3962704448202237, 1.1076952322143663), (319, 451, 3.2257112717365386, 3.0434821176563394, 1.0598752176078257), (377, 96, 3.3434612658195215, 3.0607790057577065, 1.0923563117526796)], [(16, 60, 3.275678363742696, 2.892993229604277, 1.132279996448787), (192, 326, 3.2169178556324765, 3.058801698898756, 1.0516921893925475), (205, 224, 3.4384744971873773, 2.9343711035554043, 1.1717926519318502), (314, 129, 3.0029577502969054, 3.4404919769036355, 1.1457010930518323), (414, 189, 2.9114248400587095, 3.5818481840801604, 1.2302732788416861), (478, 335, 3.186924079010081, 3.560956746642364, 1.1173647876006085)], [(115, 385, 3.2016711374251052, 2.933036898791437, 1.0915891098214139), (142, 177, 3.4339388333119576, 2.9885618757958388, 1.1490271829816197), (145, 62, 3.322432720751312, 3.213869533024464, 1.0337795876936806), (285, 304, 3.0203257462374817, 3.3579319369060863, 1.111778072643049), (455, 452, 3.011513096358524, 3.309464821301094, 1.0989375491352997), (468, 532, 3.9007400491518682, 3.148261551982456, 1.2390139715982549), (513, 37, 3.103837140880805, 3.190457899044174, 1.0279076363326163)], [(156, 512, 3.1734830337780044, 3.0546269397946677, 1.0389101832485397), (181, 476, 2.8976582478437454, 3.5647963538210146, 1.2302335365026953), (198, 6, 3.1940000634227967, 3.045622388587091, 1.0487183425600377), (206, 459, 3.0685606306227906, 3.0415664165780254, 1.0088751026108236), (282, 130, 3.1897788551448696, 2.9768957020202467, 1.0715117943098078), (330, 383, 3.0346015623523397, 3.5939184854393775, 1.184313133567845)], [(14, 401, 2.9936814390045714, 3.3724089410369094, 1.1265089521877347), (87, 252, 3.2455910841335984, 3.2170249092898984, 1.00887968717967), (102, 515, 2.9497667229417783, 3.194631336885613, 1.0830115181785063), (130, 282, 3.3936783275755507, 2.9768957020202467, 1.140005786992288), (451, 319, 3.112557037246623, 3.0434821176563394, 1.0226960162471648), (525, 61, 3.0476161020408923, 3.305610723129084, 1.0846545668647114)], [(34, 295, 3.0895549219632876, 3.07162976274557, 1.0058357160863343), (62, 211, 3.1320717708240733, 2.895029503481432, 1.0818790506478724), (129, 355, 3.4404919769036355, 2.9493032173810203, 1.1665440015214137), (218, 513, 3.089808114125526, 3.315143376395405, 1.0729285618869744), (228, 123, 2.8758463135908254, 3.5808298907629754, 1.2451395173102615), (310, 205, 3.0029577502969054, 3.4384744971873773, 1.1450292621823974)], [(74, 127, 3.1369662879401305, 2.7729076703834727, 1.1312912872812355), (138, 1, 3.5591587845414234, 2.9766155957930542, 1.1957065566584062), (247, 414, 3.321169291275405, 3.1477674854956144, 1.0550872345491835), (258, 364, 2.8495653270435537, 3.622503945686218, 1.2712479027264814), (326, 308, 2.9754396404389114, 3.525171188960263, 1.1847564108006101), (512, 393, 3.424839637326926, 3.1792830587419942, 1.077236463079225)], [(42, 184, 3.162126716443487, 2.8629129738292907, 1.1045137401483716), (109, 440, 3.341528126298138, 3.1658211225777917, 1.0555012418317797), (118, 510, 3.172951957017077, 3.346309910822877, 1.0546361735551695), (340, 146, 2.8655628893584533, 3.369415553740101, 1.17583025877839), (383, 3, 3.5865210815797957, 2.985689929339029, 1.2012369557657911), (420, 386, 3.4138015733001112, 3.109625319869406, 1.0978176539427842)], [(127, 74, 3.3876511537322975, 2.7729076703834727, 1.2216963406011316), (270, 56, 2.887524204671388, 3.0795045638868705, 1.0664861471654161), (342, 256, 3.0066855550598617, 2.878727982873827, 1.0444493446227925)], [(19, 319, 3.3168786486747996, 3.0745729576743233, 1.078809543418271), (96, 397, 3.3027819678555606, 2.974564914733546, 1.110341196958351), (132, 104, 3.1044817962266333, 3.137544757187182, 1.0106500740319158), (234, 514, 3.134770912191493, 2.922832941983249, 1.0725111473748605), (369, 168, 2.985247183390596, 3.3536889089337403, 1.123420843537879), (389, 401, 2.9038180944479026, 3.490305212602498, 1.2019710254151106)], [(7, 106, 3.2937563409351998, 2.986028732316778, 1.1030558096404064), (19, 318, 3.0745729576743233, 3.3168786486747996, 1.078809543418271), (70, 437, 2.9823743972681838, 3.164414220873724, 1.0610385549756216), (97, 76, 3.2659642435249667, 3.450176908273754, 1.056403760425119), (309, 66, 3.2257112717365386, 3.1090535846118237, 1.0375219287638235), (451, 313, 3.0434821176563394, 3.112557037246623, 1.0226960162471648)], [(36, 494, 3.0455042176619225, 3.077568949388999, 1.0105285461570277), (188, 195, 2.9341048773387417, 3.3221297840003476, 1.1322464338812415), (270, 15, 2.921339561248391, 3.6844097202822934, 1.2612055678689456), (303, 160, 3.170698728151715, 2.9943540428130055, 1.058892396429196), (342, 376, 3.2339308003943428, 3.011712636236097, 1.0737846504625237), (415, 243, 3.4035965528464205, 2.8907404467069613, 1.1774134051791776)], [(14, 240, 3.0739316644302273, 3.1261454948224388, 1.0169860088291487), (87, 99, 3.291857348924555, 2.9892034529681553, 1.1012490118917388), (114, 432, 3.2176235365785955, 3.1727956403744177, 1.0141288318836972), (238, 475, 3.411229790093119, 2.9452721120778285, 1.1582053067709819), (373, 292, 3.263169832281719, 3.047812101965144, 1.0706597792487662), (536, 474, 2.821342850488343, 3.391980989157446, 1.2022576371993683)], [(48, 123, 3.516994263410728, 3.09407719269586, 1.136686011491), (52, 191, 3.017535830608052, 3.1757888868798094, 1.0524444663312806), (180, 229, 3.0837141854361168, 3.481597715289835, 1.1290273695703894), (250, 95, 2.8700086817590424, 3.048482905819008, 1.0621859526747417), (286, 75, 2.9314608769303043, 3.5543682459888606, 1.212490425494212), (359, 473, 3.4581484975110053, 3.0304708179470605, 1.1411258201303742)], [(41, 100, 3.445453115205379, 2.9973258262687343, 1.1495090340226717), (51, 454, 2.8722794723512184, 3.5150334477700254, 1.223778354998532), (65, 152, 2.9263605379473026, 3.520885617677185, 1.203161938530962), (416, 521, 3.4144909846861813, 3.018016802491453, 1.1313691102936962), (446, 297, 3.3418251517534565, 2.9193179814989922, 1.1447280402244904)], [(20, 394, 3.0636997056461075, 3.185813171002452, 1.0398581705417478), (38, 11, 3.0796662184477857, 3.2406885688551395, 1.0522856501275364), (327, 260, 3.071709328210865, 3.3686674438791635, 1.0966752006581506), (422, 497, 3.468919414288381, 2.923890215085142, 1.1864054937464088), (445, 81, 3.140389686724783, 3.2178384123229913, 1.0246621385637595), (449, 186, 3.298521965913676, 3.0011543505702614, 1.0990844123984862)], [(4, 445, 3.0235980412432073, 3.4527904914873053, 1.1419475883995571), (78, 416, 3.0988343895170893, 3.3201328519079913, 1.0714134524708783), (100, 511, 2.9908183859185455, 2.9811117561715075, 1.00325604356393), (394, 32, 3.1516737559072543, 3.2637076959866436, 1.0355474420121693), (501, 375, 3.278577997272235, 3.025685445242558, 1.0835819045325128), (509, 535, 3.484062457121131, 3.0621738495350717, 1.1377742180281876)], [(85, 352, 3.670686467563941, 2.9973612644970866, 1.2246393222739629), (184, 42, 3.528025370917465, 2.8629129738292907, 1.2323201589318844), (192, 310, 3.058801698898756, 3.2169178556324765, 1.0516921893925475), (295, 63, 3.007889682584272, 3.299438532626326, 1.0969280395255605), (308, 195, 3.525171188960263, 2.9107983575516987, 1.211066778231014), (315, 74, 2.9754396404389114, 3.1369662879401305, 1.0542866490403389)], [(260, 40, 3.3686674438791635, 2.9757449671413876, 1.1320417176459958), (306, 402, 2.9450350056462007, 3.180520120096998, 1.0799600391843651), (324, 445, 3.071709328210865, 3.140389686724783, 1.022359003139767), (437, 70, 3.346776810378991, 3.164414220873724, 1.0576291777171054), (461, 32, 3.2780115129259384, 3.130450488688438, 1.0471373129109363), (530, 289, 2.9725959444296794, 3.4331798474336974, 1.154943326174922)], [(49, 433, 3.004385266108441, 3.3252551078936508, 1.106800497727387), (69, 477, 3.8006114238665196, 2.7849501053822183, 1.364696414675949), (301, 476, 3.0963684429204994, 3.1462459279269455, 1.0161083817788175), (360, 6, 3.0098183851171116, 3.0230843061052886, 1.0044075486593391), (418, 166, 3.2617784944761694, 3.3170847750298806, 1.0169558664536456), (522, 177, 3.3462948556511924, 3.249469307542079, 1.0297973419488466)], [(111, 22, 3.004512920858778, 3.321571228071647, 1.1055273568676296), (160, 179, 3.1488729823189874, 3.0512097312048057, 1.0320080426184333), (346, 145, 3.0980453100091365, 3.2395582855693483, 1.0456781490906581), (354, 81, 3.1534483454189997, 3.0128464704994413, 1.0466674542816152), (374, 415, 3.0392002857597227, 3.1050739357002395, 1.0216746656181792), (399, 434, 3.362036519848573, 3.4713076061902726, 1.0325014572853661)], [(174, 237, 3.410976762988592, 3.1034638143259983, 1.099087009567527), (219, 64, 3.495315045832326, 3.283238194919659, 1.0645938059690052), (312, 282, 3.0346015623523397, 3.1897788551448696, 1.051135969452359), (364, 143, 3.0373359699183102, 2.9898111245947003, 1.0158956012079365), (383, 5, 3.5939184854393775, 2.8348379712712966, 1.267768571558842), (502, 85, 2.89350683377068, 3.7045030454841044, 1.280281422614293)], [(75, 265, 3.0997946146784106, 3.1695144510532582, 1.0224917599523222), (126, 158, 3.1340390727766527, 3.0247302443652977, 1.0361383725424718), (225, 3, 3.2575635316482447, 3.0629574486549846, 1.063535353087819), (262, 250, 3.1114446883698683, 3.178583225599958, 1.021577930496738), (352, 393, 3.3312093099184645, 3.1010651555817783, 1.074214549772499)], [(2, 441, 3.185476740560508, 3.2378763170187073, 1.0164495241139255), (73, 367, 2.9716826510510073, 3.460162683449088, 1.164378263010459), (376, 252, 3.2532322912133718, 3.002209529518101, 1.0836126723425474), (392, 415, 3.0499540147004445, 3.1023432693563335, 1.017177063786332), (443, 507, 3.288181856526635, 3.2015314287797194, 1.0270653059869985), (448, 22, 3.196709630771691, 3.0202977558449122, 1.058408769329244)], [(51, 428, 3.286308026959834, 2.935853915562805, 1.119370418786606), (118, 363, 3.288114951185095, 2.8688106452151203, 1.1461596312288267), (163, 25, 3.401176617330192, 2.957286324089136, 1.1501005464453218), (197, 278, 3.2705462574969792, 3.0723075884555997, 1.0645243561504956), (340, 112, 3.012836219994319, 3.883754868164459, 1.2890693634092671), (516, 153, 2.932606238867815, 3.360420558372048, 1.1458819509534284)], [(16, 287, 3.128632121115204, 3.4145516921573895, 1.0913880443509187), (215, 190, 3.4317090643370167, 3.5976678374446123, 1.0483603854511652), (246, 274, 3.677694905995369, 3.0723624474724454, 1.1970250804949512), (344, 155, 2.8722035622700584, 3.3182828120155867, 1.1553090649998943), (409, 299, 3.0796800332461265, 3.040957046472888, 1.0127338157630843), (478, 349, 3.0391300358622146, 3.1867930576100907, 1.0485872667524683)], [(122, 407, 2.9739545879481826, 3.396126268716366, 1.1419563306309435), (254, 350, 3.1089477977100253, 3.4178013436710275, 1.0993434325878666), (285, 468, 3.140999542418886, 3.0394688377208783, 1.0334040946358707), (478, 310, 3.560956746642364, 3.186924079010081, 1.1173647876006085), (513, 191, 3.5890192385511375, 2.9215310517796964, 1.2284720493950698)], [(429, 284, 3.1221794369380085, 3.320777572973555, 1.0636088155875871), (439, 408, 2.870153491373596, 3.16733391068969, 1.1035416468872783), (463, 201, 3.715348803558551, 3.0455962777832166, 1.2199085054906962), (496, 103, 3.042748022773085, 3.1375064422967984, 1.0311423814310305), (531, 178, 3.1407232747039346, 3.185339147081731, 1.0142056044023815)], [(109, 5, 3.098682299497014, 3.3645028736306775, 1.0857850364901274), (118, 424, 3.5866980649813254, 3.0559218970508684, 1.173687740004967), (172, 86, 3.2353000717461717, 2.8904003288323894, 1.1193259423178616), (465, 35, 2.9440209979967995, 3.5027637251728874, 1.1897889748599868), (483, 361, 3.2988395029546784, 3.017589794629283, 1.0932034263987653), (516, 290, 3.0448360231728593, 3.2284297622478224, 1.0602967574206672)], [(0, 443, 3.102057576273999, 3.035237755535734, 1.0220146908150434), (79, 523, 2.8771110934075033, 3.065079966624696, 1.0653325044166342), (386, 420, 3.5691269250207465, 3.109625319869406, 1.1477675146955126), (441, 185, 3.3310830782579157, 3.0743803175583073, 1.08349739921035), (528, 203, 3.5687633935730907, 2.8657481118419277, 1.245316494784082), (529, 436, 2.98246172462166, 3.1392209953438908, 1.0525603629471947)], [(120, 279, 3.473187056864409, 2.844312023537952, 1.2210991720044198), (222, 385, 3.3101128367031714, 3.0857804099344457, 1.0726987656174443), (235, 233, 3.813324297458974, 2.9933460387622945, 1.2739336675674586), (298, 102, 2.912216084512726, 3.3506137930496003, 1.1505374930343562), (493, 452, 2.875861860153412, 3.1861949433414414, 1.1079095931163656)], [(112, 369, 3.883754868164459, 3.022632870993381, 1.2848913625716218), (316, 42, 2.8655628893584533, 3.162126716443487, 1.1034923463680915), (333, 51, 3.012836219994319, 3.286308026959834, 1.0907688924975918)], [(43, 512, 3.1519412680291485, 3.2552456555829954, 1.0327748453315697), (61, 208, 3.379598187480926, 2.999047205943419, 1.1268906273910404), (128, 538, 3.3709084045383895, 2.878920604129509, 1.1708931464463368), (317, 342, 3.7177382613326406, 3.0066855550598617, 1.236490545237153), (435, 130, 2.9109775703228946, 3.4032749814090018, 1.1691175555954214), (493, 271, 3.3382796900115266, 3.1274406644300687, 1.0674158355678605)], [(220, 176, 2.9246402355738406, 3.1616155406872832, 1.0810271643776883), (256, 425, 2.878727982873827, 3.194195550728337, 1.109585750974491), (317, 341, 3.0066855550598617, 3.7177382613326406, 1.236490545237153), (320, 36, 3.2339308003943428, 3.0455042176619225, 1.0618704061021063), (376, 495, 3.011712636236097, 3.3767202501444666, 1.1211960296333385), (392, 136, 3.044790592107404, 3.4464082105535496, 1.1319031986919574)], [(18, 183, 2.88205664816274, 3.9697454362400424, 1.3774002113284918), (216, 105, 3.2962451021039727, 3.3961414617425776, 1.0303061078725129), (267, 103, 3.133914635979198, 2.9780404180252473, 1.0523412029636965), (272, 487, 3.178924956809618, 3.1583318679618064, 1.0065202422382233)], [(37, 291, 3.588840422510059, 2.8019706415195613, 1.2808272753934937), (48, 105, 2.9180771418688347, 3.2887539358062767, 1.1270277569495808), (155, 534, 3.3182828120155867, 2.8472809837667254, 1.165421618355967), (218, 145, 3.241789867054261, 2.952159980558611, 1.0981077883322725), (228, 457, 3.1328416909197405, 3.162972285844094, 1.0096176563953692), (334, 246, 2.8722035622700584, 3.677694905995369, 1.2804436824417442)], [(56, 244, 3.1857967632202504, 3.0145031775174886, 1.0568231564591768), (149, 1, 2.9831983446226693, 3.2551179306418545, 1.091150354286476), (245, 127, 2.985119264420868, 3.3713278169596106, 1.1293779304371176), (482, 368, 3.1965356573462325, 2.846856195892433, 1.122830040364642), (538, 357, 3.5269870427718897, 2.9943475132211, 1.177881667775366)], [(115, 271, 3.742689264607889, 3.094290643406826, 1.2095467736951735), (145, 218, 3.2395582855693483, 2.952159980558611, 1.097351873510715), (161, 113, 2.8275996074375875, 3.34798364041703, 1.1840373833730378), (211, 243, 3.252580874895862, 3.1330579498590096, 1.038148967222975), (329, 160, 3.0980453100091365, 3.1488729823189874, 1.0164063682818445), (395, 300, 3.3367863311644963, 2.97586413575663, 1.1212831564019303)], [(175, 444, 2.9686098004008095, 3.3146274966796736, 1.1165588337787427), (201, 83, 3.118274239974471, 3.137631146705446, 1.006207570355048), (526, 517, 3.014494102273366, 2.8482080889984753, 1.0583826771355607)], [(29, 425, 3.0442734927040664, 3.1184179142998456, 1.0243553746972716), (53, 159, 3.2993229664989334, 3.4952607218571887, 1.0593872613708304), (60, 123, 2.9115639298468405, 3.0214975824526973, 1.0377575953180735), (90, 176, 3.460921519198629, 2.905531477377357, 1.1911492083791113), (187, 205, 3.2644519455338115, 3.071251954799175, 1.0629059398506007)], [(17, 231, 3.4579590317432674, 3.0831525375371696, 1.1215659911868954), (145, 535, 3.1371255653115204, 3.067297402623201, 1.0227653707881739), (285, 227, 3.100394308524693, 3.2359940073884235, 1.0437362752508261), (395, 111, 3.3105327913654996, 3.157735145800857, 1.0483883665062386), (414, 247, 3.1887672138631316, 3.1477674854956144, 1.013025018066435), (478, 334, 3.1867930576100907, 3.0391300358622146, 1.0485872667524683)], [(156, 371, 2.909995393435746, 3.5423003903093675, 1.2172872844747282), (198, 485, 3.0174072964615104, 3.2722066794872653, 1.0844431520148294), (254, 335, 3.4178013436710275, 3.1089477977100253, 1.0993434325878666), (277, 405, 3.1821674428203273, 3.0793864378240983, 1.0333771051706178), (288, 225, 3.00495655811838, 3.2467974565804463, 1.0804806637915259), (393, 93, 3.288349748425666, 3.1496268005280887, 1.0440442492660775)], [(20, 477, 3.0476135717573363, 3.586096311790499, 1.1766899665441046), (38, 110, 3.48211300565591, 2.828566332947847, 1.2310522702244555), (171, 166, 2.9388719636411316, 3.066333256722596, 1.0433708220903728), (185, 441, 3.229131495583902, 3.0743803175583073, 1.0503357301443106), (307, 22, 3.3987988580447452, 3.0829645094065925, 1.1024450160468906), (533, 508, 3.087519286360008, 3.174059134900897, 1.0280289256566602)], [(85, 326, 2.9973612644970866, 3.670686467563941, 1.2246393222739629), (288, 219, 2.80915442138788, 3.6086836458650424, 1.2846156189883466), (291, 37, 3.4187402807798977, 2.8019706415195613, 1.2201199506236977), (331, 126, 3.3312093099184645, 3.1340390727766527, 1.062912501268571), (393, 452, 3.1010651555817783, 2.95614769419274, 1.049022402254706), (398, 255, 3.150879467424001, 2.983470831116972, 1.0561120405673123)], [(45, 53, 3.4342162407358168, 3.13962136190888, 1.0938313397918225), (104, 467, 3.105072923156939, 2.9474027330231807, 1.0534946203201876), (137, 80, 3.301591968646269, 3.042575567561341, 1.0851306386097528), (144, 389, 2.8149725066754336, 3.480052570340591, 1.2362652075954506), (249, 520, 3.4159211306402733, 2.9421384818614675, 1.1610334291535616), (470, 450, 3.093457498754077, 3.4840431962712333, 1.1262618599655785)], [(81, 445, 3.0128464704994413, 3.2178384123229913, 1.068039292353841), (135, 71, 3.331922527861232, 3.265405628849882, 1.0203701795647293), (161, 535, 3.023606603281613, 3.0751574581586096, 1.0170494583591154), (165, 8, 3.138441254092852, 3.5671471053931905, 1.1365983354766516), (329, 111, 3.1534483454189997, 3.004512920858778, 1.0495705721637074), (385, 222, 3.3218137458710184, 3.0857804099344457, 1.0764906456650902)], [(41, 166, 3.4292607814026264, 2.907414386585913, 1.1794881380598456), (44, 50, 3.05019794342888, 3.2199867155264656, 1.0556648372487976), (65, 134, 3.3203393469333964, 3.0607587106275296, 1.084809245303968), (129, 314, 2.9493032173810203, 3.4404919769036355, 1.1665440015214137), (162, 508, 3.340544281800449, 3.196567400501707, 1.0450410904134682), (365, 217, 3.243894348882502, 3.634003068001854, 1.1202593787475665)], [(46, 216, 3.411824626277363, 3.096390401316349, 1.101871593719873), (284, 84, 3.1338139361798816, 3.1259377624169575, 1.0025196195067025), (404, 125, 3.147884651538419, 3.149710630874017, 1.0005800655162842), (472, 439, 3.0512526694826603, 3.4202928936685315, 1.120947120465262), (479, 387, 3.304044022894998, 2.9636227254717786, 1.1148666105497715), (539, 147, 3.016537387801347, 3.584041604758573, 1.1881310071780218)], [(27, 117, 3.169839083784378, 2.978827605861224, 1.0641230387241325), (67, 493, 3.235161893017223, 3.007931449815717, 1.075543757227389), (279, 66, 3.236736798443314, 3.1042302508244948, 1.0426857987044051), (358, 106, 3.1764504281928736, 3.1317976422666005, 1.0142578771130168), (471, 169, 3.2090211368760597, 3.079449136784337, 1.0420763566262459), (538, 345, 2.9943475132211, 3.5269870427718897, 1.177881667775366)], [(24, 202, 2.87742930552752, 3.163734676694957, 1.0995004014929042), (91, 309, 3.5711697122237127, 2.931539171743158, 1.218189320697433), (106, 79, 3.1317976422666005, 3.081674890455086, 1.0162647760043606), (203, 482, 3.289238619241666, 3.346759033695243, 1.017487455643105), (357, 67, 3.1764504281928736, 3.235161893017223, 1.0184833562341318), (375, 501, 3.146759004845712, 3.025685445242558, 1.0400152500298814)], [(92, 217, 3.124443377854147, 3.126436143745149, 1.0006377986892407), (209, 265, 3.1505084083885975, 2.872580339488628, 1.0967520612319743), (322, 250, 3.4581484975110053, 2.8700086817590424, 1.2049261451681355), (362, 182, 3.2332423573305014, 2.96734800702388, 1.0896067295366887), (464, 23, 3.010156147040547, 3.9016401787835564, 1.296158733366532), (473, 469, 3.0304708179470605, 3.519831841249142, 1.161480196543714)], [(6, 254, 3.0230843061052886, 3.259401316558566, 1.0781708303589226), (54, 266, 2.980906007456529, 3.499960000892225, 1.1741262529369658), (328, 522, 3.0098183851171116, 3.3462948556511924, 1.1117929481054016), (370, 378, 3.238852375125205, 3.0897242574658446, 1.0482658338520066), (407, 231, 3.10654096819586, 3.22917791183968, 1.0394770083186904), (485, 181, 3.073577049139953, 3.301182238128192, 1.0740522151712213)], [(82, 367, 3.328510619054082, 3.1923083249683812, 1.0426657704145947), (90, 140, 2.9132184594383976, 3.207385638772465, 1.1009766975700053), (187, 246, 3.2483343648453022, 2.898755790494772, 1.120596076253413), (287, 457, 3.6767431624295486, 2.810868747235087, 1.3080451252112641), (483, 337, 3.017589794629283, 3.2988395029546784, 1.0932034263987653)], [(180, 80, 3.015138496149572, 3.070815142469946, 1.018465701124998), (182, 47, 2.96734800702388, 3.4601819597784864, 1.1660856601881682), (216, 46, 3.0213073853775305, 3.096390401316349, 1.024851167511854), (267, 30, 3.422718168388607, 2.9426803211503123, 1.1631294584695646), (280, 194, 3.1229014127964048, 3.3808066626868283, 1.0825851398425934), (359, 92, 3.2332423573305014, 3.124443377854147, 1.0348218758731602)], [(42, 474, 2.8056164804160364, 3.725014117288715, 1.3276989721475916), (82, 192, 3.1182711826832055, 2.988749157646431, 1.0433365325107344), (118, 333, 2.8688106452151203, 3.288114951185095, 1.1461596312288267), (143, 459, 3.1312985828551896, 3.0126371025145566, 1.0393879104262473), (224, 273, 3.3473518933602904, 3.1480692910314505, 1.0633031181672452), (483, 475, 3.0754848317742627, 3.427637026744537, 1.1145029854584312)], [(42, 400, 3.2040972148680726, 3.3520239412124013, 1.0461679894286289), (143, 297, 2.9898111245947003, 3.647527033142582, 1.219985772056769), (206, 0, 3.1418041578635356, 2.8675345151801683, 1.0956465009336198), (258, 315, 3.622503945686218, 2.8495653270435537, 1.2712479027264814), (330, 219, 3.0373359699183102, 3.495315045832326, 1.1507831469583305)], [(57, 218, 3.125629006355453, 3.0483891808498598, 1.0253379148537916), (100, 418, 2.9126757945281554, 3.481468509706455, 1.1952818491666155), (177, 480, 3.9922742340762154, 3.0629140493906934, 1.303423527300872), (217, 154, 3.634003068001854, 3.0627772618118527, 1.186505827019259), (355, 65, 3.243894348882502, 3.3203393469333964, 1.023565810050882), (509, 113, 2.8974169572899404, 3.4181004362623217, 1.1797060922358222), (511, 504, 3.6720185739065716, 3.651034742351997, 1.0057473656197138), (534, 488, 3.7156762935854437, 2.7381465882522193, 1.3570041536589863)], [(183, 283, 3.1674241905261806, 2.7968257604428253, 1.1325067994313232), (381, 91, 3.190171728935117, 3.099734100953242, 1.0291759309142237), (449, 272, 3.049206440639224, 3.1095535974346697, 1.0197911023639301), (477, 266, 2.9792715731147723, 3.4915324946305537, 1.1719416672647343), (501, 522, 3.5132692009249924, 3.1576436462917497, 1.112623713904791)], [(73, 332, 3.460162683449088, 2.9716826510510073, 1.164378263010459), (82, 361, 3.1923083249683812, 3.328510619054082, 1.0426657704145947), (143, 424, 3.0855678206478374, 3.3799451909403677, 1.095404602135993), (214, 17, 2.95442797688108, 3.7720150513603423, 1.276732782412374), (247, 196, 3.250385989238422, 2.852041702288446, 1.1396698676005856), (258, 502, 3.3088360599557882, 3.315875678772114, 1.0021275211853258)], [(39, 25, 2.985321843980429, 3.1407757784455432, 1.0520727554982288), (72, 495, 3.4833562945027334, 3.0651928149213448, 1.136423221908185), (116, 431, 2.9535099091215455, 3.3526642850514214, 1.1351457717128821), (411, 498, 3.1225758557580945, 2.855314492137801, 1.0936013753848153), (482, 345, 2.846856195892433, 3.1965356573462325, 1.122830040364642), (528, 106, 3.473946735876745, 3.154855562587185, 1.101142878638756)], [(3, 225, 3.4965535415377267, 3.0629574486549846, 1.1415612525316485), (112, 340, 3.022632870993381, 3.883754868164459, 1.2848913625716218), (168, 485, 3.3536889089337403, 3.2001342499988508, 1.047983817846062), (318, 19, 2.985247183390596, 3.3168786486747996, 1.1110901191463622), (378, 391, 3.4691879750730155, 3.0865705903404836, 1.123961974474177), (444, 264, 2.927071438464555, 3.248108464429941, 1.1096785755710121)], [(49, 456, 2.9253914593024737, 3.4467780534866264, 1.1782279744224287), (64, 190, 3.314433775477454, 2.8941873122502497, 1.1452036160370214), (168, 174, 2.9894057925463793, 3.3222692373399756, 1.111347695125078), (173, 273, 3.0946842194410475, 3.1530476263692866, 1.0188592446885518), (360, 54, 3.238852375125205, 2.980906007456529, 1.0865328752478076), (378, 132, 3.0897242574658446, 3.2089228877121156, 1.0385790511752775)], [(73, 204, 2.952897820531345, 3.241952225849331, 1.0978883872337897), (74, 88, 3.5802724869024773, 2.991722323200221, 1.196726199867603), (156, 350, 3.5423003903093675, 2.909995393435746, 1.2172872844747282), (206, 386, 2.893935180896006, 3.2875798487998757, 1.1360240099717753), (258, 184, 2.970155669117857, 3.03577605706026, 1.0220932487225132), (443, 276, 3.284945911017667, 2.8257946704320465, 1.162485705486669)], [(50, 23, 3.3223938706022893, 3.1166044902166976, 1.0660299954747492), (86, 274, 2.952112913154284, 3.202585151156336, 1.0848450738066202), (428, 41, 3.2363988564049233, 3.0829160513391756, 1.0497849446789436), (454, 516, 3.3720772235472243, 3.0260897014876855, 1.114334853289195), (497, 422, 3.1804936759101197, 2.923890215085142, 1.08776097662665)], [(5, 383, 3.7158129231622987, 2.8348379712712966, 1.3107673033940366), (292, 505, 3.047812101965144, 3.148423981358476, 1.033011181801024), (321, 114, 3.263169832281719, 3.2176235365785955, 1.0141552593662198), (436, 529, 2.886802082855901, 3.1392209953438908, 1.0874389394364967), (440, 465, 3.1885275761174894, 3.3739098552632516, 1.0581404032802793), (518, 9, 3.247165256651446, 3.160926302000579, 1.0272828109267482)], [(165, 387, 2.9362923747194083, 3.306649538048984, 1.1261308875499727), (211, 62, 3.316338328284098, 2.895029503481432, 1.1455283354784533), (303, 125, 2.9422319968455835, 3.449869274325318, 1.1725347552551877), (329, 399, 3.0392002857597227, 3.362036519848573, 1.106224073352951), (415, 245, 3.1050739357002395, 2.936872962817174, 1.0572721309408357), (537, 448, 3.2565045223019746, 2.943513623693817, 1.1063324103849008)], [(358, 203, 3.146759004845712, 3.289238619241666, 1.0452782097950777), (381, 186, 2.8698072605774385, 3.4709898351454567, 1.2094853486596355), (412, 523, 3.197342863426018, 3.2398145652119354, 1.013283436778628), (471, 141, 3.232783051297128, 3.1110632067527693, 1.0391248381839873), (501, 325, 3.025685445242558, 3.278577997272235, 1.0835819045325128), (527, 68, 3.268209964321089, 3.0131714734575383, 1.0846412137875778)], [(252, 238, 3.002209529518101, 3.348906509518689, 1.1154806073965924), (276, 435, 3.031858243012121, 3.7348930650674763, 1.2318824845045846), (332, 73, 3.2532322912133718, 2.9716826510510073, 1.0947441814026768), (342, 320, 3.011712636236097, 3.2339308003943428, 1.0737846504625237), (380, 426, 3.3130227911993138, 3.1858210881400995, 1.0399274471290147), (495, 72, 3.3767202501444666, 3.0651928149213448, 1.1016338788563669)], [(19, 518, 3.0503922175727975, 3.3604554697014994, 1.1016470112736583), (30, 272, 3.016247817801816, 3.2841098483848894, 1.0888063736017175), (96, 444, 3.0607790057577065, 3.0969526171237685, 1.0118184329211664), (103, 280, 3.151348552946932, 3.1032296190093924, 1.015506082322358), (309, 91, 3.3434612658195215, 2.931539171743158, 1.140513931400557), (505, 384, 3.2495160632417037, 3.0123931085167337, 1.0787158070620226)], [(132, 417, 3.2089228877121156, 3.1562320514194386, 1.0166942212848324), (257, 396, 3.261896222241576, 3.2422203862305774, 1.0060686300334671), (369, 112, 3.4691879750730155, 3.022632870993381, 1.1477371295617766), (370, 360, 3.0897242574658446, 3.238852375125205, 1.0482658338520066), (391, 274, 3.0865705903404836, 3.2101355137372476, 1.0400330787131402), (433, 150, 3.376504623630983, 2.9171697386390814, 1.157459087453098)], [(153, 465, 3.0715219878472495, 3.1547900201437886, 1.0271096976111507), (290, 197, 3.0945099510617884, 3.4259981458165476, 1.107121385937382), (458, 116, 3.194254999983586, 2.936113536996145, 1.087919441715983), (472, 193, 3.348437167898343, 3.060664614438363, 1.0940228968905783), (498, 241, 2.9114306965063483, 3.3962704448202237, 1.1665297233060268), (539, 261, 3.769566238402216, 2.9116332517426615, 1.2946569545275206)], [(256, 303, 3.3978113173798925, 3.181676493879205, 1.0679311123919983), (289, 11, 2.8664150959691543, 3.270994141043471, 1.141144611484655), (376, 276, 3.3130227911993138, 3.031858243012121, 1.0927367065512466), (426, 260, 3.1858210881400995, 3.2048706316714615, 1.0059794768771786), (448, 204, 2.865991172318017, 3.375592196038874, 1.177809697616302), (491, 475, 3.424563619266559, 2.9611721613184683, 1.156489198433422)], [(91, 423, 3.099734100953242, 3.3872471296699556, 1.0927540941748186), (186, 30, 3.4709898351454567, 2.985424835075639, 1.1626451935298903), (226, 47, 3.0027611703910564, 3.2832403619994173, 1.0934070929030408), (375, 527, 2.8698072605774385, 3.268209964321089, 1.1388255961355005), (390, 164, 3.3843466314787767, 2.9341200768159585, 1.1534451702301816)], [(11, 185, 3.380663794831617, 2.9835662994666445, 1.1330949124328022), (35, 491, 3.1028590957340003, 3.19387010806388, 1.0293313391043142), (76, 97, 3.1304691298917766, 3.450176908273754, 1.1021277530991114), (110, 422, 3.253930202955873, 3.2256423017000744, 1.0087696956481782), (153, 516, 2.987514325396019, 3.360420558372048, 1.124821571500447), (458, 494, 3.042233270060502, 3.245500587540031, 1.0668151648592965)], [(3, 257, 2.985689929339029, 3.089247696773982, 1.034684702660292), (5, 373, 2.8348379712712966, 3.7158129231622987, 1.3107673033940366), (112, 175, 3.0284949428780124, 3.3267863922226293, 1.098494947150596), (316, 118, 3.5865210815797957, 3.172951957017077, 1.130342069519237), (330, 312, 3.5939184854393775, 3.0346015623523397, 1.184313133567845), (474, 293, 3.321428334556755, 2.974433989210993, 1.1166589497714174)], [(290, 430, 3.2498029387353338, 3.084356031551921, 1.0536406645312495), (336, 531, 3.3191986960219175, 3.1407232747039346, 1.0568262166729119), (498, 411, 3.003036263749825, 2.855314492137801, 1.0517357271917964), (505, 377, 3.0123931085167337, 3.2495160632417037, 1.0787158070620226)], [(59, 105, 3.454512501724665, 3.1076210765937273, 1.1116260369527313), (115, 311, 2.933036898791437, 3.2016711374251052, 1.0915891098214139), (222, 339, 3.0857804099344457, 3.3101128367031714, 1.0726987656174443), (354, 165, 3.3218137458710184, 3.138441254092852, 1.0584278872637904), (399, 532, 2.91040429550998, 3.0041987741071012, 1.032227302145555)], [(181, 485, 3.0894464651200884, 3.301182238128192, 1.0685351811072323), (206, 371, 3.2875798487998757, 2.893935180896006, 1.1360240099717753), (338, 529, 3.5691269250207465, 2.98246172462166, 1.1967050224168452), (420, 316, 3.109625319869406, 3.4138015733001112, 1.0978176539427842), (503, 489, 3.167594299848571, 2.929382320517427, 1.0813181596894008), (517, 500, 3.05617795148539, 3.5945960477628978, 1.1761736734000785)], [(81, 283, 3.2827906620368275, 3.0918664171169916, 1.0617504830942415), (165, 374, 3.306649538048984, 2.9362923747194083, 1.1261308875499727), (388, 482, 3.2600060772808077, 3.1855367826808774, 1.023377314305333), (431, 38, 2.853350087854791, 3.4230038077954377, 1.199643823015396), (479, 356, 2.9636227254717786, 3.304044022894998, 1.1148666105497715), (508, 162, 3.11968957298559, 3.196567400501707, 1.0246427811862524)], [(8, 415, 3.1855174717881694, 3.2284944458546883, 1.01349136347458), (77, 199, 2.985766560918502, 3.399384245430149, 1.1385298133905042), (116, 11, 3.065655673803655, 3.150100709860709, 1.0275455057717817), (387, 81, 3.2600060772808077, 3.2827906620368275, 1.006989123399127), (482, 203, 3.1855367826808774, 3.346759033695243, 1.050610701433711), (494, 472, 3.464467831972416, 2.851738926758309, 1.2148615006320447)], [(3, 219, 3.1549015481471754, 3.0698019639035174, 1.027721522510021), (5, 109, 3.085719004967335, 3.3645028736306775, 1.090346485929069), (144, 353, 3.480052570340591, 2.8149725066754336, 1.2362652075954506), (318, 96, 2.9038180944479026, 3.3027819678555606, 1.1373928601693324), (401, 126, 3.490305212602498, 2.8324151008252474, 1.2322717851580332), (518, 451, 2.9808921548419507, 3.1322817464501607, 1.0507866718231667)], [(7, 499, 3.2434027105192733, 2.9654167039797557, 1.0937426453983499), (70, 417, 3.0678985125992235, 3.2222168092159738, 1.0503009783351689), (164, 269, 2.9341200768159585, 3.376493808194789, 1.1507687892101828), (381, 226, 3.3843466314787767, 3.0027611703910564, 1.1270781921820394), (412, 301, 3.449065575640786, 3.089324782571677, 1.1164464141480284), (477, 20, 3.12461261058374, 3.586096311790499, 1.1476930931033222)], [(104, 487, 3.02615423982134, 3.2841396845126662, 1.085251915218491), (173, 407, 3.356642030605456, 3.087396935437208, 1.0872078002273846), (249, 95, 3.1687908930455437, 3.0978376636386686, 1.0229041147764775), (274, 200, 3.2101355137372476, 3.2474391015351878, 1.0116205648136365), (378, 369, 3.0865705903404836, 3.4691879750730155, 1.123961974474177)], [(127, 43, 3.6171605270102027, 3.6828543536474854, 1.018161711692564), (136, 247, 3.4464082105535496, 2.8707040336162977, 1.200544594704186), (317, 61, 3.9536899322036336, 3.979943222581933, 1.0066401996182), (332, 2, 3.0499540147004445, 3.185476740560508, 1.0444343505531095), (342, 220, 3.044790592107404, 2.9246402355738406, 1.0410820979182722), (415, 8, 3.1023432693563335, 3.2284944458546883, 1.0406631908675046), (537, 111, 3.0956718264202094, 3.1854773249538106, 1.0290100190101388)], [(93, 117, 3.1496268005280887, 3.162124490867587, 1.003967990854473), (350, 198, 3.288349748425666, 3.0174072964615104, 1.0897931320978402), (352, 331, 3.1010651555817783, 3.3312093099184645, 1.074214549772499), (452, 493, 2.95614769419274, 3.1861949433414414, 1.0778199443825565), (490, 513, 3.13402937258303, 3.28780308559798, 1.049065817429851), (512, 315, 3.1792830587419942, 3.424839637326926, 1.077236463079225)], [(20, 324, 3.185813171002452, 3.0636997056461075, 1.0398581705417478), (32, 304, 3.2637076959866436, 3.2648345676273656, 1.0003452734575795), (69, 412, 3.1378152343196124, 3.2238490624033513, 1.0274183856151728), (171, 489, 3.524441809197798, 3.195103419418893, 1.103075971743914), (325, 509, 3.1516737559072543, 3.484062457121131, 1.1054641840992816), (418, 157, 2.8287032058806267, 3.1984504072422975, 1.1307126179208191)], [(111, 425, 3.157735145800857, 3.285750838754249, 1.0405403515629317), (189, 136, 3.1281491189372574, 3.089841060776554, 1.0123980675404307), (300, 296, 2.97586413575663, 3.530570332397716, 1.1864017210920312), (346, 161, 3.3367863311644963, 2.8275996074375875, 1.180077378136412), (349, 478, 3.3105327913654996, 3.1867930576100907, 1.0388289203341639), (468, 302, 3.8093772962903336, 3.1741542660310587, 1.2001235532429095)], [(190, 254, 2.8824299455578726, 3.210753938262146, 1.1139052809280778), (225, 474, 3.135638599010506, 2.8235328067957193, 1.1105373351652248), (257, 378, 3.2422203862305774, 3.261896222241576, 1.0060686300334671), (438, 240, 3.660284789151241, 3.007080393963774, 1.2172221256533977)], [(96, 318, 2.974564914733546, 3.3027819678555606, 1.110341196958351), (103, 66, 3.550936878313788, 2.8733136463853044, 1.2358333670885353), (209, 286, 3.45691442261032, 2.8875955544377367, 1.1971601830795306), (264, 68, 2.9857630944241786, 3.465128170219541, 1.1605502716175178), (473, 488, 2.936008761342656, 3.327849107421055, 1.1334602100775775), (481, 463, 3.4687935346559393, 3.045093794280748, 1.1391417700075375)], [(55, 52, 3.0958134575487923, 3.0952191523283625, 1.000192007477074), (75, 33, 3.3073730406252664, 3.0305463916862316, 1.0913454582640476), (93, 442, 3.024628397678057, 3.562320484730432, 1.1777712883556704), (133, 295, 3.5352218190275853, 3.284837138281858, 1.0762243819724626), (255, 213, 2.983470831116972, 3.230982586641851, 1.082961010693781), (352, 288, 3.150879467424001, 2.80915442138788, 1.1216469423803657)], [(207, 299, 3.0295404923234535, 3.167605438785687, 1.045572900184723), (329, 374, 3.362036519848573, 3.0392002857597227, 1.106224073352951), (385, 283, 2.91040429550998, 3.2997236821530174, 1.1337681459732791), (434, 298, 3.4713076061902726, 2.8679327352323347, 1.2103866884831445), (468, 285, 3.205259789319713, 3.0394688377208783, 1.0545460277602818), (532, 170, 3.0041987741071012, 2.9135546556875407, 1.0311111783136158)], [(42, 364, 3.3520239412124013, 3.2040972148680726, 1.0461679894286289), (109, 514, 3.0038338182328124, 3.104972040254367, 1.0336697128208827), (238, 252, 3.035598205147131, 3.348906509518689, 1.1032113880685248), (421, 308, 3.2992999358229596, 2.906389174901144, 1.135188626600627), (483, 200, 2.97744842648178, 3.225277419614823, 1.0832353604948528), (536, 401, 3.0439726284496915, 3.101343144545744, 1.0188472509771782)], [(14, 313, 3.3724089410369094, 2.9936814390045714, 1.1265089521877347), (126, 33, 2.8324151008252474, 3.5425560604879798, 1.250719239371315), (225, 288, 3.1195086569446957, 3.2467974565804463, 1.040804118094809), (389, 318, 3.490305212602498, 2.9038180944479026, 1.2019710254151106), (438, 137, 3.1789008328636488, 3.112532883671786, 1.0213228106086931), (536, 400, 3.101343144545744, 3.0439726284496915, 1.0188472509771782)], [(45, 210, 3.108107611415448, 3.0321108513094845, 1.0250639781435242), (86, 21, 3.164999972600058, 3.383813441697338, 1.0691353778804378), (104, 132, 3.185817224560477, 3.137544757187182, 1.0153854274947687), (159, 465, 2.984894306050817, 3.120641116477984, 1.04547792869985), (306, 327, 3.180520120096998, 2.9450350056462007, 1.0799600391843651), (497, 261, 3.3027656053411603, 3.4655920409523864, 1.0493000276337827)], [(96, 469, 2.9545018363495466, 3.3097645139914116, 1.1202445276124153), (146, 197, 3.6995560479066225, 3.0910156202292742, 1.1968739412686016), (175, 237, 3.1072716813554093, 3.2294285979661566, 1.0393132397606963), (234, 5, 3.3977882959836343, 3.1747895988805253, 1.070240464811193), (481, 265, 3.0556383773337377, 3.1538784563710016, 1.032150427146744), (505, 241, 3.3341533033763313, 3.095169577403205, 1.0772118360550795)], [(71, 105, 2.9780131805371033, 3.260562029233435, 1.094878306967524), (108, 457, 3.687105499657023, 3.0723647859199916, 1.2000871499876122), (125, 140, 3.149710630874017, 3.189585314960798, 1.0126597928380856), (193, 244, 3.5385316786672907, 3.170659141190159, 1.1160239940957655), (356, 539, 3.147884651538419, 3.016537387801347, 1.0435423954194072), (519, 165, 2.9272621178104443, 3.3518727086120816, 1.1450538331426365)], [(269, 164, 2.888966924829519, 3.376493808194789, 1.1687547473026327), (277, 350, 3.0793864378240983, 3.1821674428203273, 1.0333771051706178), (298, 427, 3.2701135256920644, 3.11587187111503, 1.0495019246480883), (302, 468, 3.1993324720534164, 3.1741542660310587, 1.007932256567303), (435, 276, 2.96508595743958, 3.7348930650674763, 1.2596238755562563), (493, 67, 3.3414213646605444, 3.007931449815717, 1.1108701845134334)], [(108, 123, 3.1706741115832973, 3.1610094064364813, 1.003057474339411), (193, 36, 2.7199861888244623, 3.150530354474014, 1.1582890999294473), (195, 176, 2.8245756504933506, 3.5593456951059057, 1.260134666417668), (429, 290, 3.272915854643634, 3.071636816279431, 1.0655282673060305), (447, 295, 3.349977071083289, 3.1144256692476495, 1.0756323723380246), (531, 194, 3.160786186448746, 2.88767197657496, 1.0945793746967494)], [(32, 461, 3.2410708053103465, 3.130450488688438, 1.0353368682947146), (122, 335, 3.396126268716366, 2.9739545879481826, 1.1419563306309435), (173, 391, 3.087396935437208, 3.356642030605456, 1.0872078002273846), (231, 297, 3.22917791183968, 2.891553304084559, 1.1167623668836395), (360, 485, 3.10654096819586, 3.073577049139953, 1.0107249366223405), (418, 100, 2.906931178035758, 3.481468509706455, 1.1976439401151966)], [(56, 270, 2.995784295616037, 3.0795045638868705, 1.027946026819537), (121, 279, 3.398897714183604, 3.3357559212122108, 1.0189287809008662), (411, 76, 3.0413975520219747, 3.0560376426610816, 1.0048136063729565), (439, 336, 3.16733391068969, 2.870153491373596, 1.1035416468872783), (482, 248, 3.1184957778119675, 3.234837412702951, 1.037306972072482), (494, 479, 3.1122356998345784, 3.326557849008751, 1.0688643694902555)], [(60, 253, 3.702332572027588, 2.878770210330181, 1.2860813130350368), (210, 45, 3.3056241653773113, 3.0321108513094845, 1.0902055787141502), (250, 262, 3.0025107782376232, 3.178583225599958, 1.0586417369884293), (299, 207, 3.040957046472888, 3.167605438785687, 1.0416475439729391), (334, 478, 3.0796800332461265, 3.0391300358622146, 1.0133426332224733)], [(80, 250, 3.060880382596036, 3.373019414179183, 1.1019768800368512), (101, 470, 3.1810761909498813, 2.951011634577805, 1.0779612501951357), (207, 160, 3.214091474474762, 3.3256977198315862, 1.0347240413793957), (229, 182, 2.9864000926493626, 3.2247423617637536, 1.079809222381502), (427, 158, 3.0366869423404292, 3.295730849069048, 1.085304778413862), (434, 302, 3.286316004925123, 2.9016360122181775, 1.1325734830582264)], [(76, 419, 3.0560376426610816, 3.5042965039249507, 1.1466797578034877), (368, 39, 3.1225758557580945, 2.985321843980429, 1.045976286293695), (408, 56, 3.0413975520219747, 2.995784295616037, 1.0152258146464974), (458, 212, 3.4367751173775245, 3.08547297622838, 1.113856819961058), (498, 384, 2.855314492137801, 3.003036263749825, 1.0517357271917964)], [(69, 394, 3.2238490624033513, 3.1378152343196124, 1.0274183856151728), (106, 528, 3.148663117975456, 3.154855562587185, 1.0019666901093283), (301, 462, 3.089324782571677, 2.9569681641622148, 1.0447609210046949), (375, 471, 3.197342863426018, 3.232783051297128, 1.0110842625845684), (390, 70, 3.449065575640786, 3.0678985125992235, 1.1242437001994), (523, 89, 3.2398145652119354, 3.077438120403443, 1.0527635125242438)], [(55, 291, 3.150639890119308, 3.0145882945515963, 1.0451310700746779), (92, 155, 3.1743458236989808, 3.7900227139742784, 1.1939539434168724), (154, 480, 3.098486227088526, 3.196646151861502, 1.0316799616260393), (209, 280, 3.314122523483043, 3.0198675452235997, 1.097439696891625), (255, 83, 3.005284211061149, 3.0077393645481805, 1.0008169455248177), (264, 444, 3.025198504332868, 3.248108464429941, 1.0736844077430978)], [(62, 239, 3.3000992697356835, 3.1480452859286903, 1.0483010789224199), (189, 537, 3.5818481840801604, 2.999315348624523, 1.1942219365905573), (214, 297, 3.121886577927251, 3.449564986585881, 1.1049616635579982), (247, 315, 3.1477674854956144, 3.321169291275405, 1.0550872345491835), (310, 16, 2.9114248400587095, 3.275678363742696, 1.1251117729958784), (349, 285, 3.1887672138631316, 3.100394308524693, 1.028503763245679)], [(8, 388, 3.2284944458546883, 3.1855174717881694, 1.01349136347458), (243, 196, 2.8907404467069613, 3.415431252338681, 1.1815074079824186), (245, 128, 2.936872962817174, 3.9111861785513002, 1.3317519103037823), (320, 188, 3.4035965528464205, 2.9341048773387417, 1.1600118929400751), (374, 329, 3.1050739357002395, 3.0392002857597227, 1.0216746656181792), (392, 332, 3.1023432693563335, 3.0499540147004445, 1.017177063786332)], [(25, 526, 3.21544374146933, 3.3131349517058823, 1.03038187512866), (78, 325, 3.3201328519079913, 3.0988343895170893, 1.0714134524708783), (199, 26, 3.6331585286290076, 3.1670244662786793, 1.1471836000364233), (323, 51, 3.4144909846861813, 2.8722794723512184, 1.188773939846151), (489, 131, 2.98156925142103, 3.199312171197943, 1.0730296368843808), (521, 116, 3.018016802491453, 3.3030504226293638, 1.0944440136657316)], [(9, 240, 3.0623682304285205, 3.4029994978804954, 1.1112313222385768), (19, 292, 3.5720858147372283, 2.9410499602341855, 1.2145614195730274), (30, 267, 3.7138078303502016, 2.9426803211503123, 1.2620493648791762), (70, 390, 3.2222168092159738, 3.0678985125992235, 1.0503009783351689), (132, 378, 3.1562320514194386, 3.2089228877121156, 1.0166942212848324), (487, 306, 2.9370897025884433, 3.0511846555034943, 1.0388462609141627)], [(100, 365, 3.481468509706455, 2.9126757945281554, 1.1952818491666155), (157, 504, 3.1984504072422975, 3.0565597162645055, 1.0464216976435192), (166, 446, 3.3170847750298806, 3.110684708541374, 1.0663519725807535), (328, 301, 3.2617784944761694, 3.0963684429204994, 1.053420661851099), (394, 20, 2.8287032058806267, 3.185813171002452, 1.1262451162707439), (407, 122, 2.906931178035758, 3.396126268716366, 1.1682857490321328)], [(35, 98, 2.907348597903755, 3.043039875352036, 1.0466718292901362), (76, 411, 3.5042965039249507, 3.0560376426610816, 1.1466797578034877), (252, 220, 2.8209730184315753, 3.618258504629908, 1.282627831244417), (440, 109, 3.0655056573162645, 3.1658211225777917, 1.032723953721015), (495, 2, 3.293156185452739, 3.1075904045375635, 1.059713719235399)], [(174, 168, 3.028603037637261, 3.3222692373399756, 1.0969642426073165), (316, 109, 3.4138015733001112, 3.341528126298138, 1.021628860889476), (364, 258, 3.042461781810881, 3.622503945686218, 1.1906489564940712), (386, 338, 3.109625319869406, 3.5691269250207465, 1.1477675146955126), (424, 446, 3.225967083572584, 3.1657290741692012, 1.0190281631788696)], [(176, 220, 3.125325264365802, 3.1616155406872832, 1.0116116798258583), (184, 258, 3.0823420509621258, 3.03577605706026, 1.0153390741038253), (232, 126, 3.5126677421013923, 2.9247999932243465, 1.2009941706232605), (308, 447, 2.906389174901144, 3.675187202751148, 1.2645199873744208), (361, 187, 3.427308552710056, 3.2483343648453022, 1.0550972183780338), (400, 109, 3.2992999358229596, 3.0038338182328124, 1.0983630039041152)], [(110, 382, 3.2256423017000744, 3.253930202955873, 1.0087696956481782), (272, 30, 3.0051852055567467, 3.2841098483848894, 1.0928144602576892), (306, 119, 3.218521002430938, 3.3277963817927, 1.033952047937306), (324, 20, 3.468919414288381, 3.0636997056461075, 1.1322648260518131), (466, 460, 3.379528583908741, 3.0432168620182027, 1.1105119145756517), (497, 372, 2.923890215085142, 3.1804936759101197, 1.08776097662665)], [(91, 381, 3.3872471296699556, 3.099734100953242, 1.0927540941748186), (121, 15, 2.8492184720480767, 3.636696127537174, 1.2763837393357351), (203, 500, 3.0729890156694966, 3.0799489299440057, 1.0022648679311965), (241, 505, 3.1523943164629853, 3.095169577403205, 1.0184884018884002), (499, 529, 3.261985628310783, 2.8548652815472364, 1.1426057997885286)], [(118, 337, 3.0559218970508684, 3.5866980649813254, 1.173687740004967), (143, 367, 3.3799451909403677, 3.0855678206478374, 1.095404602135993), (163, 134, 2.8946880051608352, 3.345502973928838, 1.155738707578938), (420, 174, 3.225967083572584, 3.028603037637261, 1.0651666935159965), (446, 166, 3.1657290741692012, 3.110684708541374, 1.0176952570849387), (503, 441, 3.2716228988294294, 2.9365714675410746, 1.1140961270623897)], [(29, 348, 3.1184179142998456, 3.0442734927040664, 1.0243553746972716), (111, 395, 3.285750838754249, 3.157735145800857, 1.0405403515629317), (160, 135, 3.2583854758202597, 2.939341080511237, 1.1085428286714964), (256, 342, 3.194195550728337, 2.878727982873827, 1.109585750974491), (426, 530, 2.885298605352015, 3.2911670962045645, 1.1406677596903467)], [(22, 300, 3.0756918560987794, 3.015505176090646, 1.0199590703691512), (260, 450, 3.2048706316714615, 3.286542455408161, 1.0254836569468966), (263, 53, 3.1444651982098106, 3.4841151475264738, 1.1080151720267188), (380, 376, 3.1858210881400995, 3.3130227911993138, 1.0399274471290147), (425, 160, 2.885298605352015, 3.2583854758202597, 1.129306155618069), (530, 38, 3.2911670962045645, 3.008249822504741, 1.094047133846171)], [(31, 262, 3.090983443980383, 2.994822956170686, 1.0321089056738941), (158, 114, 3.295730849069048, 3.1619824187900414, 1.0422989164911887), (235, 524, 3.144064388841977, 3.019026654169993, 1.0414165719601307), (240, 9, 2.809290099124834, 3.4029994978804954, 1.2113378746255565), (298, 405, 3.11587187111503, 3.2701135256920644, 1.0495019246480883), (410, 207, 3.0366869423404292, 3.214091474474762, 1.058420421829062)], [(21, 273, 3.277287687316394, 2.8675582106300412, 1.1428844496224995), (41, 78, 3.0829160513391756, 3.007363979491309, 1.025122357108449), (44, 275, 2.9683909470503957, 3.2488882477997443, 1.0944947298899663), (51, 333, 2.935853915562805, 3.286308026959834, 1.119370418786606), (372, 497, 3.2363988564049233, 3.1804936759101197, 1.0175775166346797), (511, 456, 3.630256445025419, 3.25867886704957, 1.1140270622346649)], [(278, 63, 3.0550117659339753, 3.140447605846574, 1.0279657973384202), (284, 23, 3.320777572973555, 2.9422042232503465, 1.1286699769960178), (290, 516, 3.071636816279431, 3.2284297622478224, 1.0510454052176355), (336, 496, 3.1221794369380085, 3.042748022773085, 1.0261051567761865), (406, 195, 3.272915854643634, 2.8245756504933506, 1.158728339979839), (472, 494, 3.3421805442616788, 2.851738926758309, 1.1719798446139231)], [(50, 488, 2.9861271819926913, 3.5109847583918303, 1.1757653121957428), (167, 65, 2.9203111406261097, 3.4048959948538573, 1.1659360358854958), (284, 108, 3.1892925570822745, 3.194918030235297, 1.0017638623777334), (290, 384, 3.084356031551921, 3.2498029387353338, 1.0536406645312495), (454, 172, 2.9988308382176925, 3.2812755239513693, 1.094184934386477), (539, 212, 3.4811857299456883, 2.833145830699028, 1.2287351015343846)], [(38, 530, 3.4230038077954377, 3.008249822504741, 1.137872188070259), (116, 368, 3.3526642850514214, 2.9535099091215455, 1.1351457717128821), (212, 539, 3.3126500761638735, 2.833145830699028, 1.1692479929091883), (387, 165, 2.853350087854791, 3.306649538048984, 1.1588656968956073), (533, 78, 2.964039885966948, 3.097957167333639, 1.0451806610298038)], [(114, 321, 3.1727956403744177, 3.2176235365785955, 1.0141288318836972), (178, 139, 3.1249274716418287, 3.7227141551470075, 1.1912961785289382), (188, 125, 3.42585678924275, 2.8411066681526678, 1.2058177285790876), (208, 235, 2.9855109693324553, 3.409699540178146, 1.1420824023770166), (244, 439, 3.3911979509571, 3.3972336288700835, 1.0017798070181305), (270, 245, 3.7805428300144, 3.0562400838633694, 1.2369914425163369)], [(49, 328, 3.3252551078936508, 3.004385266108441, 1.106800497727387), (150, 92, 2.9171697386390814, 3.648145528192271, 1.2505770507184146), (378, 257, 3.376504623630983, 3.261896222241576, 1.0351355143084988), (444, 175, 3.2145020843237644, 3.3146274966796736, 1.0311480315549313), (469, 30, 3.0705243808612455, 3.10832802994952, 1.0123117892578566), (504, 86, 2.8820887701096494, 3.8935704552319867, 1.3509543826729027)], [(222, 47, 3.173408548510842, 3.072079666271771, 1.032983806817107), (298, 525, 2.8679327352323347, 3.1922332835661495, 1.113078157081513), (302, 227, 2.9016360122181775, 3.2958706371335587, 1.1358663261881718), (399, 329, 3.4713076061902726, 3.362036519848573, 1.0325014572853661), (410, 80, 3.286316004925123, 3.060880382596036, 1.0736505822347386), (515, 450, 2.980913406047323, 3.202006176155486, 1.074169470894269)], [(130, 281, 3.4032749814090018, 2.8916348192678414, 1.1769380278353083), (276, 376, 3.7348930650674763, 3.031858243012121, 1.2318824845045846), (405, 302, 2.96508595743958, 3.1993324720534164, 1.079001593200392), (507, 156, 3.1469051054161548, 2.7967048951454974, 1.1252188641277572), (525, 158, 3.0832721091795516, 3.399659710101203, 1.1026142324511996)], [(87, 492, 3.1995743239178007, 3.1867179824160847, 1.0040343518229902), (97, 241, 2.921272624069847, 3.218991588390329, 1.1019141321721992), (373, 5, 2.886802082855901, 3.7158129231622987, 1.2871727318023343), (419, 35, 3.751010480560893, 2.907348597903755, 1.2901825681534822), (451, 124, 3.0365815346213734, 3.4183867518437587, 1.1257352100937386), (529, 338, 3.1392209953438908, 2.98246172462166, 1.0525603629471947)], [(70, 319, 3.164414220873724, 2.9823743972681838, 1.0610385549756216), (327, 530, 3.346776810378991, 2.9725959444296794, 1.125876800259546), (449, 18, 3.165586550174065, 3.085273073769242, 1.026031237587247), (450, 515, 2.9994657912547114, 3.202006176155486, 1.0675254858686185), (477, 69, 3.444861340858145, 2.7849501053822183, 1.2369562148350797), (487, 104, 3.0250655488778344, 3.2841396845126662, 1.0856424865672536)], [(31, 233, 3.267516251131788, 2.9496629943236368, 1.1077591770381332), (137, 210, 3.112532883671786, 3.7405267779509352, 1.2017629749627958), (144, 132, 3.239737747426439, 3.279692083334005, 1.012332583382499), (240, 102, 3.007080393963774, 3.3381177300953238, 1.1100859613850211), (396, 190, 3.660284789151241, 2.8824299455578726, 1.26986079741231), (401, 536, 3.1789008328636488, 3.101343144545744, 1.0250077739557146)], [(244, 432, 3.3972336288700835, 3.3911979509571, 1.0017798070181305), (305, 66, 2.9780279908864693, 3.4081645936396794, 1.1444367225793506), (336, 463, 2.870153491373596, 3.715348803558551, 1.2944773910960636), (408, 482, 3.16733391068969, 3.1184957778119675, 1.01566079814031), (472, 356, 3.4202928936685315, 3.0512526694826603, 1.120947120465262), (498, 259, 3.0057396306022124, 3.504383987601379, 1.1658973890893076)], [(109, 316, 3.1658211225777917, 3.341528126298138, 1.0555012418317797), (238, 176, 3.3443852435961494, 3.262602681808449, 1.0250666629570624), (259, 498, 2.90106176636602, 3.504383987601379, 1.2079660034233273), (373, 518, 3.1885275761174894, 3.247165256651446, 1.0183902064931667), (419, 495, 3.0655056573162645, 3.293156185452739, 1.074261982715039), (465, 159, 3.3739098552632516, 3.120641116477984, 1.0811592007321595)], [(2, 332, 3.2378763170187073, 3.185476740560508, 1.0164495241139255), (39, 248, 3.150029713250805, 3.3389567585642377, 1.0599762740391618), (72, 76, 3.18805468486839, 2.973616277035373, 1.0721136783818008), (185, 351, 3.0743803175583073, 3.229131495583902, 1.0503357301443106), (338, 79, 3.3310830782579157, 2.8771110934075033, 1.1577874368113994), (503, 424, 2.9365714675410746, 3.2716228988294294, 1.1140961270623897)], [(27, 24, 2.9186037394267608, 3.4449891696583856, 1.1803552236711008), (58, 142, 3.0629626023982492, 3.276800903645684, 1.0698142057235707), (67, 120, 2.882986538664661, 3.1466162555198802, 1.091443270136574), (93, 398, 3.562320484730432, 3.024628397678057, 1.1777712883556704), (452, 277, 3.4299534232323894, 2.910912350305406, 1.1783087260846334), (484, 88, 3.309906628992037, 3.059134699492522, 1.0819747916105542)], [(0, 338, 3.035237755535734, 3.102057576273999, 1.0220146908150434), (127, 245, 2.9442968584178857, 3.3713278169596106, 1.1450366518990172), (276, 492, 2.8257946704320465, 3.2040255535362903, 1.1338493865325376), (332, 448, 3.288181856526635, 3.196709630771691, 1.0286144931258152), (371, 258, 3.284945911017667, 2.970155669117857, 1.105984425386466), (507, 277, 3.2015314287797194, 3.377987884291869, 1.0551162652741495)], [(96, 377, 3.0969526171237685, 3.0607790057577065, 1.0118184329211664), (175, 347, 3.3146274966796736, 2.9686098004008095, 1.1165588337787427), (264, 413, 3.248108464429941, 3.025198504332868, 1.0736844077430978), (369, 3, 2.927071438464555, 3.4965535415377267, 1.1945569539539163), (433, 504, 3.2145020843237644, 2.8820887701096494, 1.1153376390281928), (462, 301, 3.4133899540508272, 2.9569681641622148, 1.1543546513014045)], [(4, 325, 3.4527904914873053, 3.0235980412432073, 1.1419475883995571), (81, 354, 3.2178384123229913, 3.0128464704994413, 1.068039292353841), (324, 327, 3.140389686724783, 3.071709328210865, 1.022359003139767), (431, 116, 2.9209266070213094, 3.3526642850514214, 1.147808464955026), (466, 261, 2.945113631390019, 3.265999695059311, 1.108955410157754)], [(166, 418, 3.110684708541374, 3.3170847750298806, 1.0663519725807535), (297, 214, 2.9193179814989922, 3.449564986585881, 1.1816338639529158), (323, 65, 3.3418251517534565, 2.9263605379473026, 1.1419731466504746), (424, 420, 3.1657290741692012, 3.225967083572584, 1.0190281631788696), (489, 185, 2.773626144831022, 3.284938671259404, 1.1843480338477748), (510, 172, 3.019743956840392, 3.5032056228928434, 1.1601002180854783)], [(12, 265, 2.8092387429973984, 3.538367548005047, 1.2595467568661265), (278, 197, 2.907141659495836, 3.0723075884555997, 1.0568138564628486), (295, 138, 3.1144256692476495, 2.987833648265123, 1.042369166387838), (308, 421, 3.675187202751148, 2.906389174901144, 1.2645199873744208), (406, 193, 3.349977071083289, 2.7199861888244623, 1.2316154710076301), (457, 228, 3.3950573920852594, 3.162972285844094, 1.0733756369854595)], [(22, 170, 3.0202977558449122, 3.130159813213813, 1.0363745783528444), (204, 214, 3.375592196038874, 2.878342707480302, 1.1727554843508763), (251, 185, 3.3976232712963927, 2.906445751749418, 1.168995935758075), (332, 443, 3.196709630771691, 3.288181856526635, 1.0286144931258152), (380, 491, 2.865991172318017, 3.424563619266559, 1.1948967785887379), (537, 374, 2.943513623693817, 3.2565045223019746, 1.1063324103849008)], [(18, 151, 3.085273073769242, 3.294217853981194, 1.0677232696153818), (186, 309, 3.0011543505702614, 3.5295348215655666, 1.1760590790323415), (272, 467, 3.1095535974346697, 2.9433991576201515, 1.056449849618378), (324, 38, 3.298521965913676, 3.0796662184477857, 1.0710647621988718), (366, 501, 3.049206440639224, 3.5132692009249924, 1.152191322339095), (437, 450, 3.165586550174065, 2.9994657912547114, 1.055383448413947)], [(124, 451, 2.8160009098688645, 3.4183867518437587, 1.21391535772016), (260, 426, 3.286542455408161, 3.2048706316714615, 1.0254836569468966), (266, 54, 2.8841524044147593, 3.499960000892225, 1.2135142357716089), (437, 449, 2.9994657912547114, 3.165586550174065, 1.055383448413947), (470, 353, 3.4840431962712333, 3.093457498754077, 1.1262618599655785), (515, 434, 3.202006176155486, 2.980913406047323, 1.074169470894269)], [(124, 450, 3.4183867518437587, 2.8160009098688645, 1.21391535772016), (281, 79, 3.1001421615061404, 3.478460778163109, 1.1220326671965166), (313, 525, 3.112557037246623, 3.0476161020408923, 1.0213087649596817), (319, 309, 3.0434821176563394, 3.2257112717365386, 1.0598752176078257), (518, 389, 3.1322817464501607, 2.9808921548419507, 1.0507866718231667)], [(43, 74, 2.8660303855669613, 3.605322539716054, 1.2579498660837978), (43, 127, 2.8660303855669613, 3.6828543536474854, 1.2850018521066513), (277, 89, 2.910912350305406, 3.13762354975632, 1.0778832105429514), (393, 352, 2.95614769419274, 3.1010651555817783, 1.049022402254706), (442, 27, 3.4299534232323894, 2.9186037394267608, 1.1752035320512753), (455, 311, 3.309464821301094, 3.011513096358524, 1.0989375491352997), (493, 339, 3.1861949433414414, 2.875861860153412, 1.1079095931163656)], [(6, 64, 3.1867427394594463, 2.9845562292124654, 1.0677442456161503), (54, 69, 3.3527425620921076, 3.0628085891785957, 1.0946627791034333), (89, 282, 3.111111594828166, 3.021147524496202, 1.0297781123240468), (227, 455, 3.1330831755778297, 3.0655946913007805, 1.0220148098731254), (294, 515, 2.9972587553559586, 3.4323016058156246, 1.1451469112175467), (304, 300, 3.1685842466260765, 3.5066457234995747, 1.1066916485599105)], [(51, 323, 3.5150334477700254, 2.8722794723512184, 1.223778354998532), (147, 110, 2.9326911743662105, 3.2784801857011536, 1.1179084297580948), (172, 200, 3.2812755239513693, 2.9401232616568467, 1.1160333196718675), (372, 221, 3.3720772235472243, 2.72654208085076, 1.2367596477715241), (430, 284, 2.9988308382176925, 3.1892925570822745, 1.063511991552608), (516, 514, 3.0260897014876855, 3.328337790845077, 1.0998807435248201)], [(58, 480, 3.0175109184766953, 3.468568995112218, 1.1494801804605321), (227, 453, 3.0655946913007805, 3.1330831755778297, 1.0220148098731254), (271, 222, 3.061968179555648, 3.4425956546298444, 1.124308109279383), (311, 145, 3.011513096358524, 3.322432720751312, 1.1032436567414392), (452, 43, 3.309464821301094, 2.8660303855669613, 1.1547207726642479), (490, 291, 3.401495787847332, 3.0102533384727734, 1.1299699411920765)], [(49, 370, 3.4467780534866264, 2.9253914593024737, 1.1782279744224287), (68, 202, 2.999811588522652, 3.1449007720146405, 1.0483660987400352), (150, 473, 3.163713408874697, 3.212711828237726, 1.0154876289443857), (217, 191, 2.8585769726262877, 3.454395092767328, 1.2084317217435774), (511, 428, 3.25867886704957, 3.630256445025419, 1.1140270622346649), (522, 501, 2.997921556710785, 3.1576436462917497, 1.0532776080226083)], [(34, 196, 3.1565278644537402, 2.8866751297091824, 1.0934821975522213), (84, 167, 3.286717100561865, 3.0053580600169183, 1.0936191411892409), (108, 404, 3.0723647859199916, 3.687105499657023, 1.2000871499876122), (228, 344, 3.162972285844094, 3.1328416909197405, 1.0096176563953692), (287, 361, 2.810868747235087, 3.6767431624295486, 1.3080451252112641), (447, 12, 3.3950573920852594, 2.8092387429973984, 1.2085328812113723)], [(116, 521, 2.936113536996145, 3.3030504226293638, 1.124973670469372), (212, 466, 3.08547297622838, 3.3897054178830555, 1.0986015576861616), (379, 290, 3.194254999983586, 3.0945099510617884, 1.0322329061787547), (382, 35, 3.042233270060502, 3.1028590957340003, 1.0199280660921484), (494, 36, 3.245500587540031, 3.077568949388999, 1.0545663284601219)], [(2, 495, 3.116380488876878, 3.1075904045375635, 1.002828585236484), (73, 136, 3.4081667665448494, 3.0250031822478287, 1.1266655144515578), (131, 296, 3.259043638617252, 3.3456786776640586, 1.0265829637935022), (143, 363, 3.0126371025145566, 3.1312985828551896, 1.0393879104262473), (206, 312, 3.0415664165780254, 3.0685606306227906, 1.0088751026108236), (503, 25, 2.8535159371536993, 3.4057142421753928, 1.1935150590301224)], [(46, 23, 3.333916060095908, 3.3464459920722014, 1.0037583225703448), (81, 113, 3.035301980251455, 3.007910635776715, 1.0091064355931794), (135, 160, 3.308296690530383, 2.939341080511237, 1.1255232380023668), (151, 182, 3.0111434331982165, 3.3427462521398987, 1.11012521532044), (466, 422, 3.0432168620182027, 3.379528583908741, 1.1105119145756517), (479, 494, 3.005724338895261, 3.326557849008751, 1.1067408298098989)], [(20, 4, 2.9941942547357097, 3.1196711017730467, 1.0419067155842943), (32, 407, 3.130450488688438, 3.2410708053103465, 1.0353368682947146), (170, 532, 3.0112431873090935, 2.9135546556875407, 1.0335289854373781), (307, 251, 3.0563451787906124, 3.1347144034757415, 1.025641483569647), (327, 306, 3.2780115129259384, 2.9450350056462007, 1.1130636840110075)], [(49, 157, 3.054939485919205, 3.0396119786127955, 1.005042586821692), (68, 154, 3.0832078877456577, 3.116026228599773, 1.0106442192836083), (168, 219, 3.461591956464979, 2.860138991215845, 1.2102880199515955), (237, 526, 3.29331922587049, 3.273137615172774, 1.0061658301820748), (301, 412, 2.9569681641622148, 3.089324782571677, 1.0447609210046949), (444, 96, 3.4133899540508272, 3.0969526171237685, 1.1021770030246518)], [(12, 139, 3.362314470515118, 3.111727276895743, 1.080529934445078), (201, 506, 3.0455962777832166, 3.081895517744509, 1.0119185987407737), (278, 84, 3.6068117681996603, 2.8247435083217547, 1.2768634594871768), (336, 439, 3.715348803558551, 2.870153491373596, 1.2944773910960636), (481, 397, 3.045093794280748, 3.4687935346559393, 1.1391417700075375)], [(23, 284, 3.9016401787835564, 2.9422042232503465, 1.326094275832862), (155, 57, 3.4026019884029313, 3.0088106883822476, 1.1308793875072327), (216, 151, 2.9098095958497163, 3.378997929624203, 1.161243654720121), (488, 221, 2.9472576128192305, 3.1150210505654576, 1.0569218778217868)], [(35, 289, 3.5027637251728874, 3.001260939592734, 1.1670973619668692), (153, 379, 3.1547900201437886, 3.0715219878472495, 1.0271096976111507), (159, 402, 3.120641116477984, 2.984894306050817, 1.04547792869985), (337, 118, 2.9440209979967995, 3.5866980649813254, 1.218299077153941), (440, 373, 3.3739098552632516, 3.1885275761174894, 1.0581404032802793), (475, 90, 2.9802851928897, 3.2967486865999818, 1.1061856410471365)], [(18, 28, 2.9940080841216705, 3.318291022061406, 1.1083106420652393), (212, 458, 3.3897054178830555, 3.08547297622838, 1.0986015576861616), (261, 50, 3.265999695059311, 3.0403540217572878, 1.074216907533552), (422, 306, 3.379528583908741, 3.218521002430938, 1.050025331932336), (445, 4, 2.945113631390019, 3.4527904914873053, 1.1723793794189448), (460, 135, 3.0432168620182027, 3.308296690530383, 1.0871051392428157)], [(104, 353, 2.9474027330231807, 3.105072923156939, 1.0534946203201876), (272, 449, 2.9433991576201515, 3.1095535974346697, 1.056449849618378), (469, 96, 2.942517665628377, 3.3097645139914116, 1.124807015656305), (497, 147, 3.4343306579968096, 3.1796514647550387, 1.0800965753840543), (504, 157, 3.4465568561604245, 3.0565597162645055, 1.127593495988537)], [(115, 13, 3.1374190947153466, 2.971593517427857, 1.0558035869693996), (285, 335, 3.0394688377208783, 3.140999542418886, 1.0334040946358707), (302, 405, 3.1741542660310587, 3.1993324720534164, 1.007932256567303), (399, 207, 3.205259789319713, 3.0295404923234535, 1.05800196347978), (532, 260, 3.148261551982456, 3.8719363928185238, 1.2298649044519097)], [(30, 186, 3.10832802994952, 2.985424835075639, 1.0411677404937134), (96, 403, 3.3097645139914116, 2.9545018363495466, 1.1202445276124153), (132, 144, 2.9518445796148325, 3.279692083334005, 1.111065299976583), (433, 49, 3.0705243808612455, 3.3252551078936508, 1.0829600079452737), (467, 497, 2.942517665628377, 3.4343306579968096, 1.1671402004186118), (473, 359, 3.519831841249142, 3.0304708179470605, 1.161480196543714)], [(9, 518, 3.234958914758185, 3.160926302000579, 1.0234211764794232), (101, 410, 2.951011634577805, 3.1810761909498813, 1.0779612501951357), (119, 263, 3.07554276082464, 3.433613003334017, 1.116425057414376), (353, 249, 3.093457498754077, 3.4159211306402733, 1.1042405244022495), (450, 266, 3.4840431962712333, 2.8841524044147593, 1.2079955244175806), (487, 272, 3.25709060932503, 3.1583318679618064, 1.0312692729871216)], [(141, 271, 3.1110632067527693, 3.193672165296996, 1.0265532883950794), (226, 28, 3.205020457075717, 3.2049530458481486, 1.000021033452473), (357, 538, 3.2090211368760597, 2.9943475132211, 1.0716929557130894), (375, 412, 3.232783051297128, 3.197342863426018, 1.0110842625845684), (480, 154, 3.0108367001341865, 3.196646151861502, 1.0617135601273342)], [(193, 188, 3.060664614438363, 3.2073873849490195, 1.0479382059107383), (356, 46, 3.0512526694826603, 3.411824626277363, 1.1181717792174313), (379, 153, 3.348437167898343, 3.0715219878472495, 1.0901556886607788), (429, 278, 3.3421805442616788, 3.0550117659339753, 1.0939992380814647), (439, 305, 3.4202928936685315, 2.9780279908864693, 1.1485093169491711), (494, 388, 2.851738926758309, 3.464467831972416, 1.2148615006320447)], [(150, 456, 3.212711828237726, 3.163713408874697, 1.0154876289443857), (267, 28, 3.2173104890362794, 3.10231849511714, 1.0370664695130851), (359, 322, 3.0304708179470605, 3.4581484975110053, 1.1411258201303742), (397, 481, 2.936008761342656, 3.4687935346559393, 1.1814656619313484), (469, 132, 3.519831841249142, 2.9518445796148325, 1.1924177395912974)], [(42, 363, 3.725014117288715, 2.8056164804160364, 1.3276989721475916), (225, 396, 2.8235328067957193, 3.135638599010506, 1.1105373351652248), (293, 139, 2.974433989210993, 3.1804758304375054, 1.069270940950069), (383, 112, 3.321428334556755, 3.0284949428780124, 1.0967257324855773), (502, 156, 2.8864383357192165, 3.4611499249062287, 1.199107523647759), (536, 321, 3.391980989157446, 2.821342850488343, 1.2022576371993683)], [(53, 45, 2.923933570547004, 3.13962136190888, 1.0737663104027104), (90, 253, 3.2967486865999818, 3.195365308638267, 1.0317282589529395), (238, 321, 2.9452721120778285, 3.411229790093119, 1.1582053067709819), (465, 153, 2.9802851928897, 3.1547900201437886, 1.0585530632002664), (483, 363, 3.427637026744537, 3.0754848317742627, 1.1145029854584312), (491, 380, 2.9611721613184683, 3.424563619266559, 1.156489198433422)], [(24, 27, 2.8066976209087193, 3.4449891696583856, 1.2274172835700796), (106, 7, 3.3444066323344708, 2.986028732316778, 1.120018235638221), (181, 312, 3.5647963538210146, 2.8976582478437454, 1.2302335365026953), (237, 175, 2.902045101487362, 3.2294285979661566, 1.1128113054862598), (301, 328, 3.1462459279269455, 3.0963684429204994, 1.0161083817788175)], [(20, 351, 3.586096311790499, 3.0476135717573363, 1.1766899665441046), (69, 328, 2.7849501053822183, 3.8006114238665196, 1.364696414675949), (266, 294, 3.4915324946305537, 2.9135073800976095, 1.198394937483762), (366, 183, 2.9792715731147723, 3.1674241905261806, 1.0631538994663379), (390, 7, 3.12461261058374, 3.2434027105192733, 1.0380175448096078), (437, 487, 3.444861340858145, 3.0250655488778344, 1.1387724613557668)], [(218, 57, 3.2861077443789393, 3.0483891808498598, 1.0779816976855974), (223, 273, 2.9138176282118526, 3.7614635037691992, 1.2909056034771567), (310, 192, 3.186924079010081, 3.2169178556324765, 1.0094115127561218), (334, 409, 3.0391300358622146, 3.0796800332461265, 1.0133426332224733), (335, 254, 3.560956746642364, 3.1089477977100253, 1.1453896875545087), (349, 395, 3.1867930576100907, 3.3105327913654996, 1.0388289203341639)], [(212, 110, 3.0313304458151533, 3.1962993527197385, 1.0544212878976391), (356, 284, 3.304044022894998, 3.1338139361798816, 1.054320419202241), (387, 508, 2.9636227254717786, 3.11968957298559, 1.0526608350558408), (460, 151, 3.005724338895261, 3.0111434331982165, 1.0018029245838782), (494, 408, 3.326557849008751, 3.1122356998345784, 1.0688643694902555), (519, 303, 3.1714219869932405, 3.191227033373396, 1.0062448474095786)], [(27, 242, 3.3331643809653517, 3.2394840724468446, 1.028918280325962), (58, 455, 3.468568995112218, 3.0175109184766953, 1.1494801804605321), (154, 413, 3.196646151861502, 3.098486227088526, 1.0316799616260393), (177, 365, 3.0629140493906934, 3.9922742340762154, 1.303423527300872), (177, 509, 3.0629140493906934, 3.342922791689961, 1.0914190662173395), (471, 226, 3.0108367001341865, 3.205020457075717, 1.0644949481759924), (527, 301, 2.8548439916389876, 3.5445050311947877, 1.2415757363889648)], [(213, 202, 3.276017620252854, 3.0248307292823786, 1.0830416355331287), (265, 75, 3.1538784563710016, 3.1695144510532582, 1.004957703633338), (397, 473, 3.4687935346559393, 2.936008761342656, 1.1814656619313484), (403, 234, 3.0556383773337377, 3.3977882959836343, 1.1119733019417195), (463, 278, 3.045093794280748, 3.6068117681996603, 1.1844665589526087), (496, 305, 2.984871923918773, 3.3895001371271025, 1.1355596566693227)], [(203, 358, 3.346759033695243, 3.289238619241666, 1.017487455643105), (248, 10, 3.234837412702951, 2.8263874375632896, 1.144513087523414), (345, 245, 3.1965356573462325, 2.985119264420868, 1.0708234325660622), (368, 72, 2.846856195892433, 3.4833562945027334, 1.2235799966042085), (388, 387, 3.1855367826808774, 3.2600060772808077, 1.023377314305333), (408, 439, 3.1184957778119675, 3.16733391068969, 1.01566079814031)], [(200, 274, 3.225277419614823, 3.2474391015351878, 1.006871248279477), (337, 516, 3.2988395029546784, 3.0448360231728593, 1.0834210702476963), (363, 143, 3.0754848317742627, 3.1312985828551896, 1.0181479519925734), (400, 536, 2.97744842648178, 3.0439726284496915, 1.0223426882481783), (475, 491, 3.427637026744537, 2.9611721613184683, 1.1575271007607928)], [(43, 507, 2.838550026015317, 3.3263598361885505, 1.1718517573065317), (88, 243, 3.059134699492522, 3.1138555369376864, 1.0178876848588072), (128, 15, 3.3755193280258093, 3.038507275641341, 1.1109136894573783), (242, 500, 3.5563582499848794, 2.992581283359904, 1.1883915300011494), (268, 133, 3.0091321088587852, 3.677456638811395, 1.2220987666128331), (442, 58, 3.309906628992037, 3.0629626023982492, 1.0806226058393382)], [(168, 369, 3.2001342499988508, 3.3536889089337403, 1.047983817846062), (181, 386, 3.301182238128192, 3.0894464651200884, 1.0685351811072323), (198, 350, 3.2722066794872653, 3.0174072964615104, 1.0844431520148294), (301, 527, 3.0018180010786932, 3.5445050311947877, 1.1807861202514882), (360, 407, 3.073577049139953, 3.10654096819586, 1.0107249366223405), (523, 164, 3.146700520573147, 3.084328565224637, 1.020222214990888)], [(8, 36, 3.639768024261627, 2.9051824962686053, 1.252853488184139), (26, 152, 3.003139989722938, 3.4189587735602154, 1.1384613388853844), (77, 521, 3.2614726317137017, 3.2656070472777854, 1.0012676529994093), (149, 242, 3.035752923401623, 3.1735042237025324, 1.0453763213860487), (230, 161, 3.3559258782101136, 3.0810536976116962, 1.089213693617702), (243, 136, 3.216672771382615, 3.117601552007677, 1.0317780247803439)], [(104, 391, 3.2841396845126662, 3.02615423982134, 1.085251915218491), (272, 343, 3.1583318679618064, 3.178924956809618, 1.0065202422382233), (306, 98, 3.0511846555034943, 3.205905299912042, 1.0507083844071758), (417, 19, 2.9370897025884433, 3.5720858147372283, 1.2161990870041068), (437, 477, 3.0250655488778344, 3.444861340858145, 1.1387724613557668), (470, 101, 3.25709060932503, 2.951011634577805, 1.1037200162685956)], [(50, 430, 3.5109847583918303, 2.9861271819926913, 1.1757653121957428), (473, 397, 3.327849107421055, 2.936008761342656, 1.1334602100775775), (534, 365, 2.7381465882522193, 3.7156762935854437, 1.3570041536589863)], [(131, 204, 3.199312171197943, 3.0359070108694097, 1.053824165148503), (171, 394, 3.195103419418893, 3.524441809197798, 1.103075971743914), (185, 11, 3.284938671259404, 2.9835662994666445, 1.1010107842572945), (416, 199, 2.98156925142103, 3.6331585286290076, 1.2185390384266361), (446, 510, 2.773626144831022, 3.019743956840392, 1.0887350346289602), (503, 386, 2.929382320517427, 3.167594299848571, 1.0813181596894008)], [(107, 154, 3.158648319706941, 3.136010449901598, 1.0072186844294646), (254, 6, 3.137866884388211, 3.259401316558566, 1.0387315449151218), (291, 262, 3.0102533384727734, 3.4474034430476546, 1.1452203703216106), (393, 512, 3.13402937258303, 3.1792830587419942, 1.0144394582114802), (455, 271, 3.401495787847332, 3.061968179555648, 1.1108854136887067), (513, 218, 3.28780308559798, 3.315143376395405, 1.0083156716158541)], [(35, 382, 3.19387010806388, 3.1028590957340003, 1.0293313391043142), (140, 125, 3.1828888208780963, 3.189585314960798, 1.0021039044904039), (252, 87, 3.1625652639491992, 3.2170249092898984, 1.0172200858466058), (263, 119, 2.9472928542921117, 3.433613003334017, 1.1650057096747892), (380, 448, 3.424563619266559, 2.865991172318017, 1.1948967785887379), (475, 483, 2.9611721613184683, 3.427637026744537, 1.1575271007607928)], [(87, 436, 3.1867179824160847, 3.1995743239178007, 1.0040343518229902), (114, 232, 2.8307377082945337, 3.307199283896692, 1.1683171048331489), (276, 443, 3.2040255535362903, 2.8257946704320465, 1.1338493865325376), (525, 298, 3.038244680856133, 3.1922332835661495, 1.0506834106154428)], [(67, 357, 3.007931449815717, 3.235161893017223, 1.075543757227389), (271, 115, 3.1274406644300687, 3.094290643406826, 1.0107132861271055), (339, 235, 2.875861860153412, 3.813324297458974, 1.3259761709331732), (405, 269, 3.3414213646605444, 2.888966924829519, 1.1566146140138747), (452, 393, 3.1861949433414414, 2.95614769419274, 1.0778199443825565)], [(36, 320, 3.077568949388999, 3.0455042176619225, 1.0105285461570277), (388, 77, 3.464467831972416, 2.985766560918502, 1.1603277621632457), (408, 121, 3.1122356998345784, 3.398897714183604, 1.092108066996424), (458, 382, 3.245500587540031, 3.042233270060502, 1.0668151648592965), (472, 429, 2.851738926758309, 3.3421805442616788, 1.1719798446139231), (479, 460, 3.326557849008751, 3.005724338895261, 1.1067408298098989)], [(2, 459, 3.1075904045375635, 3.116380488876878, 1.002828585236484), (72, 368, 3.0651928149213448, 3.4833562945027334, 1.136423221908185), (99, 130, 3.435410994004304, 3.0207640003589113, 1.1372656035347768), (289, 530, 3.024746876869135, 3.4331798474336974, 1.135030462776219), (376, 342, 3.3767202501444666, 3.011712636236097, 1.1211960296333385), (419, 440, 3.293156185452739, 3.0655056573162645, 1.074261982715039)], [(103, 267, 3.1375064422967984, 2.9780404180252473, 1.0535473001999394), (194, 33, 3.3400228459626757, 3.0216199233128562, 1.1053749084036777), (305, 121, 3.3895001371271025, 3.1800179260025065, 1.0658745378168133), (336, 429, 3.042748022773085, 3.1221794369380085, 1.0261051567761865), (481, 213, 2.984871923918773, 3.276017620252854, 1.0975404318024613), (505, 292, 3.13278416266613, 3.148423981358476, 1.0049923064853072)], [(147, 153, 3.1796514647550387, 3.1115023796781927, 1.0219023085188494), (261, 46, 3.4655920409523864, 3.200397618115541, 1.0828629609445208), (372, 428, 3.1804936759101197, 3.2363988564049233, 1.0175775166346797), (402, 45, 3.3027656053411603, 3.108107611415448, 1.062629103706311), (422, 324, 2.923890215085142, 3.468919414288381, 1.1864054937464088), (467, 469, 3.4343306579968096, 2.942517665628377, 1.1671402004186118)], [(241, 309, 3.3962704448202237, 3.0660693898905955, 1.1076952322143663), (259, 440, 3.504383987601379, 2.90106176636602, 1.2079660034233273), (379, 539, 2.9114306965063483, 3.769566238402216, 1.2947470269258379), (384, 146, 3.003036263749825, 3.0810256359088237, 1.0259701732877562), (411, 368, 2.855314492137801, 3.1225758557580945, 1.0936013753848153), (439, 244, 3.0057396306022124, 3.3972336288700835, 1.130248806078201)], [(7, 390, 2.9654167039797557, 3.2434027105192733, 1.0937426453983499), (97, 292, 2.9363395436192707, 3.163031138245725, 1.0772021052943486), (423, 121, 3.261985628310783, 2.8492184720480767, 1.1448703075289277), (528, 517, 3.2768631840781297, 3.347703915308649, 1.021618458645062), (529, 99, 2.8548652815472364, 3.0861366116095033, 1.0810095423966646)], [(24, 527, 3.3810140938077438, 2.9844632888820466, 1.1328717315448171), (203, 423, 3.0799489299440057, 3.0729890156694966, 1.0022648679311965), (242, 484, 2.992581283359904, 3.5563582499848794, 1.1883915300011494), (248, 77, 3.1397777724992073, 3.170240749098018, 1.0097022715638129), (506, 201, 3.07818936967174, 3.081895517744509, 1.0012040026222182), (517, 386, 3.5945960477628978, 3.05617795148539, 1.1761736734000785)], [(69, 54, 3.2904890729532625, 3.0628085891785957, 1.0743371572677116), (169, 13, 3.0717759742918753, 3.1659410096102434, 1.0306549162785466), (325, 78, 3.278577997272235, 3.0988343895170893, 1.0580036185099766), (366, 449, 3.5132692009249924, 3.049206440639224, 1.152191322339095), (375, 358, 3.025685445242558, 3.146759004845712, 1.0400152500298814), (522, 456, 3.1576436462917497, 2.997921556710785, 1.0532776080226083)], [(85, 133, 3.7045030454841044, 2.865310018588406, 1.2928803590018252), (156, 507, 3.4611499249062287, 2.7967048951454974, 1.2375813876230097), (258, 367, 3.315875678772114, 3.3088360599557882, 1.0021275211853258), (288, 190, 3.3624883812011683, 3.0720014677038416, 1.094559496976559), (330, 174, 2.89350683377068, 3.410976762988592, 1.1788383297313902), (474, 536, 2.8864383357192165, 3.391980989157446, 1.1751441030914187)], [(25, 10, 3.4057142421753928, 3.218576284688825, 1.0581430859280256), (386, 181, 3.167594299848571, 3.0894464651200884, 1.0252950927005122), (424, 118, 3.2716228988294294, 3.0559218970508684, 1.0705845924880228), (441, 72, 2.9365714675410746, 3.18805468486839, 1.0856383779884282), (459, 73, 2.8535159371536993, 3.4081667665448494, 1.1943745335953506), (489, 171, 2.929382320517427, 3.195103419418893, 1.090708917385195)], [(86, 172, 3.8935704552319867, 2.8904003288323894, 1.3470696139883294), (157, 418, 3.0565597162645055, 3.1984504072422975, 1.0464216976435192), (221, 50, 2.68197383390531, 3.8718860490644866, 1.4436703297088125), (391, 249, 3.1195806883122676, 3.1687908930455437, 1.0157746215437369), (433, 444, 2.8820887701096494, 3.2145020843237644, 1.1153376390281928), (467, 272, 3.4465568561604245, 2.9433991576201515, 1.1709444324727927), (511, 365, 3.651034742351997, 3.6720185739065716, 1.0057473656197138)], [(241, 423, 3.095169577403205, 3.1523943164629853, 1.0184884018884002), (292, 373, 3.148423981358476, 3.047812101965144, 1.033011181801024), (377, 30, 3.2495160632417037, 3.016247817801816, 1.0773372280829), (384, 290, 3.0123931085167337, 3.2498029387353338, 1.0788110388207262), (403, 175, 3.3341533033763313, 3.1072716813554093, 1.0730163452981218), (496, 194, 3.13278416266613, 3.3400228459626757, 1.0661515995152941)], [(10, 199, 3.2516528233106405, 2.9278889446082474, 1.1105792893199071), (94, 268, 3.6797890631069325, 2.9245739560747106, 1.2582308118635688), (201, 463, 3.081895517744509, 3.0455962777832166, 1.0119185987407737), (202, 68, 3.3723670953921556, 3.1449007720146405, 1.0723286169794792), (500, 203, 3.07818936967174, 3.0799489299440057, 1.0005716218402942)], [(43, 484, 3.3263598361885505, 2.838550026015317, 1.1718517573065317), (156, 502, 2.7967048951454974, 3.4611499249062287, 1.2375813876230097), (277, 227, 3.377987884291869, 3.0687639058669465, 1.1007649946070273), (282, 79, 3.051907528425515, 3.1394930703056216, 1.0286986224400096), (435, 525, 3.1469051054161548, 3.0832721091795516, 1.0206381383099967), (443, 332, 3.2015314287797194, 3.288181856526635, 1.0270653059869985)], [(77, 248, 3.148715476938977, 3.170240749098018, 1.0068362074365533), (113, 169, 3.1317016777688336, 3.2704644466536408, 1.044309063621816), (162, 355, 3.196567400501707, 3.340544281800449, 1.0450410904134682), (230, 211, 2.976468446276127, 3.136460660056166, 1.053752363469603), (387, 479, 3.11968957298559, 2.9636227254717786, 1.0526608350558408), (533, 351, 3.174059134900897, 3.087519286360008, 1.0280289256566602)], [(100, 157, 3.742351340702459, 3.867441398101282, 1.0334255247598807), (113, 81, 3.4181004362623217, 3.007910635776715, 1.136370341461187), (162, 199, 3.0568194301595466, 3.1436349941357413, 1.0284006189962172), (177, 480, 3.342922791689961, 3.0629140493906934, 1.0914190662173395), (325, 394, 3.484062457121131, 3.1516737559072543, 1.1054641840992816), (365, 534, 2.8974169572899404, 3.7156762935854437, 1.282409935593409), (535, 145, 3.0621738495350717, 3.067297402623201, 1.0016731751167254)], [(21, 157, 2.8353344757561993, 3.508200462951603, 1.237314501322087), (51, 167, 3.0430856446771584, 3.298687795357158, 1.0839943992792607), (118, 316, 3.346309910822877, 3.172951957017077, 1.0546361735551695), (172, 159, 3.5032056228928434, 3.3423422558139357, 1.0481289331752572), (224, 205, 3.284753526381704, 2.9343711035554043, 1.1194063090390176), (446, 489, 3.019743956840392, 2.773626144831022, 1.0887350346289602)], [(100, 325, 2.9811117561715075, 2.9908183859185455, 1.00325604356393), (157, 173, 3.0762338792636443, 3.2344023030080105, 1.0514162544046315), (428, 51, 3.630256445025419, 2.935853915562805, 1.236524891712637), (456, 68, 3.25867886704957, 2.999811588522652, 1.086294512467833)], [(43, 341, 3.2552456555829954, 3.1519412680291485, 1.0327748453315697), (85, 293, 2.958975925017555, 3.28975875278027, 1.1117896313268425), (156, 312, 3.0546269397946677, 3.1734830337780044, 1.0389101832485397), (268, 94, 3.2478332285572344, 2.9245739560747106, 1.110532090259189), (393, 490, 3.1792830587419942, 3.13402937258303, 1.0144394582114802)], [(37, 250, 3.190457899044174, 3.164685867059168, 1.0081436303846976), (191, 217, 2.9215310517796964, 3.454395092767328, 1.1823920511346369), (218, 314, 3.315143376395405, 3.089808114125526, 1.0729285618869744), (311, 115, 3.103837140880805, 3.2016711374251052, 1.0315203382470437), (335, 122, 3.5890192385511375, 2.9739545879481826, 1.206817095693215), (490, 393, 3.28780308559798, 3.13402937258303, 1.049065817429851)], [(109, 400, 3.104972040254367, 3.0038338182328124, 1.0336697128208827), (234, 318, 2.922832941983249, 3.134770912191493, 1.0725111473748605), (259, 76, 3.4065781596211426, 3.2747345415273905, 1.0402608566959624), (516, 454, 3.328337790845077, 3.0260897014876855, 1.0998807435248201)], [(101, 137, 3.1357760424842356, 3.4257515458891272, 1.092473282363356), (102, 313, 3.194631336885613, 2.9497667229417783, 1.0830115181785063), (294, 453, 3.4323016058156246, 2.9972587553559586, 1.1451469112175467), (434, 222, 2.980913406047323, 3.173408548510842, 1.0645758920983774), (450, 437, 3.202006176155486, 2.9994657912547114, 1.0675254858686185)], [(153, 382, 3.360420558372048, 2.987514325396019, 1.124821571500447), (290, 429, 3.2284297622478224, 3.071636816279431, 1.0510454052176355), (333, 163, 2.932606238867815, 3.401176617330192, 1.159779506792319), (337, 483, 3.0448360231728593, 3.2988395029546784, 1.0834210702476963), (454, 372, 3.0260897014876855, 3.3720772235472243, 1.114334853289195), (514, 234, 3.328337790845077, 2.922832941983249, 1.1387369230164341)], [(386, 206, 3.05617795148539, 3.2875798487998757, 1.075716107173019), (476, 301, 2.9842909547515726, 3.1462459279269455, 1.0542691633057792), (500, 242, 3.5945960477628978, 2.992581283359904, 1.2011690602191714), (526, 347, 2.8482080889984753, 3.014494102273366, 1.0583826771355607), (528, 499, 3.347703915308649, 3.2768631840781297, 1.021618458645062)], [(9, 470, 3.160926302000579, 3.234958914758185, 1.0234211764794232), (14, 158, 2.9764558343472927, 3.4717993803038536, 1.1664205933246057), (19, 377, 3.3604554697014994, 3.0503922175727975, 1.1016470112736583), (373, 440, 3.247165256651446, 3.1885275761174894, 1.0183902064931667), (389, 3, 2.9808921548419507, 3.1549015481471754, 1.0583749375241824), (451, 281, 3.1322817464501607, 3.1001421615061404, 1.0103671326247845)], [(36, 56, 2.87149701265598, 3.4739260348378753, 1.2097961514592281), (135, 59, 3.316883501369983, 2.9340395938380768, 1.1304835518702392), (165, 230, 3.3518727086120816, 2.8832139217596824, 1.1625473515216544), (303, 256, 3.191227033373396, 3.181676493879205, 1.0030017317953488), (404, 108, 2.9272621178104443, 3.687105499657023, 1.2595747668865855), (479, 212, 3.1714219869932405, 3.0313304458151533, 1.0462145396822335)], [(64, 219, 3.0854885892292208, 3.283238194919659, 1.064090208072958), (122, 304, 3.385914804066517, 2.9194488753109717, 1.1597787625946554), (173, 157, 2.9245051705190597, 3.2344023030080105, 1.1059656640764113), (215, 37, 3.496684518530334, 2.876625767784697, 1.2155507183762548), (249, 353, 2.9421384818614675, 3.4159211306402733, 1.1610334291535616)], [(77, 486, 3.2656070472777854, 3.2614726317137017, 1.0012676529994093), (116, 458, 3.3030504226293638, 2.936113536996145, 1.124973670469372), (185, 251, 3.107583792498345, 2.906445751749418, 1.069204126940219), (416, 323, 3.018016802491453, 3.4144909846861813, 1.1313691102936962), (533, 4, 3.435732804060332, 2.8972093637887455, 1.185876604915893)], [(100, 41, 3.2229277128491907, 2.9973258262687343, 1.075267721848345), (177, 142, 3.249469307542079, 2.9885618757958388, 1.0873020009588263), (328, 360, 3.3462948556511924, 3.0098183851171116, 1.1117929481054016), (456, 150, 2.997921556710785, 3.163713408874697, 1.0553022649284438), (501, 366, 3.1576436462917497, 3.5132692009249924, 1.112623713904791), (527, 24, 3.6159434384678164, 2.9844632888820466, 1.2115891832002788)], [(54, 32, 3.0235403483911765, 3.3469163890282965, 1.106952778324651), (79, 338, 3.065079966624696, 2.8771110934075033, 1.0653325044166342), (89, 277, 3.077438120403443, 3.13762354975632, 1.019556990911969), (164, 124, 3.084328565224637, 3.41114167505157, 1.1059592397229348), (412, 375, 3.2398145652119354, 3.197342863426018, 1.013283436778628), (485, 168, 3.146700520573147, 3.2001342499988508, 1.0169808753887934)], [(66, 309, 3.2853184474249453, 3.1090535846118237, 1.0566940575374897), (120, 67, 2.938311667693787, 3.1466162555198802, 1.0708926116029027), (182, 151, 3.164213033402539, 3.3427462521398987, 1.0564226292138679), (235, 427, 3.019026654169993, 3.144064388841977, 1.0414165719601307), (236, 244, 3.15151015189195, 2.9833897587558273, 1.0563521385842105), (280, 209, 3.560537386519781, 3.0198675452235997, 1.1790376012190789)], [(61, 15, 3.305610723129084, 2.9180032632757515, 1.1328331139075576), (158, 126, 3.399659710101203, 3.0247302443652977, 1.1239546787467587), (298, 434, 3.1922332835661495, 2.8679327352323347, 1.113078157081513), (313, 451, 3.0476161020408923, 3.112557037246623, 1.0213087649596817), (435, 507, 3.0832721091795516, 3.1469051054161548, 1.0206381383099967), (492, 220, 3.038244680856133, 3.82950071355407, 1.2604319651026172)], [(25, 416, 3.3131349517058823, 3.21544374146933, 1.03038187512866), (237, 462, 3.273137615172774, 3.29331922587049, 1.0061658301820748), (347, 146, 3.014494102273366, 3.4965771761984925, 1.1599217173991372), (506, 94, 2.785726285015436, 3.6797890631069325, 1.320944230199753)], [(24, 500, 2.9844632888820466, 3.3810140938077438, 1.1328717315448171), (68, 264, 3.0131714734575383, 3.465128170219541, 1.149993686301362), (301, 485, 3.5445050311947877, 3.0018180010786932, 1.1807861202514882), (375, 381, 3.268209964321089, 2.8698072605774385, 1.1388255961355005), (480, 58, 2.8548439916389876, 3.468568995112218, 1.2149767221153427), (522, 100, 3.6159434384678164, 3.2229277128491907, 1.1219436985979387)], [(106, 412, 3.154855562587185, 3.148663117975456, 1.0019666901093283), (203, 538, 2.8657481118419277, 3.4371961929223307, 1.1994062488322155), (338, 0, 3.5687633935730907, 3.102057576273999, 1.150450404553635), (368, 116, 3.473946735876745, 2.9535099091215455, 1.176209609166333), (499, 97, 3.2768631840781297, 2.9363395436192707, 1.1159687547711654)], [(72, 11, 3.4281753326837094, 3.0741102004064413, 1.115176460567502), (99, 276, 3.0861366116095033, 2.882490557592832, 1.070649339502689), (281, 164, 3.00833323643055, 3.194291038364821, 1.0618142297809114), (338, 386, 2.98246172462166, 3.5691269250207465, 1.1967050224168452), (436, 373, 3.1392209953438908, 2.886802082855901, 1.0874389394364967), (499, 423, 2.8548652815472364, 3.261985628310783, 1.1426057997885286)], [(38, 431, 3.008249822504741, 3.4230038077954377, 1.137872188070259), (98, 159, 3.1779215889423913, 3.21574604530358, 1.011902262312827), (289, 495, 3.4331798474336974, 3.024746876869135, 1.135030462776219), (307, 296, 2.83193656894518, 3.2052241607367087, 1.131813542677818), (327, 437, 2.9725959444296794, 3.346776810378991, 1.125876800259546), (426, 425, 3.2911670962045645, 2.885298605352015, 1.1406677596903467)], [(12, 83, 2.901443117861705, 3.974590146027631, 1.3698666437951093), (178, 232, 3.185339147081731, 3.3193961011549935, 1.042085614084792), (194, 280, 2.88767197657496, 3.3808066626868283, 1.1707724042454333), (244, 56, 3.708078687206466, 3.0145031775174886, 1.2300795417506085), (336, 384, 3.1407232747039346, 3.3191986960219175, 1.0568262166729119), (406, 108, 3.160786186448746, 3.1706741115832973, 1.0031283119297798)], [(40, 148, 3.2513107618704047, 2.964149625378467, 1.0968780840323715), (111, 537, 2.913754861413271, 3.1854773249538106, 1.0932550871519593), (170, 461, 2.9135546556875407, 3.0112431873090935, 1.0335289854373781), (399, 385, 3.0041987741071012, 2.91040429550998, 1.032227302145555), (468, 311, 3.148261551982456, 3.9007400491518682, 1.2390139715982549)], [(78, 41, 3.097957167333639, 3.007363979491309, 1.0301237856342396), (351, 307, 3.087519286360008, 3.3987988580447452, 1.1008186646994897), (431, 212, 2.964039885966948, 3.3126500761638735, 1.117613191322896), (508, 230, 3.174059134900897, 2.976468446276127, 1.0663842712231593), (521, 39, 3.435732804060332, 3.0251419395034724, 1.1357261486461858)], [(44, 167, 3.1740998023237177, 3.2660908617788658, 1.0289817791450044), (155, 344, 2.8472809837667254, 3.3182828120155867, 1.165421618355967), (365, 509, 3.7156762935854437, 2.8974169572899404, 1.282409935593409), (511, 157, 3.1837615376182113, 3.0762338792636443, 1.0349543183564138)], [(57, 155, 3.345474459818699, 3.0088106883822476, 1.1118926400841342), (142, 58, 3.089094945862324, 3.276800903645684, 1.0607640623137151), (145, 349, 3.067297402623201, 3.1371255653115204, 1.0227653707881739), (161, 354, 3.0751574581586096, 3.023606603281613, 1.0170494583591154), (239, 26, 3.425529336320636, 2.9063389361475354, 1.1786406924931156), (509, 325, 3.0621738495350717, 3.484062457121131, 1.1377742180281876)], [(5, 234, 2.9237407698223867, 3.1747895988805253, 1.0858656251776349), (232, 178, 3.230791796494345, 3.3193961011549935, 1.027424950365663), (321, 87, 2.821342850488343, 3.291857348924555, 1.1667696991717158), (400, 483, 3.0439726284496915, 2.97744842648178, 1.0223426882481783), (401, 438, 3.101343144545744, 3.1789008328636488, 1.0250077739557146), (474, 502, 3.391980989157446, 2.8864383357192165, 1.1751441030914187)], [(111, 532, 3.1854773249538106, 2.913754861413271, 1.0932550871519593), (189, 414, 2.999315348624523, 3.5818481840801604, 1.1942219365905573), (256, 140, 3.219095949691826, 2.9498032339900373, 1.0912917555309378), (374, 165, 3.2565045223019746, 2.9362923747194083, 1.1090532231529449), (392, 127, 3.0956718264202094, 3.6171605270102027, 1.168457359122926), (448, 251, 2.943513623693817, 3.3976232712963927, 1.1542746885719228)], [(121, 305, 3.1944587443144403, 3.1800179260025065, 1.0045411122352026), (128, 341, 2.878920604129509, 3.3709084045383895, 1.1708931464463368), (203, 528, 3.4371961929223307, 2.8657481118419277, 1.1994062488322155), (242, 94, 2.905957443141829, 3.4075559351677263, 1.172610405293336), (345, 8, 3.5269870427718897, 3.1725789213690248, 1.11170978884583), (357, 471, 2.9943475132211, 3.2090211368760597, 1.0716929557130894)], [(147, 159, 3.584041604758573, 2.9231171875187254, 1.2261026071968295), (212, 431, 2.833145830699028, 3.3126500761638735, 1.1692479929091883), (356, 404, 3.016537387801347, 3.147884651538419, 1.0435423954194072), (379, 498, 3.769566238402216, 2.9114306965063483, 1.2947470269258379), (430, 167, 3.4811857299456883, 2.9203111406261097, 1.1920598738664976)]]
    """
    ...


def altbc(atoms: Atoms, r_min: float, r_max: float, n_r: int,
          cutoff_angle: float, filter: fastatomstruct.Filter = None) -> np.ndarray:
    """Angular-limited three-body correlation.

    The Angular-limited three-body correlation (ALTBC) is a higher-order
    correlation function that includes three atoms. It is defined as

    .. math::

        g^{(3)}(r, r') = \frac{V}{\rho (N - 1)(N - 2)} \sum_{i_1} \sum_{i_2} \sum_{i_3} \langle\delta(r - r_{i_1 i_2}) \delta(r' - r_{i_2 i_3}) \Theta(\beta - \delta)\rangle,

    with :math:`i_1`, :math:`i_2` and :math:`i_3` distinct indices of atoms, :math:`V` the unit cell
    volume, :math:`N` the number of particles, :math:`\rho = V/N`
    the atomic density, and

    .. math::

        \cos\beta = \frac{\vec{r}_{i_1 i_2} \cdot \vec{r}_{i_2 i_3}}{r_{i_1 i_2} r_{i_2 i_3}}

    the alignment angle and :math:`\delta` an angular threshold (typically
    25° in phase-change materials).

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        cell (np.ndarray): unit cell array, shape is 3x3
        r_min (float): minimum radius / atomic distance
        r_max (float): maximum radius / atomic distance
        n_r (int): number of bins for each :math:`r`
        cutoff_angle (float): maximum bond angle in degrees
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        2D array of shape `(n_r, n_r)` containing the ALTBC

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.altbc(atoms, 2.5, 4.5, 100, 25)
    array([[ 0.,  0.,  0., ...,  0.,  0.,  0.],
           [ 0.,  0.,  0., ...,  0.,  0.,  0.],
           [ 0.,  0.,  0., ...,  0.,  0.,  0.],
           ...,
           [ 0.,  0.,  0., ..., 12., 12.,  5.],
           [ 0.,  0.,  0., ..., 12.,  4., 10.],
           [ 0.,  0.,  0., ...,  5., 10.,  6.]])
    """
    ...


def tbc(atoms: Atoms, r_min: float, r_max: float, n_r: int, cutoff_angle: float,
        filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Three-body correlation.

    The three-body correlation (TBC) is a higher-order correlation function
    that includes three atoms. It is defined as

    .. math::

        g^{(3)}(r, r') = \frac{V}{\rho (N - 1)(N - 2)} \sum_{i_1} \sum_{i_2} \sum_{i_3} \langle\delta(r - r_{i_1 i_2}) \delta(r' - r_{i_2 i_3})\rangle,

    with :math:`i_1`, :math:`i_2` and :math:`i_3` distinct indices of atoms, :math:`V` the unit cell
    volume, :math:`N` the number of particles, and :math:`\rho = V/N` the atomic density.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_min (float): minimum radius / atomic distance
        r_max (float): maximum radius / atomic distance
        n_r (int): number of bins for each :math:`r`
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        2D array of shape (n_r, n_r) containing the TBC

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.tbc(atoms, 2.5, 4.5, 100)
    array([[  0.,   0.,   0., ...,   0.,   0.,   0.],
           [  0.,   0.,   0., ...,   0.,   0.,   0.],
           [  0.,   0.,   0., ...,   0.,   0.,   0.],
           ...,
           [  0.,   0.,   0., ..., 100.,  91.,  83.],
           [  0.,   0.,   0., ...,  91.,  74.,  80.],
           [  0.,   0.,   0., ...,  83.,  80.,  96.]])
    """
    ...


def distances(atoms: Atoms, i: int, filter: Optional[fastatomstruct.Filter] = None]) -> np.ndarray:
    """Distances of one atom to all other atoms.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        i (int): Index of atom to calculate distances for
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        Distances (array of floats) for this atom

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.distances(atoms, 0)
    array([ 0.        , 13.45305427,  2.91828225, 11.63855946, 14.23152337,
           12.98086183, 10.09229625,  7.97052848, 12.48575892, 11.68376213,
           10.90332209,  7.74104435, 16.15190625, 16.24227574,  9.87187816,
           11.66458514, 14.68663104, 11.71660523, 16.71265741, 11.72043018,
           12.65826558, 13.49519165,  9.41687702, 19.99444408, 10.37973443,
            7.99728229, 13.61816274, 13.50774181, 15.42548069, 12.78331659,
           14.69245551, 14.52264404, 12.80632247, 14.6919333 , 12.7525009 ,
            8.81892663, 13.08456575, 15.78227991, 10.55052004,  7.38462546,
           13.89727903, 12.97814224,  9.07994266,  7.3494056 , 16.6598059 ,
           15.77506224, 18.8148011 , 14.94695883, 19.54346046, 13.09210181,
           18.52370806, 13.21267724, 17.53237126, 13.69773798,  9.92318978,
           15.24722512, 12.68169007, 17.47706352, 13.38001098, 17.3142893 ,
           16.07975324,  9.50590849, 12.25494589, 14.84074353, 11.61823972,
           13.01514489, 14.26862774, 11.93514476, 12.9590854 , 11.37955313,
           10.93736885, 16.57102791,  5.76662189,  4.87728443,  6.60290485,
           14.95309546,  8.13378436, 11.65879917, 12.06845717,  4.60399339,
           16.01095523, 16.53982762, 10.06996221, 14.30801786, 17.24454067,
           10.02425226, 15.83478322,  6.43750202,  9.41839457,  6.45394489,
           12.61765436, 10.34916904, 18.71061698, 12.20284299, 13.0766057 ,
           15.68238841, 15.78768235,  8.78146163, 11.14209921,  3.47372848,
           15.44425666, 13.18397683,  8.86465355, 15.85267195, 17.08010766,
           18.56117449,  7.00701062, 14.20679296, 16.71142292, 12.98804769,
           12.0010191 , 11.06064295, 10.75372142, 16.04795803,  9.83823783,
           14.45299861,  9.77181582, 14.27364656, 10.04037206, 14.36592624,
           13.24780974, 12.56878422, 12.9189609 , 17.70972824,  9.74302193,
           12.43186957, 13.17125831,  5.9644202 , 10.61360209, 16.05632274,
            4.38268784,  6.64795492, 15.3985803 , 12.27211505, 12.23781216,
           15.19248623,  6.97400373, 15.83467251, 10.95976576, 13.82044358,
           10.46063164, 14.13769976, 16.12717333,  6.69155193, 14.47967226,
           14.12287337, 14.14127043, 13.73241727, 15.14378575, 12.56603332,
           16.41958784, 18.96442397, 15.20458993, 11.84955346, 15.45205798,
           19.22840534,  5.57137442, 15.48363621, 10.50039429, 13.37120077,
           12.9950963 , 15.33644961, 14.00486711,  9.55342674,  7.76769412,
           13.4878439 , 11.54713584, 15.3088453 ,  9.68247744, 15.84401275,
           12.37422573, 10.22809455, 14.26282979, 14.42670532,  8.05950516,
           11.99431538, 11.65799056, 16.19542765, 13.61548693, 14.15165814,
           17.28402126,  5.34348235, 16.54436515, 15.77318309,  9.63491383,
            6.82168988, 13.0779276 , 13.90074491, 11.41176921,  8.76332105,
           12.02279105, 16.55599001, 11.01432106, 14.10949256, 16.18903138,
           13.83397709, 10.89930383, 14.03046053,  7.23702111, 12.34905376,
           13.55005226, 13.19399613, 11.89725048,  9.46956465,  7.3034463 ,
           11.32760465,  2.86753452, 15.52769025, 11.61745584, 18.35310755,
           17.04176701, 11.2150203 , 13.2986454 , 13.82744104,  8.69116788,
           15.2131513 , 20.58778579, 16.85423355, 15.31951296,  9.23490219,
            9.34737321, 18.3200486 , 13.2038814 , 13.42948245, 11.14406463,
           11.72218223, 14.30153769, 10.4520376 , 15.9071007 , 14.88436765,
           13.93347168, 11.0300734 , 12.97898727, 15.52743615, 13.72518434,
           13.32411846, 13.91096185,  9.46029323, 10.58043842, 14.67039595,
           11.33730806, 11.30891943, 12.87706341,  9.70928192, 13.28087182,
            9.25479839, 15.54783952,  8.61200567, 10.4780869 , 15.80128269,
           17.28461507,  6.18083802,  7.43525484, 14.40223132, 10.80872322,
           15.47823572,  8.95531222, 13.65618381,  6.84293173, 10.79021807,
           13.22834315, 17.33153406, 15.39201146, 11.34007454, 15.16520504,
           17.00098423, 10.4366974 , 16.73077809, 10.55015716,  7.22746613,
            9.78536784, 11.8135512 , 16.22434369, 13.19639842, 15.19389069,
           18.1126051 ,  4.17592756,  7.63884459, 16.49653812, 13.48689902,
           17.26272358,  5.5249288 ,  3.44397171, 17.60236508, 18.57300126,
           13.57520375, 16.55796129, 13.73036349,  9.82404664,  6.89819332,
           15.37062647, 13.65269438, 11.12700634, 11.84702993,  9.9336438 ,
           12.81712995,  9.80413585,  9.05876343,  9.59115285, 16.63543699,
            9.8931421 ,  9.47635956, 10.78592697, 11.24607161, 12.71695417,
           13.79395519, 13.99265941,  9.29879488, 13.68382268, 11.56198213,
           12.22030669, 14.33775085,  4.59554466,  7.22303138, 13.79472222,
            8.13255815,  9.89700931,  7.06525584, 14.36192485,  9.28105027,
           10.26096821,  9.71585209, 18.3987585 , 11.17951846, 13.35381558,
           14.49531169, 10.72890458, 12.91393083, 11.82131475, 12.73821845,
            6.91370051, 13.42814544,  4.04684835, 12.04687045, 16.14922856,
           13.69096176, 15.78890843, 13.31954903,  3.10205758, 10.5318126 ,
           11.23001924,  7.39202829,  7.54312267, 18.10321074, 17.39011203,
           12.21433378, 13.08225024, 11.01887903, 14.76007905, 12.36980948,
            7.87089367,  9.86418651, 11.24515449, 16.17805612, 15.27997295,
           15.02977129, 16.71976536, 12.94306732,  9.99203251, 20.34685389,
           10.48159708, 13.25482269, 18.44021135,  9.59041856,  5.98710024,
           17.10962816, 13.69695405,  7.34353324,  8.22716915, 12.27095652,
           12.01303339,  4.13135534, 16.63964055,  9.45073445, 10.86616122,
           11.41424525,  5.36565446, 13.77216842, 14.17634097, 13.00724394,
            7.33745139, 11.68663778,  9.74553484, 10.38228877, 13.72793066,
           15.47584264,  4.63098232, 14.10755248, 12.09380609, 13.3478497 ,
            9.52825233, 16.52457824,  6.67705599,  9.98144777, 13.41689489,
           10.77728382, 13.52304351, 17.30058891, 13.40607422, 14.54679752,
           12.40746108, 13.03618451, 15.63029831, 14.61102195, 14.91205602,
            7.83490691, 15.65940363, 13.39435922, 12.15381242, 18.54443755,
           13.69602952,  9.66399217,  8.78247606, 17.01606846, 10.24575702,
            9.45659483,  9.47314765, 12.99511121, 14.06910108,  7.27698167,
            7.07690237, 12.68106448, 14.95951706,  9.86968325,  7.56505568,
           11.16134479, 10.23313075, 11.85848785, 14.74191108, 17.05510993,
           16.9862184 , 12.2564015 , 10.93968714, 15.30167956, 11.61724376,
            5.69614198,  6.92692222, 12.8101883 , 13.8865047 , 14.02129903,
            9.99974618,  4.52931737, 11.63853582,  3.03523776, 13.33358058,
           14.61811649,  9.17010358, 15.43046322,  6.67194401, 14.58522259,
           11.61075149,  7.79561779,  9.06501566,  9.55569018, 15.58967849,
           11.71028543, 14.76017023, 15.16642806, 11.07121076,  4.11604658,
           17.44311726, 12.35099306, 11.07666243, 15.49798333, 21.6357812 ,
           11.65830427, 15.98344232, 17.88262168, 13.54676507, 16.6484437 ,
           13.96117514, 14.12836155, 15.12835964, 18.82867792, 10.39025482,
           11.62023945,  8.01949447, 12.01556067, 14.05326401, 15.22040944,
           14.4366836 , 16.25801028, 10.70484594, 12.47194826,  9.73241794,
            7.83197886, 12.79703477, 14.86020936, 20.60785388,  7.60261554,
           12.35394925,  9.30835578,  7.14509754,  9.48693599, 13.48572795,
            4.43154687, 15.53838671, 16.13946126, 11.83453271,  7.02849507,
            9.9992658 , 13.5412093 ,  7.8772154 ,  5.27826641, 17.07009366,
           13.06332155, 11.04734786,  4.62950826, 13.80205684, 16.32485245,
           11.58476043, 17.03575044,  7.99200628, 15.13789789, 13.68418964,
           11.44898685, 14.36557244,  6.7666514 , 10.58012836, 13.90014809,
           14.29593232,  9.02228555, 14.041183  ,  6.99391147, 14.77841219,
            7.55808524,  8.77282081, 11.9569199 ,  6.61068791,  4.47887047,
            9.98397752, 15.87118545, 13.22975652, 11.8398227 , 19.08623473,
           16.32066344, 12.52116397,  8.4370716 , 12.90248561, 15.19129564])
    """
    ...


def all_distances(atoms: Atoms, filter: Optional[fastatomstruct.Filter] = None) -> List[np.ndarray]:
    """Distances of each atom to all other atoms.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        List containing one array of floats for each atom

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> import numpy as np
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> d = np.array(fs.all_distances(atoms))
    >>> d
    array([[ 0.        , 13.45305427,  2.91828225, ...,  8.4370716 ,
            12.90248561, 15.19129564],
           [13.45305427,  0.        , 13.41206116, ..., 11.12712855,
             7.16928015, 13.45660093],
           [ 2.91828225, 13.41206116,  0.        , ...,  6.8906441 ,
            12.97953841, 13.27298255],
           ...,
           [ 8.4370716 , 11.12712855,  6.8906441 , ...,  0.        ,
            12.12511465, 13.76358893],
           [12.90248561,  7.16928015, 12.97953841, ..., 12.12511465,
             0.        , 13.02275278],
           [15.19129564, 13.45660093, 13.27298255, ..., 13.76358893,
            13.02275278,  0.        ]])
    >>> d.shape
    (540, 540)
    """
    ...


def distance_vectors(atoms: Atoms, i: int, filter: Optional[fastatomstruct.Filter] = None]) -> np.ndarray:
    """Distance vectors between one atom and all other atoms.
    
    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        i (int): Atom index
        filter (fastatomstruct.Filter): Filter instance
    
    Returns:
        Distance vectors (2D array of floats) for this atom
    
    Examples
    --------
    
    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.
    
    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.distance_vectors(atoms, 0)
    array([[  0.        ,   0.        ,   0.        ],
           [-12.51621245,  -1.81873614,   4.58489846],
           [  0.70907609,  -2.7956268 ,  -0.4450317 ],
           ...,
           [ -4.8092365 ,  -6.69770233,  -1.78779332],
           [ 12.63852603,   2.2464042 ,  -1.30209939],
           [  9.35835886,  -7.42100408,  -9.38750666]])
    """
    ...


def all_distance_vectors(atoms: Atoms, i: int, filter: Optional[fastatomstruct.Filter] = None]) -> List[np.ndarray]:
    """Distance vectors between all atoms and all other atoms.
    
    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        filter (fastatomstruct.Filter): Filter instance
    
    Returns:
        Distance vectors (list of 2D arrays of floats) for all atoms
    
    Examples
    --------
    
    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.
    
    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.all_distance_vectors(atoms)
    [..., array([[-12.63852603,  -2.2464042 ,   1.30209939],
        [  0.46526151,  -4.06514034,   5.88699785],
        [-11.92944994,  -5.04203099,   0.85706769],
        ...,
        [  8.17223747,  -8.94410653,  -0.48569392],
        [  0.        ,   0.        ,   0.        ],
        [ -3.28016718,  -9.66740827,  -8.08540726]]), array([[ -9.35835886,   7.42100408,   9.38750666],
        [  3.74542869,   5.60226793, -11.64759489],
        [ -8.64928276,   4.62537728,   8.94247496],
        ...,
        [ 11.45240464,   0.72330175,   7.59971334],
        [  3.28016718,   9.66740827,   8.08540726],
        [  0.        ,   0.        ,   0.        ]])]
    """


def r_theta_phi(atoms: Atoms, filter: Optional[fastatomstruct.Filter] = None) -> List[List]:
    """List of distances and angles of an atom to all other atoms.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        List of three lists (r, theta, and phi)

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> rtp = fs.r_theta_phi(atoms)
    >>> len(rtp)
    540
    >>> rtp[0]
    (array([ 0.        , 13.45305427,  2.91828225, 11.63855946, 14.23152337,
            12.98086183, 10.09229625,  7.97052848, 12.48575892, 11.68376213,
            10.90332209,  7.74104435, 16.15190625, 16.24227574,  9.87187816,
            11.66458514, 14.68663104, 11.71660523, 16.71265741, 11.72043018,
            12.65826558, 13.49519165,  9.41687702, 19.99444408, 10.37973443,
             7.99728229, 13.61816274, 13.50774181, 15.42548069, 12.78331659,
            14.69245551, 14.52264404, 12.80632247, 14.6919333 , 12.7525009 ,
             8.81892663, 13.08456575, 15.78227991, 10.55052004,  7.38462546,
            13.89727903, 12.97814224,  9.07994266,  7.3494056 , 16.6598059 ,
            15.77506224, 18.8148011 , 14.94695883, 19.54346046, 13.09210181,
            18.52370806, 13.21267724, 17.53237126, 13.69773798,  9.92318978,
            15.24722512, 12.68169007, 17.47706352, 13.38001098, 17.3142893 ,
            16.07975324,  9.50590849, 12.25494589, 14.84074353, 11.61823972,
            13.01514489, 14.26862774, 11.93514476, 12.9590854 , 11.37955313,
            10.93736885, 16.57102791,  5.76662189,  4.87728443,  6.60290485,
            14.95309546,  8.13378436, 11.65879917, 12.06845717,  4.60399339,
            16.01095523, 16.53982762, 10.06996221, 14.30801786, 17.24454067,
            10.02425226, 15.83478322,  6.43750202,  9.41839457,  6.45394489,
            12.61765436, 10.34916904, 18.71061698, 12.20284299, 13.0766057 ,
            15.68238841, 15.78768235,  8.78146163, 11.14209921,  3.47372848,
            15.44425666, 13.18397683,  8.86465355, 15.85267195, 17.08010766,
            18.56117449,  7.00701062, 14.20679296, 16.71142292, 12.98804769,
            12.0010191 , 11.06064295, 10.75372142, 16.04795803,  9.83823783,
            14.45299861,  9.77181582, 14.27364656, 10.04037206, 14.36592624,
            13.24780974, 12.56878422, 12.9189609 , 17.70972824,  9.74302193,
            12.43186957, 13.17125831,  5.9644202 , 10.61360209, 16.05632274,
             4.38268784,  6.64795492, 15.3985803 , 12.27211505, 12.23781216,
            15.19248623,  6.97400373, 15.83467251, 10.95976576, 13.82044358,
            10.46063164, 14.13769976, 16.12717333,  6.69155193, 14.47967226,
            14.12287337, 14.14127043, 13.73241727, 15.14378575, 12.56603332,
            16.41958784, 18.96442397, 15.20458993, 11.84955346, 15.45205798,
            19.22840534,  5.57137442, 15.48363621, 10.50039429, 13.37120077,
            12.9950963 , 15.33644961, 14.00486711,  9.55342674,  7.76769412,
            13.4878439 , 11.54713584, 15.3088453 ,  9.68247744, 15.84401275,
            12.37422573, 10.22809455, 14.26282979, 14.42670532,  8.05950516,
            11.99431538, 11.65799056, 16.19542765, 13.61548693, 14.15165814,
            17.28402126,  5.34348235, 16.54436515, 15.77318309,  9.63491383,
             6.82168988, 13.0779276 , 13.90074491, 11.41176921,  8.76332105,
            12.02279105, 16.55599001, 11.01432106, 14.10949256, 16.18903138,
            13.83397709, 10.89930383, 14.03046053,  7.23702111, 12.34905376,
            13.55005226, 13.19399613, 11.89725048,  9.46956465,  7.3034463 ,
            11.32760465,  2.86753452, 15.52769025, 11.61745584, 18.35310755,
            17.04176701, 11.2150203 , 13.2986454 , 13.82744104,  8.69116788,
            15.2131513 , 20.58778579, 16.85423355, 15.31951296,  9.23490219,
             9.34737321, 18.3200486 , 13.2038814 , 13.42948245, 11.14406463,
            11.72218223, 14.30153769, 10.4520376 , 15.9071007 , 14.88436765,
            13.93347168, 11.0300734 , 12.97898727, 15.52743615, 13.72518434,
            13.32411846, 13.91096185,  9.46029323, 10.58043842, 14.67039595,
            11.33730806, 11.30891943, 12.87706341,  9.70928192, 13.28087182,
             9.25479839, 15.54783952,  8.61200567, 10.4780869 , 15.80128269,
            17.28461507,  6.18083802,  7.43525484, 14.40223132, 10.80872322,
            15.47823572,  8.95531222, 13.65618381,  6.84293173, 10.79021807,
            13.22834315, 17.33153406, 15.39201146, 11.34007454, 15.16520504,
            17.00098423, 10.4366974 , 16.73077809, 10.55015716,  7.22746613,
             9.78536784, 11.8135512 , 16.22434369, 13.19639842, 15.19389069,
            18.1126051 ,  4.17592756,  7.63884459, 16.49653812, 13.48689902,
            17.26272358,  5.5249288 ,  3.44397171, 17.60236508, 18.57300126,
            13.57520375, 16.55796129, 13.73036349,  9.82404664,  6.89819332,
            15.37062647, 13.65269438, 11.12700634, 11.84702993,  9.9336438 ,
            12.81712995,  9.80413585,  9.05876343,  9.59115285, 16.63543699,
             9.8931421 ,  9.47635956, 10.78592697, 11.24607161, 12.71695417,
            13.79395519, 13.99265941,  9.29879488, 13.68382268, 11.56198213,
            12.22030669, 14.33775085,  4.59554466,  7.22303138, 13.79472222,
             8.13255815,  9.89700931,  7.06525584, 14.36192485,  9.28105027,
            10.26096821,  9.71585209, 18.3987585 , 11.17951846, 13.35381558,
            14.49531169, 10.72890458, 12.91393083, 11.82131475, 12.73821845,
             6.91370051, 13.42814544,  4.04684835, 12.04687045, 16.14922856,
            13.69096176, 15.78890843, 13.31954903,  3.10205758, 10.5318126 ,
            11.23001924,  7.39202829,  7.54312267, 18.10321074, 17.39011203,
            12.21433378, 13.08225024, 11.01887903, 14.76007905, 12.36980948,
             7.87089367,  9.86418651, 11.24515449, 16.17805612, 15.27997295,
            15.02977129, 16.71976536, 12.94306732,  9.99203251, 20.34685389,
            10.48159708, 13.25482269, 18.44021135,  9.59041856,  5.98710024,
            17.10962816, 13.69695405,  7.34353324,  8.22716915, 12.27095652,
            12.01303339,  4.13135534, 16.63964055,  9.45073445, 10.86616122,
            11.41424525,  5.36565446, 13.77216842, 14.17634097, 13.00724394,
             7.33745139, 11.68663778,  9.74553484, 10.38228877, 13.72793066,
            15.47584264,  4.63098232, 14.10755248, 12.09380609, 13.3478497 ,
             9.52825233, 16.52457824,  6.67705599,  9.98144777, 13.41689489,
            10.77728382, 13.52304351, 17.30058891, 13.40607422, 14.54679752,
            12.40746108, 13.03618451, 15.63029831, 14.61102195, 14.91205602,
             7.83490691, 15.65940363, 13.39435922, 12.15381242, 18.54443755,
            13.69602952,  9.66399217,  8.78247606, 17.01606846, 10.24575702,
             9.45659483,  9.47314765, 12.99511121, 14.06910108,  7.27698167,
             7.07690237, 12.68106448, 14.95951706,  9.86968325,  7.56505568,
            11.16134479, 10.23313075, 11.85848785, 14.74191108, 17.05510993,
            16.9862184 , 12.2564015 , 10.93968714, 15.30167956, 11.61724376,
             5.69614198,  6.92692222, 12.8101883 , 13.8865047 , 14.02129903,
             9.99974618,  4.52931737, 11.63853582,  3.03523776, 13.33358058,
            14.61811649,  9.17010358, 15.43046322,  6.67194401, 14.58522259,
            11.61075149,  7.79561779,  9.06501566,  9.55569018, 15.58967849,
            11.71028543, 14.76017023, 15.16642806, 11.07121076,  4.11604658,
            17.44311726, 12.35099306, 11.07666243, 15.49798333, 21.6357812 ,
            11.65830427, 15.98344232, 17.88262168, 13.54676507, 16.6484437 ,
            13.96117514, 14.12836155, 15.12835964, 18.82867792, 10.39025482,
            11.62023945,  8.01949447, 12.01556067, 14.05326401, 15.22040944,
            14.4366836 , 16.25801028, 10.70484594, 12.47194826,  9.73241794,
             7.83197886, 12.79703477, 14.86020936, 20.60785388,  7.60261554,
            12.35394925,  9.30835578,  7.14509754,  9.48693599, 13.48572795,
             4.43154687, 15.53838671, 16.13946126, 11.83453271,  7.02849507,
             9.9992658 , 13.5412093 ,  7.8772154 ,  5.27826641, 17.07009366,
            13.06332155, 11.04734786,  4.62950826, 13.80205684, 16.32485245,
            11.58476043, 17.03575044,  7.99200628, 15.13789789, 13.68418964,
            11.44898685, 14.36557244,  6.7666514 , 10.58012836, 13.90014809,
            14.29593232,  9.02228555, 14.041183  ,  6.99391147, 14.77841219,
             7.55808524,  8.77282081, 11.9569199 ,  6.61068791,  4.47887047,
             9.98397752, 15.87118545, 13.22975652, 11.8398227 , 19.08623473,
            16.32066344, 12.52116397,  8.4370716 , 12.90248561, 15.19129564]), array([0.        , 1.22302091, 1.7238915 , 0.47646215, 1.62384316,
            2.95326345, 1.10723667, 1.97824638, 1.79039151, 2.39433773,
            1.2131014 , 2.01979774, 0.77866993, 1.57821039, 2.66511577,
            1.97935002, 0.75387289, 1.18201577, 1.97027174, 2.49732039,
            1.69102013, 0.90792801, 1.76572344, 2.25929041, 1.31648984,
            1.10970093, 1.28627299, 1.40759739, 2.0220961 , 2.30083338,
            2.26100634, 2.41869043, 1.47723175, 2.44552962, 0.95726838,
            2.46909512, 2.01322501, 0.9848478 , 1.97446629, 1.51123056,
            2.01738023, 1.1315978 , 0.15840277, 1.37127406, 0.9727314 ,
            2.38549664, 2.10231654, 1.82194757, 0.87050439, 1.00498966,
            0.82312658, 0.79746472, 1.02458   , 2.5442081 , 1.45667487,
            0.97644397, 1.95252473, 1.19202475, 1.39691644, 1.97138786,
            2.4206387 , 2.11263437, 1.23473594, 1.02928194, 0.87030377,
            1.02659501, 2.08872068, 1.67049213, 1.09728166, 1.54595987,
            2.1562678 , 2.16282083, 2.18634168, 1.15958952, 1.23419981,
            0.71061967, 2.43593949, 1.5317282 , 1.3572686 , 1.51646963,
            2.28299442, 1.76999577, 0.5594528 , 1.00609091, 0.93199362,
            0.71603707, 0.74104541, 2.95048048, 1.3623519 , 1.40197129,
            2.62120414, 1.9403318 , 1.04912261, 1.13070678, 1.19519909,
            0.7870274 , 0.6380334 , 2.45801875, 2.28248212, 2.93587653,
            1.22999389, 2.22557633, 2.19917033, 2.24772066, 2.34285022,
            2.13029904, 1.60718437, 1.18233418, 2.38568283, 2.95961546,
            2.18742368, 1.78419664, 0.47331643, 1.59197429, 2.6579909 ,
            1.61082153, 1.88798121, 1.2023511 , 0.49729666, 2.22021477,
            1.8984104 , 1.92201191, 1.21511248, 2.3241126 , 2.20329608,
            2.30130918, 2.62659774, 1.66093286, 1.72307745, 1.15828406,
            2.18012325, 1.26145036, 2.48133935, 0.8349398 , 0.92599819,
            1.9790025 , 1.32220431, 2.37099303, 1.08925441, 0.65266139,
            2.41423862, 1.62965686, 1.39481077, 0.58485455, 2.50791136,
            1.31159722, 0.63233387, 2.40028931, 2.20125791, 1.46410673,
            0.94252098, 1.9372445 , 1.07546467, 2.51827937, 1.1937498 ,
            1.05846694, 0.83237417, 1.07195918, 2.50680572, 2.46794706,
            2.0199967 , 1.56230372, 1.34579508, 0.82875857, 1.89105444,
            1.75287859, 1.10716431, 0.85515772, 0.78320891, 1.55331883,
            1.75029388, 1.36241211, 0.5840545 , 1.02807456, 0.61534178,
            0.71542827, 2.722047  , 1.35778715, 2.50432546, 2.22848318,
            2.23182758, 0.91976347, 2.01743627, 1.74288568, 0.39544338,
            1.63979555, 2.07963519, 0.66750854, 2.34167012, 1.44621272,
            0.85474408, 1.20217374, 0.75335583, 2.22580185, 2.31158974,
            2.51757698, 1.14527508, 0.76528402, 0.98340388, 1.25519079,
            0.57437818, 0.9276409 , 1.07399901, 1.68738768, 1.32428387,
            0.80006258, 0.04782869, 2.00106456, 2.31464642, 0.86119476,
            2.37432318, 1.4649733 , 2.08993288, 0.89274159, 1.00690789,
            0.97314198, 2.08723753, 1.18667011, 1.1280919 , 0.65933608,
            2.51583271, 0.90189469, 1.90567992, 0.97459249, 0.72467736,
            0.46590933, 1.82097291, 1.39166894, 0.89390057, 2.07480425,
            1.54478666, 1.14974382, 2.74020996, 2.32631452, 0.36808817,
            2.18533038, 2.14227248, 0.89798061, 2.93138663, 1.32406887,
            2.40343412, 2.26136956, 1.45359315, 1.45525469, 2.17635922,
            1.72734781, 0.80535713, 1.02119857, 1.46800377, 0.8215235 ,
            0.81739778, 1.70710033, 2.80271241, 2.47774945, 1.11414251,
            1.00347968, 2.15156538, 0.65570324, 0.5231689 , 2.63245244,
            1.9796373 , 2.19805588, 0.70110851, 2.33818557, 0.90772044,
            0.84479394, 1.78537261, 2.19521117, 1.11817448, 1.90068736,
            2.04316663, 1.65296857, 2.15212189, 0.91759358, 0.76945467,
            0.98915659, 2.36096369, 1.38276015, 0.86428906, 1.87937089,
            2.23663568, 2.1776537 , 1.43978589, 1.77779368, 2.29306394,
            1.3788257 , 0.81485191, 0.65995997, 0.71020451, 2.1822808 ,
            2.47812418, 0.9067476 , 2.65595263, 0.46611543, 1.73038152,
            0.89305432, 1.39403462, 0.93851792, 2.09934207, 2.18406997,
            1.45974351, 1.20308517, 1.69774085, 2.02890379, 1.44236564,
            2.14180396, 2.16675634, 1.72622341, 0.55420182, 2.18147231,
            0.91475956, 1.3910534 , 0.67953891, 2.50019276, 1.02469631,
            0.85997206, 0.26999289, 2.15911905, 2.63711985, 2.31942982,
            2.09306493, 2.9390576 , 0.8642066 , 0.95146373, 1.91639733,
            1.41930901, 0.71167437, 1.96570719, 1.22049595, 1.77612122,
            0.3335231 , 0.58774261, 1.81334148, 0.62008523, 0.89475238,
            1.15040084, 2.34482331, 0.37300647, 1.65967357, 1.96575657,
            0.50320774, 1.79942307, 2.24251415, 2.11552818, 0.97663611,
            1.70114442, 1.53776515, 0.74239722, 2.46944736, 1.24508309,
            0.9705605 , 1.64133836, 0.79287013, 2.36267897, 1.76010388,
            1.12659205, 2.15343821, 1.63883704, 1.59134054, 0.91505453,
            1.16893309, 0.47638906, 2.14777958, 0.43809548, 0.13252755,
            1.21708536, 1.78465189, 0.7528546 , 1.87088635, 0.60715976,
            0.91865922, 0.75122381, 0.78396458, 2.88214306, 1.76421523,
            1.58689834, 2.46287124, 2.37551272, 0.74961164, 2.34149992,
            2.19276542, 1.83402182, 2.32989978, 0.25525475, 2.51900063,
            1.79279969, 0.81391856, 1.79095151, 1.77307293, 2.72950402,
            1.88864462, 0.86027407, 1.82916189, 1.06092249, 1.46289744,
            1.49709851, 0.66301639, 0.77924025, 0.91384556, 1.811294  ,
            0.1276993 , 2.75092982, 2.35105093, 0.58939071, 2.18471084,
            1.78231963, 2.37516393, 1.23776859, 2.00949013, 2.31828135,
            2.14810196, 2.16868096, 1.53178233, 1.03077874, 1.10618988,
            1.75685873, 1.23964235, 2.33682559, 1.26223319, 2.75896605,
            0.4772932 , 0.35437454, 2.13014865, 2.00379702, 0.6771264 ,
            2.077447  , 2.06516742, 2.29955034, 0.9492543 , 2.34127067,
            2.41438855, 1.84730638, 2.37440368, 0.83354872, 1.94599066,
            1.9402149 , 2.72560449, 2.03462606, 2.53062944, 2.1949429 ,
            2.87193928, 1.61057202, 1.42689054, 1.61804163, 0.75940871,
            1.8539287 , 0.92894221, 0.77296501, 1.80376766, 1.98079022,
            2.02180853, 2.49251142, 1.35744012, 1.42560564, 0.61685182,
            1.39472516, 1.11573945, 0.81272365, 2.13968972, 0.86974214,
            1.93547254, 1.71278327, 0.96688879, 0.78864187, 0.94713233,
            2.66627438, 2.01992385, 2.2898705 , 1.60077403, 2.40170347,
            2.21992742, 1.5991143 , 2.17810733, 0.82794101, 0.28337881,
            2.70621422, 1.18341575, 1.77335606, 1.03963344, 1.97724589,
            1.39185332, 0.74868966, 1.76779094, 0.35063194, 1.40492085,
            1.08270857, 1.50385546, 2.19386743, 0.9313968 , 1.20041681,
            1.11613366, 2.53308361, 2.55426397, 1.69759207, 2.01964201,
            2.47572556, 2.33621761, 2.30231849, 2.34163468, 2.15703724,
            1.38199887, 1.53144935, 0.44028149, 1.04932209, 0.91289674,
            2.46872048, 1.1105031 , 1.3561764 , 1.57168665, 1.38284512,
            0.75672232, 1.08250932, 0.95224188, 1.18715148, 2.75779514,
            2.01260113, 0.48351781, 1.19450625, 2.64000735, 1.99364176,
            0.98931373, 1.55812588, 1.30538253, 1.48551471, 2.07780508,
            2.31695416, 0.94710817, 1.32388268, 1.69981202, 2.38133815,
            2.00307074, 2.35683828, 1.7821809 , 1.61342967, 1.04499877,
            1.37802239, 2.96502261, 1.78431234, 1.67188691, 2.23693274]), array([ 0.00000000e+00, -2.99729221e+00, -1.32239697e+00,  1.51887047e+00,
            -9.42639036e-01,  1.12525959e+00,  1.66504779e+00,  6.97584494e-01,
            -2.80240835e+00,  1.48973032e+00, -2.51346684e-01, -7.98195502e-01,
             3.09267947e+00,  2.34389877e+00,  1.91602782e+00,  2.90474375e+00,
            -2.20727943e+00, -1.74417610e+00,  8.91849207e-01,  8.45792477e-01,
            -1.13288599e+00, -1.18406697e+00, -1.71265255e+00, -2.52038376e+00,
             4.40006127e-01, -3.67319565e-01, -2.74684556e+00,  3.57066602e-01,
             6.90189973e-01, -1.95966833e+00,  8.72938420e-01,  2.24471860e+00,
            -1.44688775e+00,  2.71594405e+00, -2.66748115e+00, -1.16334695e+00,
            -2.79792913e+00,  2.14350878e+00, -9.88422605e-01, -3.24930838e-01,
            -1.84017708e+00, -9.01070823e-01, -1.94271288e+00,  2.75276356e+00,
            -7.45623482e-01, -1.58218857e+00, -6.65881799e-01,  2.48381201e+00,
            -2.23356165e+00,  1.04935114e+00, -7.46199907e-01, -7.93702876e-01,
             2.32411438e+00, -1.61857376e+00,  1.36483873e+00,  2.46479295e+00,
            -3.08378895e+00, -2.36818881e+00,  2.47561493e+00,  2.20897694e+00,
            -2.22159123e+00,  2.75745153e+00, -2.42497928e+00, -1.67405501e-01,
             1.64654311e+00, -5.00712290e-01,  4.97043409e-01,  2.66573488e+00,
             6.41423843e-01,  1.12613280e+00,  1.04440484e+00, -2.38209962e+00,
            -4.33598720e-01, -2.22383332e+00, -3.01804741e+00,  2.71902862e+00,
            -3.17920290e-01, -4.48153017e-01, -8.56598837e-01,  9.39942931e-01,
             2.10032318e+00, -7.31249051e-01, -2.17919854e+00,  3.03251157e+00,
            -4.09666526e-01,  2.83775854e+00, -1.20258340e+00,  2.24441522e+00,
            -3.02584720e+00,  1.73016630e+00, -2.20208584e+00,  5.39050002e-01,
             6.85654800e-01,  2.68348719e+00,  3.04734411e+00,  1.90033836e+00,
             6.99327659e-01,  3.18343212e-01, -1.26202480e+00,  1.83039949e+00,
            -9.76144674e-01,  1.81543047e+00,  1.80321981e+00,  5.13266032e-01,
             1.32153477e+00, -2.21061865e+00,  6.10758750e-01,  2.46960392e+00,
            -2.61983499e+00, -4.35767967e-01, -9.45337043e-01, -2.00186832e+00,
             6.49158932e-01, -7.71235256e-01,  2.99929662e+00,  2.19231481e+00,
            -5.34189973e-01,  3.24073104e-01, -9.20993938e-01, -1.56866040e+00,
             2.66507897e+00,  1.77363750e-01,  1.68520959e+00, -2.38872993e+00,
             1.40833356e+00, -2.46618739e+00,  2.46231699e+00, -3.09479007e+00,
             2.88551829e+00, -2.54379176e+00,  1.76983797e+00, -1.44084525e+00,
             1.19053942e+00,  2.99193253e+00, -2.43716871e-01, -2.34389574e+00,
            -2.58301054e+00,  1.86667583e+00, -2.98853147e+00,  3.00645762e+00,
            -2.19722451e+00,  2.47731413e+00,  2.35207308e+00, -1.52186508e+00,
             1.53939294e+00, -2.23275421e+00, -4.28197899e-02, -8.86769424e-01,
            -1.81032183e+00, -3.01125516e+00,  8.31446155e-01,  2.36994257e+00,
            -2.75210726e+00, -6.40398255e-01,  5.42118417e-01, -2.36736214e+00,
             2.44719798e+00,  1.18126282e+00,  2.49097745e+00, -1.20098037e+00,
            -2.19776430e+00, -2.35352408e+00, -6.69289160e-01, -4.48630289e-01,
             1.32365036e+00, -2.53507778e+00, -1.12973453e+00, -5.62474666e-01,
             1.14749878e+00,  7.89138869e-01, -1.61719200e+00, -1.12015097e+00,
            -1.14006344e+00,  1.41214381e+00,  7.01800276e-01,  4.20680951e-01,
            -2.65572364e+00,  7.53434149e-01,  3.04590999e+00, -2.19017329e+00,
             2.30769451e+00,  8.33037921e-01,  2.48177597e+00,  8.17980557e-01,
            -2.71223984e+00, -9.03154678e-01,  8.68453161e-01, -1.90093872e+00,
            -2.76684559e+00, -2.27854088e+00,  1.96652840e+00,  2.32666383e+00,
            -2.49101511e+00, -2.86323624e+00,  2.86891339e+00, -2.83385889e+00,
            -2.66833532e+00, -3.16354039e-01,  1.67428691e+00, -4.80051909e-01,
            -1.51673582e+00,  6.97344445e-02,  3.72373832e-01,  1.76259035e-01,
            -1.88219445e+00, -1.87449741e+00, -7.68713276e-01, -2.03117656e+00,
             2.73224027e+00,  2.63122887e+00, -1.82571270e+00, -2.44392594e+00,
            -6.95553336e-01,  3.55014062e-01, -1.90122778e+00,  1.92425160e+00,
             2.36310998e+00,  7.23944149e-01, -2.24254762e+00,  1.56405781e+00,
            -2.61683512e+00, -1.00342729e+00,  2.31693045e+00, -1.79711009e+00,
            -1.51065746e+00,  2.10422497e+00,  6.65758682e-01,  2.01395175e+00,
            -2.39698125e+00,  2.29405535e+00, -2.54066240e+00, -1.45764145e+00,
             2.90398495e+00,  2.47570149e+00,  5.56221653e-01,  2.47491850e+00,
             2.89369365e+00,  4.84501873e-01, -2.01988546e+00, -2.55176129e+00,
             1.92796494e+00,  2.63411045e-01,  1.17590235e-01, -2.70944565e+00,
            -3.13408597e+00, -3.02095940e+00, -1.77249166e+00, -2.52674071e+00,
            -1.88180456e-01,  1.63068214e+00,  2.11581135e+00, -1.34053292e+00,
            -2.08378903e+00, -2.44587174e+00,  2.00639321e+00,  2.81149579e+00,
            -2.21029280e+00,  1.61137232e+00, -2.87770939e+00, -2.15835008e-01,
            -1.60827069e+00, -8.15065308e-01,  2.26889754e+00, -1.63155728e+00,
             6.14573549e-01,  2.82075644e+00,  1.45933227e+00,  7.23217488e-01,
             2.93905655e+00,  1.79738020e+00, -3.05660008e+00,  2.30880550e+00,
             1.06442049e+00, -1.45473119e+00, -1.49589029e+00, -2.52721882e+00,
            -3.12877125e+00,  2.15333189e+00, -2.21751274e-01,  4.34557720e-01,
             2.63667549e+00,  1.09961189e+00,  1.68967369e+00,  2.37608409e+00,
            -4.55239810e-01,  1.94277250e+00,  2.46793206e+00, -2.53426890e+00,
             2.15872652e+00, -1.26932164e+00, -3.59448614e-01,  2.26258558e+00,
             4.72992731e-01,  2.89649384e+00,  1.74171690e+00, -2.95853671e+00,
            -1.40288912e+00, -1.43628464e+00,  2.21336461e+00, -2.06177104e+00,
            -1.76097467e+00,  8.91053210e-01,  2.06371690e+00, -2.42168496e+00,
             1.70018284e+00,  2.21744169e-01, -1.28810938e+00, -1.37912257e+00,
            -2.87339184e+00,  5.85683519e-01, -2.21824133e+00,  2.15596521e+00,
             1.57339281e+00,  1.89267898e+00, -2.41719499e+00, -2.97603316e+00,
            -2.85614889e-01, -3.07077824e+00,  9.68908449e-01,  8.07764606e-01,
            -2.73271863e+00,  2.42830377e+00,  2.30505133e+00, -7.08796982e-01,
            -1.08562684e+00, -9.90818263e-01, -2.90850602e+00, -1.32894064e+00,
             1.09461943e+00, -2.21680356e+00,  1.47077582e+00,  2.47637280e+00,
            -2.23388209e+00, -5.33804078e-01, -2.05451152e+00,  1.90838719e+00,
             4.92920730e-03, -9.66502880e-01,  2.93644167e-01,  2.48248928e+00,
            -1.14641716e-01,  2.80193062e+00, -2.57243424e+00,  8.92606776e-01,
            -2.21777573e+00, -3.04689659e+00, -2.23543268e+00,  5.46489428e-02,
            -1.96942332e+00, -2.03886036e+00,  2.15013537e+00, -1.03545321e+00,
             2.47585925e+00,  1.57257439e+00, -2.35236239e+00, -6.83073959e-01,
            -4.97177247e-01,  4.06591473e-01,  4.98525562e-01,  2.48385793e+00,
             1.34976100e+00, -2.15753124e+00,  2.46943716e+00, -1.56001449e+00,
            -1.15706590e+00, -8.20253771e-01,  9.78454604e-01, -2.17839476e+00,
            -2.38967151e-01,  1.01220773e+00,  1.29278026e+00, -2.85445344e+00,
            -9.56258905e-01,  7.72612572e-01, -2.42443785e+00,  7.62138251e-01,
            -2.22537704e+00,  6.44911506e-01,  1.28167515e+00, -4.14154565e-01,
            -1.76218976e+00,  7.83100544e-01, -7.32265035e-01,  1.19584842e+00,
            -4.33207696e-02,  2.22022148e+00,  2.91581954e-02, -5.85111089e-01,
            -3.84870644e-01,  1.36577342e+00,  1.03532555e+00,  1.37907373e+00,
            -2.54127778e+00,  2.46937035e+00, -1.20084859e+00, -2.03569106e+00,
             2.00281569e+00,  6.00219500e-01,  2.66567778e+00,  2.03204122e+00,
            -2.10226305e+00,  2.05171728e+00, -1.30245974e+00,  3.99840192e-01,
            -2.57155254e+00,  2.16989098e+00, -2.88827088e+00,  1.41385368e+00,
            -1.03999328e-01, -2.04458458e+00,  2.08664475e+00, -1.34298916e-01,
             9.35203330e-01,  5.12290424e-01, -2.21215385e+00, -2.70316091e+00,
            -6.96345105e-01,  1.17825812e+00, -1.20306060e+00, -9.43208700e-01,
             1.36433291e-03, -2.74329641e+00, -1.03791684e+00,  1.83212413e-01,
            -7.37562156e-01, -1.95361742e+00, -1.66154407e+00,  2.23150660e+00,
            -9.19875868e-01, -2.38401802e-01, -5.93782112e-01, -7.45481849e-01,
             3.08512524e+00,  1.00028372e+00,  2.08362366e+00,  2.44710980e+00,
             5.40786782e-01,  1.25085421e+00,  1.92519512e+00, -2.91872741e-02,
            -5.58849229e-01, -5.33071055e-01,  2.67821317e+00,  3.13288790e+00,
             7.73220909e-01, -8.79006862e-01, -1.03560798e+00, -2.94444154e+00,
            -1.87078718e+00,  1.04728590e+00,  1.50237065e+00,  1.20113420e+00,
             2.46473452e+00,  1.71029852e+00, -7.86643787e-01,  2.27136171e+00,
             8.34624258e-01, -2.63757334e+00, -4.97889756e-01, -1.38339540e+00,
            -6.87090879e-01, -1.37273260e+00,  8.24785961e-01,  5.56740645e-02,
             2.32708692e+00, -1.07611994e+00, -8.35281566e-01,  1.11224599e+00,
             1.97733884e+00,  9.61717333e-01,  1.55077629e+00,  6.24233707e-01,
            -3.01444235e-01,  7.81285929e-01,  2.48148211e+00, -1.66004000e+00,
             5.45851919e-01,  1.17437789e+00, -2.03052952e+00, -5.60767197e-01,
             5.89675946e-01,  3.17495528e-01, -1.27153168e-01, -1.67766441e+00,
             2.93530477e+00,  1.18458516e+00, -2.77439786e+00,  1.28169107e+00,
            -8.71338202e-01, -9.85736731e-01,  2.27034153e+00, -1.74534548e+00,
             2.98366747e+00,  2.50251561e+00, -3.47619040e-01, -1.09828294e+00,
             2.75090590e-01, -1.02713709e+00, -1.13625145e-01,  3.17353750e-01,
             1.06991648e-01,  9.26368935e-01,  2.45803671e+00, -7.20747518e-01,
             1.15243640e+00,  3.12817983e-01,  5.29084513e-02,  2.41606871e+00,
            -6.25851494e-01, -7.98460973e-01, -1.08547036e+00,  9.95684524e-01,
             2.77356530e+00,  2.15372294e+00, -3.15174223e-02,  1.81092793e+00,
            -5.38714318e-01,  1.47165951e-01,  1.30664587e+00, -2.57361178e+00,
             1.64878302e+00, -6.34515656e-01,  9.10702818e-01,  1.27921121e+00,
             2.65846926e+00,  2.45348962e+00,  4.44149352e-02,  6.75189287e-01,
             1.54321306e-01,  3.80716290e-01, -1.30317636e+00,  3.10940192e+00,
            -1.83717001e+00, -8.16877080e-01, -8.51124222e-01, -2.36046052e+00,
             2.47195378e+00, -2.19352915e+00,  1.75905504e-01, -6.70446604e-01]))
    """
    ...


def neighbour_lists(atoms: Atoms, r_cut: float, filter: Optional[fastatomstruct.Filter] = None) -> List[np.ndarray]:
    """List of nearest neighbours for each atom.

    Currently only supports one cutoff radius.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_cut (float): cutoff radius
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        List of 1D arrays

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> n = fs.neighbour_lists(atoms, 3.2)
    >>> len(n)
    540
    >>> n[0]
    array([  2, 206, 338, 443], dtype=uint64)
    """
    ...


def coordination_numbers(atoms: Atoms, r_cut: float, filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Coordination number of each atom.

    Currently only supports one cutoff radius.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_cut (float): cutoff radius
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        1D array of length n containing the coordination numbers

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.coordination_numbers(atoms, 3.2)
    array([4, 3, 4, 5, 4, 4, 5, 3, 3, 4, 2, 3, 3, 5, 4, 3, 3, 3, 3, 3, 5, 2,
           5, 4, 4, 2, 3, 3, 3, 4, 4, 4, 2, 4, 5, 5, 5, 5, 3, 4, 4, 4, 3, 3,
           3, 4, 2, 3, 4, 4, 3, 3, 5, 3, 3, 6, 4, 4, 3, 3, 4, 3, 4, 4, 3, 3,
           4, 4, 5, 3, 4, 3, 4, 3, 4, 4, 3, 3, 4, 5, 4, 4, 4, 4, 4, 3, 3, 3,
           4, 5, 3, 3, 4, 4, 3, 4, 4, 3, 4, 4, 4, 4, 4, 5, 5, 3, 5, 5, 4, 4,
           3, 4, 3, 3, 3, 4, 4, 3, 3, 3, 4, 3, 4, 5, 3, 4, 4, 3, 3, 3, 3, 3,
           4, 3, 3, 2, 5, 2, 4, 4, 4, 4, 4, 4, 3, 3, 3, 4, 3, 4, 3, 1, 3, 5,
           5, 3, 4, 4, 3, 3, 4, 5, 5, 3, 4, 3, 3, 3, 2, 5, 4, 4, 2, 4, 3, 3,
           3, 4, 3, 3, 5, 3, 3, 4, 4, 4, 3, 3, 3, 4, 3, 4, 3, 4, 2, 2, 4, 4,
           4, 4, 3, 5, 4, 3, 3, 3, 5, 3, 3, 4, 3, 4, 4, 3, 3, 3, 3, 4, 3, 2,
           4, 3, 4, 3, 4, 4, 2, 4, 3, 4, 4, 3, 2, 5, 4, 4, 4, 2, 2, 4, 4, 4,
           3, 4, 3, 3, 3, 3, 4, 3, 5, 4, 3, 3, 3, 4, 4, 2, 3, 2, 2, 2, 5, 3,
           3, 4, 3, 4, 3, 3, 4, 5, 5, 4, 2, 2, 4, 5, 5, 2, 4, 4, 5, 2, 5, 4,
           4, 3, 3, 4, 3, 3, 5, 3, 3, 4, 3, 3, 4, 4, 3, 5, 4, 5, 2, 4, 3, 4,
           3, 3, 3, 3, 6, 4, 5, 3, 3, 4, 4, 3, 4, 2, 4, 3, 4, 4, 3, 3, 3, 5,
           3, 3, 4, 2, 4, 4, 4, 3, 3, 3, 4, 3, 5, 3, 3, 5, 2, 4, 3, 4, 4, 3,
           4, 3, 4, 2, 4, 3, 4, 4, 5, 2, 4, 5, 4, 3, 4, 3, 4, 3, 4, 3, 3, 3,
           4, 4, 3, 4, 2, 4, 3, 4, 4, 3, 3, 3, 4, 3, 4, 4, 3, 4, 4, 5, 4, 3,
           4, 3, 4, 3, 4, 4, 5, 3, 4, 4, 4, 3, 5, 4, 4, 5, 3, 5, 4, 4, 2, 3,
           3, 3, 3, 3, 2, 4, 3, 4, 4, 5, 3, 3, 5, 3, 3, 3, 4, 4, 5, 4, 3, 4,
           4, 4, 3, 3, 2, 3, 5, 3, 4, 5, 3, 5, 4, 5, 3, 4, 4, 4, 4, 4, 4, 4,
           3, 3, 4, 4, 3, 3, 4, 4, 3, 3, 3, 2, 3, 4, 4, 3, 4, 4, 4, 4, 4, 3,
           3, 3, 2, 4, 3, 5, 4, 5, 4, 4, 3, 3, 4, 3, 4, 3, 4, 3, 2, 4, 4, 4,
           3, 3, 6, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 3, 3, 2, 5, 4, 4, 3, 3,
           2, 5, 4, 5, 4, 5, 4, 4, 4, 4, 4, 3], dtype=uint64)
    """
    ...


def bond_angle_distribution(atoms: Atoms, r_cut: float,
                            n_bins: int, filter: Optional[fastatomstruct.Filter] = None) -> np.ndarray:
    """Distribution of bond angles.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE
        r_cut (float): cutoff radius
        n_bins (int): Number of bins for the bond angle histogram
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        Two 1D arrays of length n_bins containing the bond angles and histogram of bond angles.

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.bond_angle_distribution(atoms, 3.2, 180)
    (array([  0.,   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.,
             11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,  20.,  21.,
             22.,  23.,  24.,  25.,  26.,  27.,  28.,  29.,  30.,  31.,  32.,
             33.,  34.,  35.,  36.,  37.,  38.,  39.,  40.,  41.,  42.,  43.,
             44.,  45.,  46.,  47.,  48.,  49.,  50.,  51.,  52.,  53.,  54.,
             55.,  56.,  57.,  58.,  59.,  60.,  61.,  62.,  63.,  64.,  65.,
             66.,  67.,  68.,  69.,  70.,  71.,  72.,  73.,  74.,  75.,  76.,
             77.,  78.,  79.,  80.,  81.,  82.,  83.,  84.,  85.,  86.,  87.,
             88.,  89.,  90.,  91.,  92.,  93.,  94.,  95.,  96.,  97.,  98.,
             99., 100., 101., 102., 103., 104., 105., 106., 107., 108., 109.,
             110., 111., 112., 113., 114., 115., 116., 117., 118., 119., 120.,
             121., 122., 123., 124., 125., 126., 127., 128., 129., 130., 131.,
             132., 133., 134., 135., 136., 137., 138., 139., 140., 141., 142.,
             143., 144., 145., 146., 147., 148., 149., 150., 151., 152., 153.,
             154., 155., 156., 157., 158., 159., 160., 161., 162., 163., 164.,
             165., 166., 167., 168., 169., 170., 171., 172., 173., 174., 175.,
             176., 177., 178., 179.]),
      array([  0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,
               0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,
               0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,
               0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,
               0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,
               0.,   0.,   0.,   0.,   0.,   0.,   0.,   2.,   2.,   0.,   0.,
               0.,   0.,   0.,   0.,   0.,   0.,   4.,   4.,   2.,   4.,  12.,
               6.,  12.,  24.,  36.,  40.,  60.,  72., 142., 138., 156., 210.,
       ...
               0.,   0.,   0.,   0.,   0.,   0.,   0.,   2.,   0.,   0.,   4.,
               0.,   2.,   0.,   4.,   0.,   0.,   0.,   0.,   2.,   6.,   2.,
               4.,   8.,   4.,   6.,   6.,  10.,   4.,   8.,  26.,  26.,  42.,
               50.,  38.,  48.,  56.,  52.,  56.,  52.,  76.,  50.,  50.,  40.,
               36.,  16.,  10.,   4.]))
    """
    ...


def radial_distribution_function(atoms: Atoms, r_max: float,
                                 n_bins: int, filter: Optional[fastatomstruct.Filter] = None) -> Tuple[np.ndarray, np.ndarray]:
    """Radial distribution function (RDF).

    Estimated using a rectangular window function. The RDF is defined as

    .. math::

        g(r) = \frac{d n_r}{4 \pi \rho r^2 dr}.

    Arguments:
        atoms (ase.Atoms): Atoms object from ASE or list of Atoms objects
        r_max (float): Maximum radius for the RDF computation
        n_bins (int): Number of bins
        filter (fastatomstruct.Filter): Filter instance

    Returns:
        (1D array of length n_bins containing the radii, 1D array of length n_bins containing the RDF)

    Examples
    --------

    The exemplary file "Sb540.traj" `can be found here <https://zivgitlab.uni-muenster.de/ag-salinga/fastatomstruct/-/raw/master/regressiontesting/Structures/Sb540.traj>`_.

    >>> import fastatomstruct as fs
    >>> from ase import io
    >>> atoms = io.read("Sb540.traj")
    >>> fs.radial_distribution_function(atoms, 10, 100)
    (array([0. , 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1. , 1.1, 1.2,
           1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2. , 2.1, 2.2, 2.3, 2.4, 2.5,
           2.6, 2.7, 2.8, 2.9, 3. , 3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7, 3.8,
           3.9, 4. , 4.1, 4.2, 4.3, 4.4, 4.5, 4.6, 4.7, 4.8, 4.9, 5. , 5.1,
           5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9, 6. , 6.1, 6.2, 6.3, 6.4,
           6.5, 6.6, 6.7, 6.8, 6.9, 7. , 7.1, 7.2, 7.3, 7.4, 7.5, 7.6, 7.7,
           7.8, 7.9, 8. , 8.1, 8.2, 8.3, 8.4, 8.5, 8.6, 8.7, 8.8, 8.9, 9. ,
           9.1, 9.2, 9.3, 9.4, 9.5, 9.6, 9.7, 9.8, 9.9]), array([       nan, 0.        , 0.        , 0.        , 0.        ,
           0.        , 0.        , 0.        , 0.        , 0.        ,
           0.        , 0.        , 0.        , 0.        , 0.        ,
           0.        , 0.        , 0.        , 0.        , 0.        ,
           0.        , 0.        , 0.        , 0.        , 0.        ,
           0.        , 0.01357761, 0.13849533, 1.35803798, 2.55381745,
           3.11047932, 2.85573424, 1.90022878, 1.50024473, 1.06393955,
           0.53946888, 0.31869665, 0.26147559, 0.20975713, 0.46465595,
           0.69412131, 1.1903064 , 1.83673335, 2.37776314, 2.71655972,
           2.17563581, 1.52251451, 1.07615304, 0.66129555, 0.47402311,
           0.52500812, 0.68106246, 0.80447332, 0.85609023, 0.97261497,
           0.78282433, 0.5853612 , 0.52827722, 0.35742531, 0.40341997,
           0.51246422, 0.73013309, 0.97897244, 0.98282868, 0.8067009 ,
           0.83203587, 0.90815376, 1.24724054, 1.54231535, 1.71963653,
           1.68396709, 1.66781841, 1.47308674, 1.37444435, 1.47666662,
           1.45060517, 1.42698412, 1.45517891, 1.22349342, 0.99270356,
           0.68408236, 0.49522574, 0.36855817, 0.33841338, 0.35511913,
           0.41287206, 0.55721067, 0.63542276, 0.75380977, 0.82271293,
           0.98696812, 0.97537108, 1.08549646, 1.19705248, 1.21846285,
           1.19599703, 1.29171737, 1.34423667, 1.40868964, 1.46559489]))
    """


def mean_squared_displacement(atoms: List[Atoms]) -> np.ndarray:
    """Mean squared displacement (MSD) with temporal averaging.

    The (averaged) MSD is defined as

    .. math::

        \mathrm{MSD} = \langle |\vec{x}(t) - \vec{x}(t_0)|^2 \rangle_{N, t_0} = \left\langle \frac{1}{N} \sum_{i = 1}^N |\vec{x}^i(t) - \vec{x}^i(t_0)|^2 \right\rangle_{t_0}

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE

    Returns:
        1D array containing the MSD
    """
    ...


def mean_squared_displacement_single(atoms: List[Atoms]) -> np.ndarray:
    """Mean squared displacement (MSD) without temporal averaging.

    The (non-averaged) MSD is defined as

    .. math::

        \mathrm{MSD} = \langle |\vec{x}(t) - \vec{x}(t_0)|^2 \rangle = \frac{1}{N} \sum_{i = 1}^N |\vec{x}^i(t) - \vec{x}^i(t_0)|^2

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE

    Returns:
        1D array containing the MSD
    """


def squared_displacement_single(atoms: List[Atoms]) -> np.ndarray:
    """Squared displacement (SD) for each atom without temporal averaging.

    The (non-averaged) SD is defined as

    .. math::

        \mathrm{SD}_i = |\vec{x}^i(t) - \vec{x}^i(t_0)|^2 = |\vec{x}^i(t) - \vec{x}^i(t_0)|^2

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE

    Returns:
        2D array of shape (n_atoms, n_times) containing the SD
    """
    ...


def non_gaussian_alpha2(atoms: List[Atoms]) -> np.ndarray:
    """(Temporally averaged) non-Gaussian parameter.

    It is defined as

    .. math::

        \alpha_2(t) = \left\langle \frac{3}{5} \frac{\sum_{i = 1}^N |\vec{x}^i(t) - \vec{x}^i(t_0)|^4}{(\sum_{i = 1}^N |\vec{x}^i(t) - \vec{x}^i(t_0)|^2)^2} \right\rangle_{t_0}

    can be used to identify non-Gaussian behaviour in molecular dynamics.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE

    Returns:
        1D array containing the non-Gaussian parameter for each time step
    """
    ...


def non_gaussian_alpha2_single(atoms: List[Atoms]) -> np.ndarray:
    """(Non-averaged) non-Gaussian parameter.

    It is defined as

    .. math::

        \alpha_2(t) = \frac{3}{5} \frac{\sum_{i = 1}^N |\vec{x}^i(t) - \vec{x}^i(t_0)|^4}{(\sum_{i = 1}^N |\vec{x}^i(t) - \vec{x}^i(t_0)|^2)^2}

    can be used to identify non-Gaussian behaviour in molecular dynamics.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE

    Returns:
        2D array of shape (n_atoms, n_times) containing the non-Gaussian
        parameter for each time step and atom
    """
    ...


def overlap_q(atoms: List[Atoms], a: float) -> np.ndarray:
    """Time-averaged overlap parameter.

    It is defined as

    .. math::

        Q(t) = \left\langle \sum_{i, j = 1}^N w(|\vec{r}_i(t_0) - \vec{r}_j(t)|) \right\rangle_{t_0}

    (see Lacevic et al., "Spatially heterogeneous dynamics investigated via a time-dependent four-point density correlation function").
    Here, :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero otherwise.
    The overlap parameter can be split into a self-part

    .. math::

        Q_\mathrm{s}(t) = \left\langle \sum_{i = 1}^N w(|\vec{r}_i(0) - \vec{r}_i(t)|) \right\rangle_{t_0}

    and a distinct part

    .. math::

        Q_\mathrm{d}(t) = \left\langle \sum_{i, j = 1, i \neq j}^N w(|\vec{r}_i(0) - \vec{r}_j(t)|) \right\rangle_{t_0}.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Tuple of three arrays of length n_times (full q, self part, distinct part)
    """
    ...


def overlap_q_self(atoms: List[Atoms], a: float) -> np.ndarray:
    """Self part of the overlap parameter.

    It is defined as

    .. math::

        Q_\mathrm{s}(t) = \left\langle \sum_{i = 1}^N w(|\vec{r}_i(0) - \vec{r}_i(t)|) \right\rangle_{t_0},

    where :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero otherwise.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Array of length n_times containing the self part of the overlap parameter
    """
    ...


def overlap_q_self_atomic(atoms: List[Atoms], a: float) -> np.ndarray:
    """Self part of the overlap parameter for each individual atom.

    It is defined as

    .. math::

        Q_\mathrm{s}(i, t) = \left\langle w(|\vec{r}_i(0) - \vec{r}_i(t)|) \right\rangle_{t_0},

    where :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero otherwise.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Array of shape (n_atoms, n_times) containing the self part of the overlap parameter
    """
    ...


def overlap_q_distinct(atoms: List[Atoms], a: float) -> np.ndarray:
    """Distinct part of the overlap parameter.

    It is defined as

    .. math::

        Q_\mathrm{d}(t) = \left\langle \sum_{i, j = 1, i \neq j}^N w(|\vec{r}_i(0) - \vec{r}_j(t)|) \right\rangle_{t_0},

    where :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero otherwise.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Array of length n_times containing the self part of the overlap parameter
    """
    ...


def overlap_q_single(atoms: List[Atoms], a: float) -> \
        Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Non-averaged overlap parameter.

    It is defined as

    .. math::

        Q(t) = \sum_{i, j = 1}^N w(|\vec{r}_i(0) - \vec{r}_j(t)|)

    (see Lacevic et al., "Spatially heterogeneous dynamics investigated via a
    time-dependent four-point density correlation function").
    Here, :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero
    otherwise. The overlap parameter can be split into a self-part

    .. math::

        Q_\mathrm{s}(t) = \sum_{i = 1}^N w(|\vec{r}_i(0) - \vec{r}_i(t)|)

    and a distinct part

    .. math::

        Q_\mathrm{d}(t) = \sum_{i, j = 1, i \neq j}^N w(|\vec{r}_i(0) - \vec{r}_j(t)|).

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Tuple of three arrays of length n_times (full q, self part, distinct part)
    """
    ...


def overlap_q_single_self(atoms: List[Atoms], a: float) -> np.ndarray:
    """Self part of the overlap parameter.

    It is defined as

    .. math::

        Q_\mathrm{s}(t) = \sum_{i = 1}^N w(|\vec{r}_i(0) - \vec{r}_i(t)|),

    where :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero
    otherwise.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Array of length n_times containing the self part of the overlap parameter
    """
    ...


def overlap_q_single_distinct(atoms: List[Atoms], a: float) -> np.ndarray:
    """Distinct part of the overlap parameter.

    It is defined as

    .. math::

        Q_\mathrm{d}(t) = \sum_{i, j = 1, i \neq j}^N w(|\vec{r}_i(0) - \vec{r}_j(t)|),

    where :math:`w(x)` is an "overlap" function that is one for :math:`x \leq a` and zero
    otherwise.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function

    Returns:
        Array of length n_times containing the self part of the overlap parameter
    """
    ...


def fourpoint_susceptibility(
    atoms: List[Atoms], a: float, temperature: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Four-point susceptibility.

    It is defined as

    .. math::

        \chi_4(t) = \frac{\beta V}{N^2} \left[\langle Q^2(t) \rangle - \langle Q(t) \rangle^2\right]

    and gives information about dynamical heterogeneities. It can be split into three different parts

    .. math::

        \chi_4(t) = \chi_\mathrm{ss}(t) + \chi_\mathrm{dd}(t) + \chi_\mathrm{sd}(t),

    (see Lacevic et al., "Spatially heterogeneous dynamics investigated via a time-dependent four-point density correlation function"), where :math:`\chi_\mathrm{ss}(t) \propto \langle Q_\mathrm{s}^2(t)\rangle - \langle Q_\mathrm{s}(t) \rangle^2`, :math:`\chi_\mathrm{dd}(t) \propto \langle Q_\mathrm{d}^2(t)\rangle - \langle Q_\mathrm{d}(t) \rangle^2`, and :math:`\chi_\mathrm{sd}(t) \propto \langle Q_\mathrm{s}(t) Q_\mathrm{d}(t) \rangle - \langle Q_\mathrm{s}(t) \rangle \langle Q_\mathrm{d}(t) \rangle`.

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function
        temperature (float): Temperature of the system (in K)

    Returns:
        Array of length n_times containing the four-point susceptibility
    """
    ...


def fourpoint_susceptibility_self(
        atoms: List[Atoms], a: float, temperature: float) -> np.ndarray:
    """Self part of the four-point susceptibility.

    It is defined as

    .. math::

        \chi_\mathrm{ss}(t) = \frac{\beta V}{N^2} \left[\langle Q_\mathrm{s}^2(t) \rangle - \langle Q_\mathrm{s}(t) \rangle^2\right]

    Arguments:
        atoms (List[ase.Atoms]): List of Atoms objects from ASE
        a (float): Cutoff for the overlap function
        temperature (float): Temperature of the system (in K)

    Returns:
        Array of length n_times containing the self part of the four-point susceptibility
    """
    ...
